########################################################################
# metaf2xml/XML.pm -- write data of METAR, TAF, and/or SYNOP message to XML file
#
# copyright (c) metaf2xml 2006-2008
#
# This program is free software; you can redistribute it and/or
# modify it under the terms of the GNU General Public License as
# published by the Free Software Foundation; either version 2 of the
# License, or (at your option) any later version.
#
# This program is distributed in the hope that it will be useful, but
# WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
# General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program; if not, write to the Free Software
# Foundation, 59 Temple Place, Suite 330, Boston, MA 02111-1307, USA
########################################################################

package metaf2xml::XML;

########################################################################
# some things strictly perl
########################################################################
use strict;
use warnings;

use IO::File;
use XML::Writer;

########################################################################
# export the functions provided by this module
########################################################################
BEGIN {
    use Exporter();
    our ($VERSION, @ISA, @EXPORT, @EXPORT_OK, %EXPORT_TAGS);

    $VERSION     = sprintf "%d.%03d", '$Revision: 1.32 $0.0' =~ /(\d+)/g;
    @ISA         = qw(Exporter);
    @EXPORT      = ();
    %EXPORT_TAGS = ();
    @EXPORT_OK   = qw(startXML printReport finishXML);
}
our @EXPORT_OK;
my $VERSION_ID = 'XML.pm: $Id: XML.pm,v 1.32 2008/04/06 06:54:24 metaf2xml Exp $';

END { }

=head1 NAME

metaf2xml::XML -- write data of METAR, TAF, and/or SYNOP messages to an XML file

=head1 SYNOPSIS

  use metaf2xml::XML;
  metaf2xml::XML::startXML(\%opts,\@versions);
  metaf2xml::XML::printReport(\%report);
  ...
  metaf2xml::XML::finishXML();

=head1 DESCRIPTION

This module writes the data of METAR, TAF, and/or SYNOP messages to an XML file.

=head1 DEPENDENCIES

The modules C<IO::File> and C<XML::Writer> are required.

=head1 FUNCTIONS

=cut

########################################################################
# the one and only global variable, the IO handle for XML::Writer
########################################################################
my $writer;

########################################################################
# helper functions
########################################################################

sub _printXML;
sub _printXML {
    my ($r, $tag, $xml_tag) = @_;

    return unless exists $r->{$tag};
    $r = $r->{$tag};

    $tag = $xml_tag if $xml_tag;
    if (ref $r eq 'HASH') {
        my (@subnodes, @attrs);

        # sort, so XML looks nicer
        for (qw/errorType warningType s v u q/) {
            push @attrs, $_, $r->{$_} if exists $r->{$_};
        }

        # special tag sequence to comply with the sequence of tags in DTD
        if ($tag eq 'cloudOpacityLvl') {
            @subnodes = qw/locationAnd oktas weather cloudType cloudBase/;
        } elsif ($tag =~ /^(?:tempMinNighttime|tempMaxDaytime|tempMinPeriod|tempMaxPeriod)$/) {
            @subnodes = qw/period notAvailable temp/;
        } elsif ($tag eq 'pressureChange') {
            @subnodes = qw/period notAvailable pressureChangeVal pressureTendency/;
        } elsif ($tag eq 'reportModifier') {
            @subnodes = qw/modifierType isLastSegment segment bulletinSeq sequenceLost over24hLate/;
        } elsif ($tag =~ /^(?:phenomenonAtLoc|phenomenonOnly)$/) {
            @subnodes = qw/phenomDescrPre phenomDescrPost weather cloudType cloudCover lightningType otherPhenom locationAnd MOV MOVD cloudTypeAsoctd cloudTypeEmbd/;
        } elsif ($tag eq 'cloudInfo') {
            @subnodes = qw/cloudBaseNotAvailable cloudBase cloudBaseFrom cloudBaseTo visVert visVertFrom visVertTo cloudTypeNotAvailable cloudType totalCloudCover/;
        } elsif ($tag eq 'cloudLocation') {
            @subnodes = qw/cloudTypeNotAvailable cloudType cloudNone cloudInvisible cloudDir topsInvisible elevationAngle/;
        } elsif ($tag =~ /^(?:radShortWave|radDirectSolar)$/) {
            @subnodes = qw/radiationPeriod notAvailable radiationValue/;
        } elsif ($tag eq 'location') {
            @subnodes = qw/inDistance inVicinity distance isGrid isQuadrant obscgMtns locationSpec compassDir quadrant/;
        } elsif ($tag eq 'climate') {
            @subnodes = qw/precipAmountTraces precipAmount1Inch precipAmount1MM precipAmount2Inch precipAmount2MM temp1 temp2/;
        } else {
            for (sort keys %$r) {
                if (!/^(?:[vuqs]|errorType|warningType)$/) {
                    push @subnodes, $_;
                }
            }
        }

        # *Arr are arrays with different subnodes. Suppress tag in XML.
        if ($#subnodes > -1) {
            $writer->startTag($tag, @attrs) unless $tag =~ /Arr$/;;
            for (@subnodes) {
                _printXML $r, $_;
            }
            $writer->endTag($tag) unless $tag =~ /Arr$/;;
        } else {
            $writer->emptyTag($tag, @attrs);
        }
    } elsif (ref $r eq 'ARRAY') {
        if ($#$r > -1) {
            for (@$r) {
                _printXML { $tag => $_ }, $tag;
            }
        } else {
            $writer->emptyTag($tag);
        }
    } elsif ($tag eq 'version') {
        $writer->dataElement('version', $r);
    } else {
        $writer->emptyTag($tag, defined $r ? ('v',  $r) : ());
    }
}

########################################################################
# variables shared by startXML, finishXML
########################################################################
my ($opt_X, $output);

########################################################################
# startXML
########################################################################

=head2 startXML(\%opts,\@versions)

This function opens the XML file and writes the necessary meta data into it.

The following arguments are expected:

=over

=item C<opts>

Reference to hash of options. The following keys of the hash are recognised:

=for html <ul>

=over

=item C<x>, with value B<E<lt>filenameE<gt>>

print parsed content of message as XML to <filename>

=item C<X>

with option B<x>, produce complete XML file, not just data

=item C<D>

with option B<X>, add DOCTYPE, reference to DTD

=item C<S>, with value B<E<lt>filenameE<gt>>

with option B<X>, add reference to stylesheet <filename>

=item C<O>, with value B<E<lt>optionsE<gt>>

with option B<S>, add <options> (a space separated list of the options passed to
metaf2xml.pl) to XML file

=for html </ul>

=back

=item C<versions>

Reference to array of version strings. They are written to the XML file if the
option C<X> is given.

=back

=cut

sub startXML {
    my ($opts, $versions) = @_;

    $output = new IO::File(">$opts->{x}");
    $writer = new XML::Writer(OUTPUT=>$output, DATA_MODE=>1, DATA_INDENT=>1);

    # XML header
    $writer->xmlDecl("UTF-8");

    # write additional info to XML, if required
    $opt_X = exists $opts->{X};
    if ($opt_X) {

        # XML meta data
        $writer->doctype('data', undef, 'metaf.dtd') if defined $opts->{D};
        $writer->pi('xml-stylesheet',
                    'href="' . $opts->{S} . '" type="text/xsl"')
            if defined $opts->{S};

        # opening data tag
        $writer->startTag('data');

        # version stuff
        _printXML { version => [ @$versions, $VERSION_ID ] }, 'version';

        if (exists $opts->{O}) {
            my (%options, $msg);
            (@options{'type_metaf', 'type_synop', 'lang', 'format', 'src_metaf', 'src_synop', 'mode'}, $msg)
                                                     = split / /, $opts->{O}, 8;
            @options{'msg_metaf', 'msg_synop'} = split '  ', $msg, 2;
            _printXML { options => \%options }, 'options';
        }
    }
    $writer->startTag('reports', query_start => gmtime().'');
    return;
}

########################################################################
# printReport
########################################################################

=head2 printReport(\%report)

This function writes the content of C<report> to the XML file.
The data for each message will be enclosed in a tag C<<metar>>, C<<taf>>, or
C<<synop>>. The function can be invoked repeatedly.
L<startXML()|/startxml___opts___versions_> must have been invoked already.
L<finishXML()|/finishxml__> must be invoked to complete the XML file.

The following argument is expected:

=over

=item C<report>

reference to hash of data to be written to the file

=back

=cut

sub printReport {
    my $report = shift;

    return unless $writer;

    if (exists $report->{isSynop}) {
        $writer->startTag('synop', 's', $report->{msg});

        for (qw/ERROR warning
                obsStationType callSign obsTime reportModifier
                windIndicator stationPosition stationId precipInd wxInd
                baseLowestCloud visPrev visibilityAtLoc totalCloudCover
                sfcWind temperature stationPressure SLPhPa pressureChange
                precipHourly weatherSynop cloudTypes exactObsTime/)
        {
            _printXML $report, $_;
        }

        if (exists $report->{section3}) {
            if ($#{$report->{section3}} > -1) {
                $writer->startTag('section3');
                _printXML { section3Arr => $report->{section3}}, 'section3Arr';
                $writer->endTag('section3');
            } else {
                $writer->emptyTag('section3');
            }
        }

        $writer->endTag('synop');
    } else {
        my $is_taf;

        $is_taf = exists $report->{isTaf};
        $writer->startTag($is_taf ? 'taf' : 'metar', 's', $report->{msg});
        for (qw/ERROR warning isSpeci
                obsStation obsTime issueTime fcstPeriod reportModifier
                fcstNotAvbl
                sfcWind CAVOK visPrev visMin visRwy RVRNO weather cloud visVert
                temperature QNH
                somePressure cloudMaxCover recentWeather windShear rwyState
                colourCode NEFO_PLAYA RH windAtLoc rwyWind TAFsupplArr/)
        {
            _printXML $report, $_;
        }

        if (exists $report->{trend}) {
            for my $td (@{$report->{trend}}) {
                $writer->startTag('trend', 's', $td->{s});
                for (qw/trendType trendTime1 trendTime2 probability
                        sfcWind CAVOK visPrev visVert weather cloud
                        rwyState colourCode TAFsupplArr/)
                {
                    _printXML $td, $_;
                }
                $writer->endTag('trend');
            }
        }

        _printXML $report, 'TAFinfoArr';
        _printXML $report, 'remark', $is_taf ? 'tafRemark' :'remark';

        $writer->endTag($is_taf ? 'taf' : 'metar');
    }
    return;
}

########################################################################
# finishXML
########################################################################

=head2 finishXML()

This function must be invoked to complete the writing to the XML file.
L<startXML()|/startxml___opts___versions_> must have been invoked already.

No arguments are expected.

=cut

sub finishXML {
    return unless $writer;

    $writer->endTag('reports');

    # closing data tag, if required
    $writer->endTag('data') if $opt_X;

    # close XML writer and file
    $writer->end();
    $output->close();
    $writer = undef;
    $output = undef;
    return;
}

=head1 SEE ALSO

=begin html

<a href=parser.pm.html>metaf2xml::parser(3pm)</a>,
<a href=metaf2xml.pl.html>metaf2xml(1)</a>,
<a href=metaf.pl.html>metaf(1)</a>, <!--

=end html

B<metaf2xml::parser>(3pm),
B<metaf2xml>(1),
B<metaf>(1),

=for html -->

L<http://metaf2xml.sourceforge.net/>

=head1 COPYRIGHT, LICENSE, DISCLAIMER

copyright (c) 2006-2008 metaf2xml @ L<http://metaf2xml.sourceforge.net/>

This program is free software; you can redistribute it and/or
modify it under the terms of the GNU General Public License as
published by the Free Software Foundation; either version 2 of the
License, or (at your option) any later version.

This program is distributed in the hope that it will be useful, but
WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
General Public License for more details.

You should have received a copy of the GNU General Public License
along with this program; if not, write to the Free Software
Foundation, 59 Temple Place, Suite 330, Boston, MA 02111-1307, USA

=cut

1;
