########################################################################
# metaf2xml/parser.pm -- parse a METAR, TAF, or SYNOP message
#
# copyright (c) metaf2xml 2006-2008
#
# This program is free software; you can redistribute it and/or
# modify it under the terms of the GNU General Public License as
# published by the Free Software Foundation; either version 2 of the
# License, or (at your option) any later version.
#
# This program is distributed in the hope that it will be useful, but
# WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
# General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program; if not, write to the Free Software
# Foundation, 59 Temple Place, Suite 330, Boston, MA 02111-1307, USA
########################################################################

package metaf2xml::parser;

########################################################################
# some things strictly perl
########################################################################
use strict;
use warnings;

use POSIX qw(floor);

########################################################################
# export the functions provided by this module
########################################################################
BEGIN {
    use Exporter();
    our ($VERSION, $VERSION_ID, @ISA, @EXPORT, @EXPORT_OK, %EXPORT_TAGS);

    $VERSION     = sprintf "%d.%03d", '$Revision: 1.41 $0.0' =~ /(\d+)/g;
    $VERSION_ID  = 'parser.pm: $Id: parser.pm,v 1.41 2008/04/06 06:55:30 metaf2xml Exp $';
    @ISA         = qw(Exporter);
    @EXPORT      = ();
    %EXPORT_TAGS = ();
    @EXPORT_OK   = qw(parseReport);
}
our @EXPORT_OK;

END { }

=head1 NAME

metaf2xml::parser -- parse a METAR, TAF, or SYNOP message

=head1 SYNOPSIS

  use metaf2xml::parser;
  my %report = metaf2xml::parser::parseReport($msg,$is_taf,$is_synop);

=head1 DESCRIPTION

This module contains functions to analyze a string (per default as a METAR
message). Its function L<parseReport()|/parsereport__msg__is_taf__is_synop_>
returns a hash with all its components.

=cut

########################################################################
# define lots of regular expressions
########################################################################
my $re_day = '(?:[0-2]\d|3[01])';
my $re_hour = '(?:[01]\d|2[0-3])';
my $re_min = '[0-5]\d';
my $re_bbb = '(?:(?:RR|CC|AA)[A-Z]|P[A-Z]{2})';
my $re_rwy = '(?:0[1-9]|[12]\d|3[0-6])';
my $re_rwy_des = "$re_rwy(?:[LCR])?";
my $re_rwy_des2 = '(?:[05][1-9]|[1267]\d|[38][0-6])'; # add 50 for rrR
my $re_wind_speed = 'P?[1-9]?\d\d';
my $re_wind_speed_unit = '(?:KTS?|TK|K T| KT|MPS|KMH)';
my $re_wind_dir = '(?:[0-2]\d|3[0-6])';
# EXTENSION: allow wind direction not rounded to 10 degrees
my $re_wind = "(VRB|///|$re_wind_dir\\d)(//|$re_wind_speed)(?:G($re_wind_speed))?($re_wind_speed_unit)";
my $re_vis_m = '\d{4}';
my $re_vis_km = '[1-9]\d?KM';
# EXTENSION: allow P6000 (UERP ULOO UMKK UMOO URSS UTAK)
my $re_vis_m_km = "(?:P?$re_vis_m(?:NDV)?|$re_vis_km)";
my $re_vis_m_km_remark = "(?:$re_vis_m|[1-9]\\d{2,3}M|$re_vis_km)";
my $re_frac16 = '[135]/16';
my $re_frac8 =  '[1357]/8';
my $re_frac4 =  '[13]/4';
my $re_frac2 =  '1/2';
my $re_vis_frac_sm0 = "(?:$re_frac16|$re_frac8|$re_frac4|$re_frac2)";
my $re_vis_frac_sm1 = "1(?: (?:$re_frac8|$re_frac4|$re_frac2))?";
# EXTENSION: allow /8 for 2 miles
my $re_vis_frac_sm2 = "2(?: (?:$re_frac8|$re_frac4|$re_frac2))?";
# EXTENSION: allow /2 for 3 miles
my $re_vis_frac_sm3 = "3(?: (?:$re_frac2))?";
my $re_vis_whole_sm = '(?:[04-9]|1[0-5]|[2-9][05]|[1-9]\d0)';
my $re_vis_sm =   "(?:M1/4"
                 . "|$re_vis_whole_sm"
                 . "|$re_vis_frac_sm0"
                 . "|$re_vis_frac_sm1"
                 . "|$re_vis_frac_sm2"
                 . "|$re_vis_frac_sm3)";
my $re_gs_size = "(?:M1/4|[1-9]\\d*(?: $re_frac4| $re_frac2)?|$re_frac4|$re_frac2)";
my $re_rwy_vis = '[PM]?\d{4}';
my $re_rwy_wind = "(?:WIND )?(?:RWY($re_rwy_des) ?|($re_rwy_des)/)((?:///|VRB|${re_wind_dir}0)(?://|$re_wind_speed(?:G$re_wind_speed)?)$re_wind_speed_unit)(?: ($re_wind_dir\\d)V($re_wind_dir\\d))?";
my $re_rwy_wind2 = "((?:///|VRB|${re_wind_dir}0)(?://|$re_wind_speed(?:G$re_wind_speed)?)$re_wind_speed_unit)/RWY($re_rwy_des)";
my $re_rs_rwy_wind = "(RS|$re_rwy_des)($re_wind_dir)($re_wind_speed(?:G$re_wind_speed)?$re_wind_speed_unit)";

my $re_compass_dir = '(?:[NS]?[EW]|[NS])';
my $re_compass_dir16 = "(?:[NE]NE|[ES]SE|[SW]SW|[NW]NW|$re_compass_dir)";

my $re_phen_locdir_comp = "(?:(?:GRID )?(?:[1-9]\\d*(?:KM)? ?)?$re_compass_dir16(?: QUAD)?)";
my $re_phen_locdir_comp_thru = "(?:O?MTNS|OMTS|(?:OVR|ALG) (?:LK|RIVER|MT(?:N?S)?|VLYS?)|(?:(?:OV?HD|(?:OVE?R |AT |NR )?A(?:P|R?PT|D)|IN VLY|STNRY|(?:DISTANT |DSNT |VC )?(?:OBSCG MTNS )?(?:$re_phen_locdir_comp|ALQDS|V/D))(?:(?: ?- ?| THRU )(?:$re_phen_locdir_comp|OV?HD))*))";
my $re_phen_locdir_comp_and = "(?:$re_phen_locdir_comp_thru(?:(?:[/ ]| AND )$re_phen_locdir_comp_thru)*)";
my $re_phen_locdirs = "(?: $re_phen_locdir_comp_and)";
my $re_wx_mov_d2 = "(?: (MOVD?) ($re_compass_dir16(?:-$re_compass_dir16)*|OV?HD|STNRY|UNKN))";
my $re_num_quadrant1 = '(?:1(?:ST)?|2(?:ND)?|3(?:RD)?|4(?:TH)?)';
my $re_num_quadrant = "(?: $re_num_quadrant1(?:[- ]$re_num_quadrant1)* QUAD)";
my $re_weather_desc = '(?:MI|BC|PR|DR|BL|SH|TS|FZ)';
my $re_weather_ra = 'RA(?:DZ|SN|SG|PL|IC|GR|GS)*';
my $re_weather_dz = 'DZ(?:RA|SN|SG|PL|IC|GR|GS)*';
my $re_weather_sn = 'SN(?:RA|DZ|SG|PL|IC|GR|GS)*';
my $re_weather_ic = 'IC(?:RA|DZ|SN|SG|PL|GR|GS)*';
my $re_weather_prec = '(?:DZ|RA|SN|SG|PL|IC|GR|GS)+';
# EXTENSION: allow SHRADZ (ENHK)
my $re_weather_ts_sh = '(?:RA|SN|PL|GR|GS|DZ)+|UP';
my $re_weather_sg_pl_gr_gs = '(?:SG|PL|GR|GS)+';
# EXTENSION: allow SN, BC, PR, BR, PL with FZ
my $re_weather_fz = '(?:RA|DZ|FG|SN|BC|PR|BR|PL)+';
# EXTENSION: allow +/- with BLSN, BR, TS, FC
# EXTENSION: allow JP (adjacent precipitation in old METAR)
my $re_weather_w_i = "(?:$re_weather_prec|DS|SS"
                      . "|(?:TS|SH)(?:$re_weather_ts_sh)"
                      . "|FZ(?:SH)?(?:$re_weather_fz)"
                      . "|UP|BLSN|BR|FC|TS|JP)";
my $re_weather_wo_i = "(?:$re_weather_ic|FG|BR|SA|DU|HZ|FU|VA|SQ|PO|FC|TS"
                       . "|(?:FZ|MI|BC|PR)*FG"
                       . "|BL(?:$re_weather_sn|SA|DU)"
                       . "|DR(?:$re_weather_sn|SA|DU))";
# EXTENSION: allow VCTS..
my $re_weather_vc = "(?:FG|PO|FC|DS|SS|TS(?:$re_weather_ts_sh)?|SH|BL(?:$re_weather_sn|SA|DU)|VA)";
# annex 3, form:
# FZDZ FZRA DZ   RA   SHRA SN SHSN SG SHGR SHGS BLSN SS DS
# TSRA TSSN TSPL TSGR TSGS FC VA   PL UP
# annex 3, text:
# FZDZ FZRA DZ   RA   SHRA SN SHSN SG SHGR SHGS BLSN SS DS
#                          FC VA   PL    TS
# EXTENSION: allow FZFG
my $re_weather_re = "(?:(?:FZ|SH)?$re_weather_prec"
                     . "|TS(?:$re_weather_ts_sh)"
                     . "|FZFG|BLSN|DS|SS|TS|FC|VA|UP)";
# EXTENSION: allow [+-]VC...
my $re_weather = "(?:[+-]?$re_weather_w_i|$re_weather_wo_i|[+-]?VC$re_weather_vc)";

my $re_cloud_cov = '(?:FEW|SCT|BKN|OVC)';
my $re_cloud_base = '(?:///|\d{3})';
my $re_cloud_type = '(?:AC|ACC|ACSL|AS|CB|CBMAM|CC|CCSL|CI|CS|CU|CUFRA|CF'
                     . '|NS|SAC|SC|SCSL|ST|STFRA|SF|TCU|///)';

# EXTENSION: allow // for last 2 digits
# EXTENSION: allow tenths of hPa as .\d (FLLS, HUEN, OPRN)
# EXTENSION: allow QNH\d{4}INS
my $re_qnh = '(?:(?:Q[01]|A[23])\d(?:\d\d|//)|QNH[23]\d{3}INS|[AQ]////|Q[01]\d{3}\.\d)';

my $re_colour1 = 'BLU\+?|WHT|GRN|YLO[12]?|AMB|RED';
my $re_colour = "(?:BLACK|(?:BLACK)?(?:$re_colour1)(?: ?(?:$re_colour1|FCST CANCEL))?)";

my $re_be_prec_be   = "(?:[BE]$re_hour?$re_min)";
my $re_be_prec = "(?:(?:FZ|SH)?(?:TS|RA|SN|UP|PL|DZ|GS|FC)+$re_be_prec_be+)";

my $re_rsc_cond = '(?:DRY|WET|SANDED)';
my $re_rsc_deposit = '(?:LSR|SLR|PSR|IR|WR)+(?://|\d\d)?';
my $re_rsc = "(?:$re_rsc_deposit(?:P(?:[ /]$re_rsc_cond)?)?|(?:(?:P[ /])?$re_rsc_cond))";

my $re_snw_cvr_title = '(?:SNO?W? ?(?:CVR|COV(?:ER)?)?[/ ])';
my $re_snw_cvr_state =
       '(?:(?:ONE |MU?CH |TR(?:ACE)? ?)LOOSE|(?:MED(?:IUM)?|HARD) PACK(?:ED)?)';
my $re_snw_cvr = "(?:$re_snw_cvr_title?($re_snw_cvr_state)|${re_snw_cvr_title}NIL)";

my $re_temp = '(?:-?\d+\.\d)';
my $re_precip = '(?:\d{1,2}(?:\.\d)?(?: ?[MC]M)?)';

my $re_opacity_phenom =
        '(?:(?:BL)?SN|FG|AS|AC|CC|CF|CI|CS|CU|TCU|NS|ST|SF|SC|ACC|CUFRA|CB|IC)';
my $re_trace_cloud = '(?:AC|AS|CF|CI|CS|CU|CUFRA|SC|SF|TCU)';

my $re_phen_desc_when = '(?:OCNL|FRE?Q|INTMT|CON(?:TU)?S|PAST HR)';
my $re_phen_desc_how  = '(?:LOW?|LWR|(?:VRY |V|PR )?(?:THN|THIN|THK|THICK)|ISOL|CVCTV|DSIPTD|FREEZING|PTCHY)';
my $re_phen_desc_strength  = '(?:(?:VRY |V|PR )?(?:LGT|FBL)|MDT|MOD)';
my $re_phen_desc_where = '(?:ALOFT|ARND)';
my $re_phen_desc = "(?:$re_phen_desc_when|$re_phen_desc_how|$re_phen_desc_strength|$re_phen_desc_where)";
my $re_phen_desc_other = "(?:$re_phen_desc_when|$re_phen_desc_where)";

my $re_ltg_types = '(?:CA|CC|CG|CW|IC)';

my $re_wind_shear_lvl = "WS(\\d{3})/(${re_wind_dir}0${re_wind_speed}KT)";

my $re_phenomenon_other = "(?:LTG|VIRGA|AURBO|AURORA|FG BNK|FULYR|HZY|BINOVC|ICG|SHS?|SHWR|DEW|(?:VIS|CIG|SC) (?:HYR|HIER|LWR|RDCD|REDUCED)|CLDS?(?: EMBD)?|(?:GRASS )?FIRES?)";
my $re_phenomenon4 = "(?:($re_phenomenon_other|PCPN|VLY FG|HIR CLDS)|($re_cloud_type(?:[/-]$re_cloud_type)*)|($re_weather|SMOKE|HAZE|TSTMS)|($re_cloud_cov))";

my $re_loc_quadr3 = "(?:(?: TO)?($re_phen_locdirs)|(?: (DSNT))?($re_num_quadrant))";

my $re_estmd = '(?:EST(?:MD)?|ESTM?D|ESMTD|ESTIMATED)';
my $re_data_estmd = '(?:WI?NDS?(?: DATA)?|(?:CIG )?BLN|CIG|CEILING|SLP|ALSTG|CLD HGTS)';

my $re_synop_tt = '(?:[0-6][0-9]|7[0-5])';
my $re_synop_period = '(?:[0-6][0-9])';

########################################################################
# helper functions
########################################################################

sub _rnd {
    my ($val, $prec) = @_;
    return $prec * floor(($val / $prec) + 0.5);
}

sub _makeErrorMsg {
    my ($errorType, $msg, $curr_msg) = @_;
    my $s;

    $s = substr($msg, 0, (length $msg) - length $curr_msg) . '<@> ' . $curr_msg;
    $s =~ s/(<@> +[^ ]+) .+/$1 .../;
    return { errorType => $errorType, s => $s };
}

# $cy: [ 'ABCD', 'AB', 'A' ]
sub _isValidCountry {
    my ($re, $cy) = @_;

    my %country = (
        cloudMaxCover =>
            ' LO ',
        NEFO_PLAYA =>
            ' SCSE ',
        relHumid =>
            ' OOSA ',
        rwyWind =>
            ' LG ',
        QNH_Alt =>
            ' MN MS MT MZ OI OM RP VT ',
        rsRwyWind =>
            ' LPMA ',
        remarkCloudMaxCov =>
            ' LI OA ',
        remarkColour =>
            ' BKPR EG ET FH L OA PGUA ',
        remarkCloudTypes =>
            ' K MM MU NS P RJ RO UT ',
        remarkCIGRAG =>
            ' CY P K ',
        remarkCIGVRBL =>
            ' CY ',
        remarkSLPhPa =>
         ' BGTL CM CW CY CZ ET EGVA K LQ MHSC MM MU NS NZ P RJ RK RO TI TJ TX ',
        remarkSPinHg =>
            ' ROTM ',
        remarkSLPinHg =>
            ' ',
        remarkGRIDWind =>
            ' NZ ',
        remarkRwyWind =>
            ' BG LC LH LT SP UL ',
        remarkRwyWind2 =>
            ' RC ',
        remarkRsc =>
            ' BG EGUN ET K LQTZ NZ PA PG PH RJ RK RO ',
        remarkOpacityPhenom =>
            ' CQ CW CY CZ ',
        remarkOpacityPhenomRev =>
            ' BG KTTS ',
        remarkCloudOpacityLvl =>
            ' RJ RO YB YM YP ',
        remarkCloudCoverLvl =>
            ' C EK K LI LL P NS TI TJ ',
        remarkCloudTypeLvl =>
            ' U ',
        remarkCloudTraces =>
            ' CW CY CZ ',
        remarkReWeather =>
            ' CY FQ TR ',
        remarkSfcVisM =>
            ' KQ ',
        remarkVisSfcTwr =>
            ' K PA ',
        remarkVisVarK =>
            ' K PA ',
        remarkVisVarCY =>
            ' CY ',
        remarkVisVarCYCZ =>
            ' CY CZ ',
        remarkPCPNPastHour =>
            ' CW CY CZ ',
        remarkVisLocM =>
            ' E KQ LQ NZ RJ RO ',
        remarkQNH =>
            ' RJ RO RP ',
        remarkQNHMB =>
            ' KQ ',
        remarkQFF =>
            ' MM ',
        remarkCorrected =>
            ' ET K LQ PA ',
        remarkSnowIncr =>
            ' K PA ',
        remarkHourly =>
            ' BG C ET K LQ MM MU NS P RJ RKSO RO TI TJ ',
        remarkPeakWind =>
            ' C ET K LP LQ NZ P TI TJ ',
        remarkHourlyTemp =>
            ' C K P NS MU RJ RO TI TJ ',
        remarkSunshine =>
            ' K P ',
        remarkQFEhPa =>
            ' UTTP UTSS ',
        remarkDA_PA =>
            ' K CY ',
        remarkKN =>
            ' KN ',
        remarkKNIP =>
            ' KNIP ',
        remarkQFEmmHg =>
            ' U ZM ',
        remarkPhenCloud =>
            ' K PA PH ',
        remarkRainfall =>
            ' YA YB ',
        remarkRSNK =>
            ' KHMS ',
        remarkLAGPK =>
            ' KNTD ',
        remarkPrecipHourly3 =>
            ' KAEG KAUN KBVS KUKT KW22 ',
        remarkCloudCoverVar =>
            ' K MH PA PH PM PP ',
        remarkC_K_P_NS_RJ_RO_TI_TJ_ET =>
            ' C K P NS RJ RO TI TJ ET ',
        remarkSNWCVR =>
            ' C ',
        remarkObsTimeOffset =>
            ' C ',
        remarkBalloon =>
            ' CY CZ ',
        remarkRVR =>
            ' CY ',
        remarkTX =>
            ' FQ ',
        remarkTAFNxtFcstBy =>
            ' CB CY CZ ',
        remarkTAFNxtFcstAt =>
            ' CW CY ',
        remarkTAFAmdAt =>
            ' LT ',
        remarkTAFFcstAutoObs =>
            ' CB CW CY CZ ',
        remarkTAFFcstAutometar =>
            ' EF ',
        remarkATIS =>
            ' ED ',
    );
    return    index($country{$re}, ' ' . $cy->[0] . ' ') > -1
           || index($country{$re}, ' ' . $cy->[1] . ' ') > -1
           || index($country{$re}, ' ' . $cy->[2] . ' ') > -1;
}

sub _parseFraction {
    my ($r, $vis, $unitDist) = @_;

    $r->{distance}{u} = $unitDist;
    if ($vis =~ /^M/) {
        $r->{distance}{v} = 0.25;
        $r->{distance}{q} = 'isLess';
    } elsif ($vis =~ /^[\d.]+$/o) {
        $r->{distance}{v} = $vis;
    } elsif ($vis =~ m@^([1357])/([1-8]+)$@o) {
        $r->{distance}{v} = $1 / $2;
    } else {
        $vis =~ m@(.)(?: (\d*)/(\d*))?@;
        $r->{distance}{v} = $1;
        $r->{distance}{v} += $2 / $3 if $2;
    }
}

sub _parseWeather {
    my ($weather, $is_recent) = @_;
    my ($w, $int, $vc, $desc, $phen, $weather_str);

    $weather_str = ($is_recent ? 'RE' : '') . $weather;
    return { notAvailable => undef, s => $weather_str } if $weather eq '//';
    return { NSW => undef, s => $weather_str } if $weather eq 'NSW';

    return { tornado => undef, s => $weather_str } if $weather eq '+FC';

    # for $re_phenomenon4, only
    $weather = 'FU' if $weather eq 'SMOKE';
    $weather = 'HZ' if $weather eq 'HAZE';
    $weather = 'TS' if $weather eq 'TSTMS';

    ($int, $vc, $desc, $phen) =
                      $weather =~ m@([+-])?(VC)?($re_weather_desc+)?([A-Z/]+)@o;
    $w->{s}             = $weather_str;
    $w->{phenomDescr}   = ($int eq '-' ? 'isLight' : 'isHeavy') if defined $int;
    $w->{inVicinity}    = undef if defined $vc;
    @{$w->{descriptor}} = $desc =~ /(..)+?/g if defined $desc;
    @{$w->{phenomenon}} = $phen =~ /(..)+?/g if defined $phen;
    return $w;
}

sub _parseOpacityPhenom {
    my ($r, $clds) = @_;

    for ($clds =~ /((?:[A-Z]+\d)+?)/g) {
        /(.*)(.)/;
        if (index($1, 'SN') >-1 || index($1, 'FG') > -1 || index($1, 'IC') > -1)
        {
            push @{$r->{opacityPhenomArr}}, { opacityWeather => {
                oktas   => $2,
                weather => _parseWeather $1
            }};
        } else {
            push @{$r->{opacityPhenomArr}}, { opacityCloud => {
                oktas     => $2,
                cloudType => $1
            }};
        }
    }
}

sub _parseRwyVis {
    my ($metar, $msg) = @_;
    my $r;

    if ($$msg =~ m@^(R($re_rwy_des)/(?:(////)|($re_rwy_vis)(?:V($re_rwy_vis))?(FT)?)(?:/)?([UDN])?) (.*)@o)
    {
        my $v;

        $v->{s}        = $1;
        $v->{rwyDesig} = $2;
        $v->{visTrend} = $7 if defined $7;
        if ($3) {
            $v->{RVR}{notAvailable} = undef;
        } else {
            $v->{RVR}{distance} = { v => $4, u => (defined $6 ? $6 : 'M') };
            if (defined $5) {
                $v->{RVRVariations}{distance} =
                                       { v => $5, u => $v->{RVR}{distance}{u} };
                $v->{RVRVariations}{distance}{q} = 'isLess'
                    if $v->{RVRVariations}{distance}{v} =~ s/^M//;
                $v->{RVRVariations}{distance}{q} = 'isEqualGreater'
                    if $v->{RVRVariations}{distance}{v} =~ s/^P//;
                $v->{RVRVariations}{distance}{v} += 0;
            }
            # corrections postponed because pattern matching changes $5
            $v->{RVR}{distance}{q} = 'isLess'
                if $v->{RVR}{distance}{v} =~ s/^M//;
            $v->{RVR}{distance}{q} = 'isEqualGreater'
                if $v->{RVR}{distance}{v} =~ s/^P//;
            $v->{RVR}{distance}{v} += 0;
        }
        push @{$metar->{visRwy}}, $v;
        $$msg = $8;
        return 1;
    }
    return 0;
}

sub _parseRwyState {
    my ($metar, $msg) = @_;
    my $r;

    if ($$msg =~ m@^((SNOCLO)|(88|99|$re_rwy_des2)(?:(CLRD)|([\d/])([\d/])(\d\d|//))(\d\d|//)) (.*)@o)
    {
        $$msg = $9;
        $r->{s} = $1;
        if (defined $2) {
            $r->{SNOCLO} = undef;
        } else {
            if ($3 == 88) {
                $r->{rwyDesigAll} = undef;
            } elsif ($3 == 99) {
                $r->{rwyDesigRep} = undef;
            } elsif ($3 > 50) {
                $r->{rwyDesig} = sprintf '%02dR', $3 - 50;
            } else {
                $r->{rwyDesig} = $3;
            }
            if (defined $4) {
                $r->{cleared} = undef;
            } else {
                if ($5 eq '/') {
                    $r->{depositType}{notAvailable} = undef;
                } else {
                    $r->{depositType}{depositTypeVal} = $5;
                }
                if ($6 eq '/') {
                    $r->{depositExtent}{notAvailable} = undef;
                } elsif (   $6 eq '0' || $6 eq '1' || $6 eq '2'
                         || $6 eq '5' || $6 eq '9')
                {
                    $r->{depositExtent}{depositExtentVal} = $6;
                } else {
                    $r->{depositExtent}{invalidFormat} = $6;
                }
                if ($7 eq '//') {
                    $r->{depositDepth}{notAvailable} = undef;
                } elsif ($7 != 91) {
                    $r->{depositDepth}{depositDepthVal} = $7 + 0;
                } else {
                    $r->{depositDepth}{invalidFormat} = $7;
                }
            }
            if ($8 eq '//') {
                $r->{friction}{notAvailable} = undef;
            } elsif ($8 >= 1 && $8 <= 90) {
                $r->{friction}{coefficient} = $8 + 0;
            } elsif ($8 == 99) {
                $r->{friction}{unreliable} = undef;
            } elsif ($8 >= 91 && $8 <= 95) {
                $r->{friction}{brakingAction} = $8;
            } else {
                $r->{friction}{invalidFormat} = $8;
            }
        }
        push @{$metar->{rwyState}}, $r;
        return 1;
    }
    return 0;
}

sub _parseWind {
    my $wind = shift;
    my ($w, $dir, $speed, $gustSpeed, $unitSpeed);

    return { notAvailable => undef } if $wind eq '/////';
    ($dir, $speed, $gustSpeed, $unitSpeed) = $wind =~ m@^$re_wind$@o;
    if ($dir eq '///' && $speed eq '//') {
        $w->{notAvailable} = undef;
    } elsif ($dir eq '000' && $speed eq '00' && !defined $gustSpeed) {
        $w->{isCalm} = undef;
    } else {
        my $isGreater;

        if ($dir eq '///') {
            $w->{dirNotAvailable} = undef;
        } elsif ($dir eq 'VRB') {
            $w->{dirVariable} = undef;
        } else {
            $w->{dir} = $dir + 0; # true, not magnetic
        }
        $unitSpeed =~ s/TK/KT/;
        $unitSpeed =~ s/KTS/KT/;
        $unitSpeed =~ s/ //;
        if ($speed eq '//') {
            $w->{speedNotAvailable} = undef;
        } else {
            $isGreater = $speed =~ s/^P//;
            $w->{speed} = { v => $speed + 0, u => $unitSpeed };
            $w->{speed}{q} = 'isGreater' if $isGreater;
        }
        if (defined $gustSpeed) {
            $isGreater = $gustSpeed =~ s/^P//;
            $w->{gustSpeed} = { v => $gustSpeed + 0, u => $unitSpeed };
            $w->{gustSpeed}{q} = 'isGreater' if $isGreater;
        }
    }
    return $w;
}

sub _parseWindAtLoc {
    my ($s, $location, $wind, $windVarLeft, $windVarRight) = @_;
    my $r;

    $r = _parseWind $wind;
    $r->{windVarLeft} = $windVarLeft + 0 if defined $windVarLeft;
    $r->{windVarRight} = $windVarRight + 0 if defined $windVarRight;
    return { windAtLoc => {
        s            => $s,
        windLocation => $location,
        wind         => $r
    }};
}

sub _parseCloud {
    my $cloud = shift;
    my $c;

    $c->{s} = $cloud;
    if ($cloud eq '//////') {
        $c->{notAvailable} = undef;
    } elsif ($cloud =~ m@^$re_cloud_cov$re_cloud_base(?: ?$re_cloud_type)?@o){
        $cloud =~ '(...)(?:(...) ?(.+)?)?';
        $c->{cloudCover} = $1;
        if (defined $2) {
            if ($2 eq '///') {
                $c->{baseBelowStation} = undef;
            } else {
                $c->{cloudBase} = { v => $2 * 100, u => 'FT' }; # AGL
            }
        }
        if (defined $3) {
            if ($3 eq '///') {
                $c->{cloudTypeNotAvailable} = undef;
            } else {
                $c->{cloudType} = $3;
            }
        }
    } else {
        $cloud =~ '(...)(.+)?';
        $c->{cloudCover} = $1;
        $c->{cloudType}  = $2 if defined $2;
    }
    return $c;
}

sub _parseQNH {
    my $qnh = shift;
    my ($q, $descr, $dig12, $dig34);

    $q->{s} = $qnh;
    ($descr, $dig12, $dig34) = $qnh =~ '([AQNH]*)(..)([^INS]*)';
    if ("$dig12$dig34" eq '////') {
        $q->{notAvailable} = undef;
    } else {
        $dig34 = '00' if $dig34 eq '//';
        if ($descr eq 'Q') {
            $q->{hPa} = "$dig12$dig34" + 0;
        } else {
            $q->{inHg} = sprintf "%.2f", "$dig12$dig34" / 100;
        }
    }
    return $q;
}

sub _parseColourCode {
    my $colour = shift;
    my $c;

    $colour =~ "^(BLACK)?($re_colour1)? ?(.+)?";
    $c->{s} = $colour;
    $c->{BLACK} = undef if defined $1;
    $c->{currentColour} =
            $2 eq 'BLU+' ? 'BLUplus' : ($2 eq 'FCST CANCEL' ? 'FCSTCANCEL' : $2)
        if defined $2;
    $c->{predictedColour} =
            $3 eq 'BLU+' ? 'BLUplus' : ($3 eq 'FCST CANCEL' ? 'FCSTCANCEL' : $3)
        if defined $3;
    return $c;
}

sub _parseLocations {
    my $loc_str = shift;
    my @loc_thru;

    return { isStationary => undef } if $loc_str =~ /^ *STNRY *$/;

    for ($loc_str =~ m@(?:[/ ]| AND )?($re_phen_locdir_comp_thru|UNKN)@og) {
        my (@loc, $is_grid);
        for ($_ =~ m@(?: ?[/-] ?| THRU )?(OV?HD|(?:OVE?R |AT |NR )?A(?:P|R?PT|D)|IN VLY|O?MTNS|OMTS|(?:OVE?R|ALG) (?:LK|RIVER|MT(?:N?S)?|VLYS?)|STNRY|UNKN|(?:DISTANT |DSNT |VC )?(?:OBSCG MTNS )?(?:ALQDS|V/D|$re_phen_locdir_comp))@og)
        {
            my $l;

            if ($_ =~ m@^(?:OV?HD|(?:OVE?R |AT |NR )?A(?:P|R?PT|D)|IN VLY|O?MTNS|OMTS|(?:OVR|ALG) (?:LK|RIVER|MT(?:N?S)?|VLYS?)|STNRY|UNKN)$@) {
                s/^OVR MT(?:N?S)?/OMTNS/;
                s/O?MTNS/OMTNS/;
                s/OMTS/OMTNS/;
                s/OV?HD/OHD/;
                s/OVER/OVR/;
                s/A(?:R?PT|D)$/AP/;
                s/^AP$/AT AP/;
                s/ /_/;
                $l->{locationSpec} = $_;
            } else {
                $_ =~ m@(?:(DISTANT|DSNT|VC) )?(OBSCG MTNS )?(?:TO ?)?(?:(?:(GRID )?(?:([1-9]\d*)(KM)? ?)?($re_compass_dir16)(?: (QUAD))?)|(ALQDS|V/D))@o;
                $l->{inDistance} = undef
                    if defined $1 and ($1 eq 'DISTANT' or $1 eq 'DSNT');
                $l->{inVicinity} = undef if defined $1 and $1 eq 'VC';
                $l->{obscgMtns}  = undef if defined $2;
                if (defined $3 || $is_grid) {
                    $is_grid = 1;
                    $l->{isGrid} = undef;
                }
                $l->{distance} = { v => $4, u => (defined $5 ? 'KM' : 'SM') }
                    if defined $4;
                $l->{compassDir} = $6 if defined $6;
                $l->{isQuadrant} = undef if defined $7;
                ($l->{locationSpec} = $8) =~ tr /\/ /_/ if defined $8;
            }
            push @loc, $l;
        }
        push @loc_thru, { location => \@loc };
    }
    return { locationThru => \@loc_thru };
}

sub _determineCeiling {
    my $cloud = shift;
    my ($ceil, $idx, $ii);

    $ceil = 100000;
    $idx = -1;
    $ii = -1;
    for (@$cloud) {
        $ii++;
        if (   exists $_->{cloudBase}
            && $_->{cloudBase}{u} eq 'FT'
            && $_->{cloudBase}{v} < $ceil
            && $_->{cloudBase}{v} < 200 * 100 # TODO: actually MSL!
            && ($_->{cloudCover} eq 'BKN' || $_->{cloudCover} eq 'OVC'))
        {
            $ceil = $_->{cloudBase}{v};
            $idx = $ii;
        }
    }
    $cloud->[$idx]{isCeiling} = undef if $idx > -1;
}

sub _parseQuadrants {
    my ($q, $inDistance) = @_;

    return { locationThru => { location => {
        $inDistance ? (inDistance => undef) : (),
        quadrant => [ $q =~ /([1-4])/g ]
    }}};
}

sub _parseTemp {
    my $temp = shift;

    $temp =~ /(.)(.*)/;
    return { v => (sprintf "%.1f", ($1 eq '1' ? -1 : 1) * $2 / 10), u => 'C' };
}

sub _parsePhenomDescr {
    my ($r, $tag, $phen_descr) = @_;

    for ($phen_descr =~ /$re_phen_desc|BBLO/g) {
        s/CONTUS/CONS/;
        s/V(LGT|THN|THIN|FBL|THK|THICK)/VRY $1/;
        s/THIN/THN/;
        s/FREQ/FRQ/;
        s/THICK/THK/;
        s/^LOW?/LOW/;
        s/^MOD/MDT/;

        push @{$r->{$tag}}, {
            FRQ      => 'isFrequent',
            OCNL     => 'isOccasional',
            INTMT    => 'isIntermittent',
            CONS     => 'isContinuous',
            THK      => 'isThick',
           'PR THK'  => 'isPrettyThick',
           'VRY THK' => 'isVeryThick',
            THN      => 'isThin',
           'PR THN'  => 'isPrettyThin',
           'VRY THN' => 'isVeryThin',
            LGT      => 'isLight',
           'PR LGT'  => 'isPrettyLight',
           'VRY LGT' => 'isVeryLight',
            FBL      => 'isFeeble',
           'PR FBL'  => 'isPrettyFeeble',
           'VRY FBL' => 'isVeryFeeble',
            MDT      => 'isModerate',
            LOW      => 'isLow',
            LWR      => 'isLower',
            ISOL     => 'isIsolated',
            CVCTV    => 'isConvective',
            DSIPTD   => 'isDissipated',
           'PAST HR' => 'inPastHour',
            BBLO     => 'baseBelowStation',
            ALOFT    => 'isAloft',
            ARND     => 'isAround',
            FREEZING => 'isFreezing',
            PTCHY    => 'isPatchy',
        }->{$_};
    }
}

# "supplementary" section of TAFs
sub _parseTAFsuppl {
    my ($metar, $msg, $base_metar) = @_;
    my $r;

    if ($$msg =~ m@^(([56])(\d)(\d{3})(\d)) (.*)@o) {
        my ($ti, $type);

        $$msg = $6;
        $type = $2 eq '5' ? 'turbulence' : 'icing';
        $r = {
            s               => $1,
            $type . 'Descr' => $3,
            layerBase       => $4 + 0,
        };
        $ti = $5;

        # if item repeated: level is layer top
        if ($$msg =~ m@^($2$3(\d{3})$5) (.*)@) {
            $$msg = $3;
            $r->{s} .= ' ' . $1;
            $r->{layerTop} = $2 + 0;
        } else {
            $r->{layerThickness} = $ti;
        }
        push @{$metar->{TAFsupplArr}}, { $type => $r };
        return 1;
    }

    if ($$msg =~ m@^($re_wind_shear_lvl) (.*)@o) {
        $$msg = $4;
        push @{$metar->{TAFsupplArr}}, { windShearLvl => {
            s     => $1,
            level => $2 + 0,
            wind  => _parseWind $3
        }};
        return 1;
    }

    if ($$msg =~ m@^($re_qnh) (.*)@o) {
        $$msg = $2;
        push @{$metar->{TAFsupplArr}}, { QNH => _parseQNH $1 };
        return 1;
    }

    return 0;
}

# additional TAF info
sub _parseTAFinfo {
    my ($metar, $msg) = @_;
    my $r;

    if ($$msg =~ m@^(T([XN])(M)?(\d\d)/($re_hour|24)Z) (.*)@o) {
        $$msg = $6;
        $r->{s} = $1;
        $r->{temp} = { v => $4 + 0, u => 'C' };
        $r->{temp}{v} *= -1 if defined $3;
        $r->{hour} = $5;
        push @{$metar->{TAFinfoArr}},
                              { ($2 eq 'N' ? 'tempMinAt' : 'tempMaxAt') => $r };
        return 1;
    }

    if ($$msg =~ m@^(T(M)?(\d\d)/($re_hour)Z) (.*)@o) {
        $$msg = $5;
        $r->{s} = $1;
        $r->{temp} = { v => $3 + 0, u => 'C' };
        $r->{temp}{v} *= -1 if defined $2;
        $r->{hour} = $4;
        push @{$metar->{TAFinfoArr}}, { tempAt => $r };
        return 1;
    }

    if ($$msg =~ m@^((COR|AMD) ($re_hour)($re_min)Z?) (.*)@o) {
        $$msg = $5;
        push @{$metar->{TAFinfoArr}},
            { $2 eq 'COR' ? 'correctedAt' : 'amendedAt' => {
                s      => $1,
                hour   => $3,
                minute => $4
        }};
        return 1;
    }

    if ($$msg =~ m@^(LIMITED METWATCH ($re_day)($re_hour) TIL ($re_day)($re_hour)) (.*)@o)
    {
        $$msg = $6;
        push @{$metar->{TAFinfoArr}}, { limMetwatch => {
            s         => $1,
            timeFrom => { day => $2, hour => $3 },
            timeTill => { day => $4, hour => $5 },
        }};
        return 1;
    }

    if ($$msg =~ m@^(AUTOMATED SENSOR(?:ED)? METWATCH(?: ($re_day)($re_hour) TIL ($re_day)($re_hour))?) (.*)@o)
    {
        $$msg = $6;
        $r->{s} = $1;
        if (defined $2) {
            $r->{timeFrom} = { day => $2, hour => $3 };
            $r->{timeTill} = { day => $4, hour => $5 };
        }
        push @{$metar->{TAFinfoArr}}, { autoMetwatch => $r };
        return 1;
    }

    if ($$msg =~ m@^(AMD (?:LTD TO CLD VIS AND WIND(?: (?:TIL ($re_hour|24)Z|(${re_hour})Z-($re_hour|24)Z))?|(NOT SKED))) (.*)@o)
    {
        $$msg = $6;
        $r->{s} = $1;
        if (defined $5) {
            $r->{isNotScheduled} = undef;
        } else {
            $r->{isLtdToCldVisWind} = undef;
            $r->{hourTill} = $2 if defined $2;
            $r->{hourFrom} = $3 if defined $3;
            $r->{hourTill} = $4 if defined $4;
        }
        push @{$metar->{TAFinfoArr}}, { amendment => $r };
        return 1;
    }

    return 0;
}

sub _getHumidity {
    my ($t, $d) = @_;

    return (
        # FGFS metar
        _rnd(100 * 10 ** (7.5 * ($d / ($d + 237.7) - $t / ($t + 237.7))), 0.01),

        # http://www.bragg.army.mil/www-wx/wxcalc.htm
        # http://www.srh.noaa.gov/bmx/tables/rh.html
        _rnd(100 * ((112 - (0.1 * $t) + $d) / (112 + (0.9 * $t))) ** 8, 0.01),

        # http://www.mattsscripts.co.uk/mweather.htm
        # http://ingrid.ldeo.columbia.edu/dochelp/QA/Basic/dewpoint.html
        _rnd(
          100 * (6.11 * exp(5417.118093 * (1 / 273.15 - (1 / ($d + 273.15)))))
              / (6.11 * exp(5417.118093 * (1 / 273.15 - (1 / ($t + 273.15))))),
          0.01),

        # http://de.wikipedia.org/wiki/Taupunkt
        _rnd(100 * (6.11213 * exp(17.5043 * $d / (241.2 + $d)))
                 / (6.11213 * exp(17.5043 * $t / (241.2 + $t))), 0.01)
        );
}

sub _IIiii2region {
    my $IIiii = shift;

    my %IIiii_region = (
         I        => [ [ 60001, 69998 ] ],
         II       => [ [ 20001, 20099 ], [ 20200, 21998 ], [ 23001, 25998 ],
                       [ 28001, 32998 ], [ 35001, 36998 ], [ 38001, 39998 ],
                       [ 40350, 48599 ], [ 48800, 49998 ], [ 50001, 59998 ] ],
         III      => [ [ 80001, 88998 ] ],
         IV       => [ [ 70001, 79998 ] ],
         V        => [ [ 48600, 48799 ], [ 90001, 98998 ] ],
         VI       => [ [ 00001, 19998 ], [ 20100, 20199 ], [ 22001, 22998 ],
                       [ 26001, 27998 ], [ 33001, 34998 ], [ 37001, 37998 ],
                       [ 40001, 40349 ] ],
        Antarctic => [ [ 89001, 89998 ] ],
    );
    for my $region (keys %IIiii_region) {
        for (@{$IIiii_region{$region}}) {
            return $region if $IIiii >= $_->[0] && $IIiii <= $_->[1];
        }
    }
    return '';
}

# C  genus of cloud
# C  genus of cloud predominating in the layer
# C´ genus of cloud whose base is below the level of the station
sub _codeTable0500 {
    my ($r, $idx) = @_;

    if ($idx eq '/') {
        $r->{cloudTypeNotAvailable} = undef;
    } else {
        $r->{cloudType} = ( 'CI', 'CC', 'CS', 'AC', 'AS',
                            'NS', 'SC', 'ST', 'CU', 'CB' )[$idx];
    }
}

# D  true direction from which surface wind is blowing
# D  true direction towards which ice has drifted in the past 12 hours
# DH true direction from which CH clouds are moving
# DK true direction from which swell is moving
# DL true direction from which CL clouds are moving
# DM true direction from which CM clouds are moving
# Da true direction in which orographic clouds or clouds with vertical development are seen
# Da true direction in which the phenomenon indicated is observed or in which conditions specified in the same group are reported
# De true direction towards which an echo pattern is moving
# Dp true direction from which the phenomenon indicated is coming
# Ds true direction of resultant displacement of the ship during the three hours preceding the time of observation
# D1 true direction of the point position from the station
sub _codeTable0700 {
    my ($r, $type, $idx) = @_;

    if ($idx eq '/') {
        $r->{"cloud${type}NA"} = undef;
    } elsif ($idx == 0) {
        $r->{"cloud${type}None"} = undef;
    } elsif ($idx == 9) {
        $r->{"cloud${type}Invisible"} = undef;
    } else {
        $r->{"cloud${type}Dir"} = {
            1 => 'NE',
            2 => 'E',
            3 => 'SE',
            4 => 'S',
            5 => 'SW',
            6 => 'W',
            7 => 'NW',
            8 => 'N',
        }->{$idx};
    }
}

# eC elevation angle of the top of the cloud indicated by C
# e´ elevation angle of the top of the phenomenon above horizon
sub _codeTable1004 {
    my ($r, $idx) = @_;

    if ($idx == 0) {
        $r->{topsInvisible} = undef;
    } elsif ($idx == 1) {
        $r->{elevationAngle}{v} = 45;
        $r->{elevationAngle}{q} = 'isEqualGreater';
    } elsif ($idx == 9) {
        $r->{elevationAngle}{v} = 5;
        $r->{elevationAngle}{q} = 'isLess';
    } else {
        $r->{elevationAngle} = {
            2 => 30,
            3 => 20,
            4 => 15,
            5 => 12,
            6 => 9,
            7 => 7,
            8 => 6,
        }->{$idx};
    }
}

# h height above surface of the base of the lowest cloud seen
sub _codeTable1600 {
    my $idx = shift;
    my @v;

    return { notAvailable => undef } if $idx eq '/';
    return { from => { v => 2500, u => 'M', q => 'isEqualGreater' }}
        if $idx == 9;
    @v = (0, 50, 100, 200, 300, 600, 1000, 1500, 2000, 2500)[$idx, $idx + 1];
    return { from => { v => $v[0], u => 'M' },
             to   => { v => $v[1], u => 'M' }};
}

# hshs height of base of cloud layer or mass whose genus is indicated by C
# htht height of the tops of the lowest clouds or height of the lowest cloud layer or fog
sub _codeTable1677 {
    my $idx = shift;

    return { v => 30,    q => 'isLess'         } if $idx == 0;
    return { v => 30 * $idx                    } if $idx <= 50;
    return { v => 300 * ($idx - 50)            } if $idx <= 80;
    return { v => 1500 * ($idx - 74)           } if $idx <= 88;
    return { v => 21000, q => 'isGreater'      } if $idx == 89;
    return { v => 50,    q => 'isLess'         } if $idx == 90;
    return { v => 2500,  q => 'isEqualGreater' } if $idx == 99;
    return [(50, 100, 200, 300, 600, 1000, 1500, 2000, 2500)
                                                        [$idx - 91, $idx - 90]];
}

# N  total cloud cover
# Nh amount of all the CL cloud present or, if no CL cloud is present, the amount of all the CM cloud present
# Ns amount of individual cloud layer or mass whose genus is indicated by C
# N′ amount of cloud whose base is below the level of the station
sub _codeTable2700 {
    my $idx = shift;

    return ( s=> $idx, notAvailable => undef ) if $idx eq '/';
    return ( s=> $idx, skyObscured  => undef ) if $idx eq '9';
    return ( s=> $idx, oktas        => $idx  );
}

# ss depth of newly fallen snow
sub _codeTable3870 {
    my ($r, $idx) = @_;

    if ($idx <= 55) {
        $$r->{snowDepthVal} = { v => $idx * 10, u => 'MM' };
    } elsif ($idx <= 90) {
        $$r->{snowDepthVal} = { v => ($idx - 50) * 100, u => 'MM' };
    } elsif ($idx <= 96) {
        $$r->{snowDepthVal} = { v => $idx + 0, u => 'MM' };
    } elsif ($idx == 97) {
        $$r->{snowDepthVal} = { v => 1, q => 'isLess', u => 'MM' };
    } elsif ($idx == 98) {
        $$r->{snowDepthVal} = { v => 4000, q => 'isGreater', u => 'MM' };
    } else {
        $$r->{snowDepthNoMeasurement} = undef;
    }
}

# tR duration of period of reference for amount of precipitation, ending at the time of the report
sub _codeTable4019 {
    my $idx = shift;
    return ('6h', '12h', '18h', '24h', '1h', '2h', '3h', '9h', '15h')[$idx - 1];
}

# tt time before observation or duration of phenomena (00-69)
# zz variation, location or intensity of phenomena (76-99)
# for both: 70-75
sub _codeTable4077 {
    my ($idx, $when) = @_;

    return {} unless defined $idx;
    return { $when => { hoursFrom => $idx / 10 }}
        if $idx <= 60;
    return { $when => { hoursFrom => $idx - 55, hoursTill => $idx - 54}}
        if $idx <= 66;
    return { $when => { hoursFrom => 12, hoursTill => 18 }}
        if $idx == 67;
    return { $when => { hoursFrom => { v => 18, q => 'isGreater' }}}
        if $idx == 68;
    return { $when => { hoursUnknown => undef }}
        if $idx == 69;
    return { phenomVariation =>
             ('startDuringObs',    'endDuringObs',
              'startEndDuringObs', 'changedDuringObs',
              'startAfterObs',     'endAfterObs' )[$idx - 70]
    } if $idx <= 75;
    return { phenomDescr =>
        ('atStation',                'atStationNotInDistance',
         'allDirections',            'allDirectionsNotAtStation',
         'approchingStation',        'recedingFromStation',
         'passingStationInDistance', 'seenInDistance',
         'inVicinityNotAtStation',   'aloftNotNearGround',
         'nearGroundNotAloft',       'isOccasional',
         'isIntermittent',           'isFrequent',
         'isSteady',                 'isIncreasing',
         'isDecreasing',             'isVariable',
         'isContinuous',             'isVeryLight',
         'isLight',                  'isModerate',
         'isHeavy',                  'isVeryHeavy')[$idx - 76]
    };
}

# VV   horizontal visibility at surface
# VsVs visibility towards the sea
sub _codeTable4377 {
    my ($r, $idx) = @_;
    my (%vis, $vis_type);

    if ($idx eq '//') {
        $r->{visPrev}{s} = $idx;
        $r->{visPrev}{notAvailable} = undef;
        return;
    }

    $vis_type = 'visPrev';
    $vis{distance}{u} = 'KM';
    if ($idx == 0) {
        $vis{distance}{v} = 0.1;
        $vis{distance}{q} = 'isLess';
    } elsif ($idx <= 50) {
        $vis{distance}{v} = $idx / 10;
    } elsif ($idx <= 80) {
        $vis{distance}{v} = $idx - 50;
    } elsif ($idx <= 88) {
        $vis{distance}{v} = 5 * ($idx - 74);
    } elsif ($idx == 89) {
        $vis{distance}{v} = 70;
        $vis{distance}{q} = 'isGreater';
    } elsif ($idx == 90) {
        $vis_type = 'visibilityAtLoc';
        $vis{distance}{v} = 0.05;
        $vis{distance}{q} = 'isLess';
    } elsif ($idx == 99) {
        $vis_type = 'visibilityAtLoc';
        $vis{distance}{v} = 50;
        $vis{distance}{q} = 'isEqualGreater';
    } else {
        $vis_type = 'visibilityAtLoc';
        $vis{distance}{v} = {
            91 => 0.05, 92 => 0.2, 93 => 0.5, 94 => 1,
            95 => 2,    96 => 4,   97 => 10,  98 => 20 }->{$idx};
    }
    if ($vis_type eq 'visPrev') {
        $r->{visPrev} = \%vis;
        $r->{visPrev}{s} = $idx;
    } else {
        $r->{visibilityAtLoc}{visibility} = \%vis;
        $r->{visibilityAtLoc}{locationAt} = 'MAR';
        $r->{visibilityAtLoc}{s} = $idx;
    }
}

sub _radiationType {
    my $idx = shift;

    return {
        0 => 'rad0PosNet',
        1 => 'rad1NegNet',
        2 => 'rad2GlobalSolar',
        3 => 'rad3DiffusedSolar',
        4 => 'rad4DownwardLongWave',
        5 => 'rad5UpwardLongWave',
        6 => 'rad6ShortWave',
    }->{$idx};
}

# FM18:
# FM18 (buoy):        ZZYY Abnnn YYMMJ    GGggi       QLLLLL LLLLLL (6QQQ/)
# Ab: maritime zone, nnn: 001..499, 500 added if drifting buoy

########################################################################
# _parseSynop
########################################################################
sub _parseSynop {
    my ($msg, $msg_hdr) = @_;
    my (%report, $corr_msg, $period, $windUnit, $windEST, $region);

    my $re_D____D = '[A-Z\d]{3,}';
    my $re_A1bw   = '(?:1[1-7]|2[1-6]|3[1-4]|4[1-8]|5[1-6]|6[1-6]|7[1-4])';
    my $re_MMM    = '(?:\d{3})'; # 001..623, 901..936
    my $re_ULaULo = '(?:\d\d)';
    my $re_IIiii  = '(?:\d{5})';

    $report{msg} = $msg_hdr . $msg;
    $report{isSynop} = undef;

=pod

=head2 Parsing of SYNOP messages

=cut

    $msg .= ' '; # this makes parsing much easier

    $corr_msg = $msg;

=for comment ###########################################################

=head3 Section 0: information about the observation and the observing station

=for comment ###########################################################

    MiMiMjMj
    FM12 (fixed land):  AAXX                     YYGGiw IIiii
    FM13 (sea):         BBXX (D....D|A1bwnbnbnb) YYGGiw 99LaLaLa QcLoLoLoLo
    FM14 (mobile land): OOXX D....D              YYGGiw 99LaLaLa QcLoLoLoLo MMMULaULo h0h0h0h0im

=over

=begin html

=item

group M<sub>i</sub>M<sub>i</sub>M<sub>j</sub>M<sub>j</sub> <!--

=end html

=item group MiMiMjMj

=for html --><dd>

station type

=for html </dd>

=cut

    if ($msg !~ /^((?:AA|BB|OO)XX) (.*)/) {
        $report{ERROR} = _makeErrorMsg 'obsStationType', $corr_msg, $msg;
        return %report;
    }
    $report{obsStationType} = { s => $1, stationType => $1 };
    $msg = $2;

=begin html

=item

group D....D or A<sub>1</sub>b<sub>w</sub>n<sub>b</sub>n<sub>b</sub>n<sub>b</sub> <!--

=end html

=item group D....D or A1bwnbnbnb

=for html --><dd>

for sea stations: call sign or station id

for mobile land stations: call sign

=for html </dd>

=cut

    # group (D....D|A1bwnbnbnb)
    if ($report{obsStationType}{s} eq 'BBXX') {
        if ($msg !~ m@^(?:($re_D____D)|(${re_A1bw}\d{3})) (.*)@) {
            $report{ERROR} = _makeErrorMsg 'stationId', $corr_msg, $msg;
            return %report;
        }
        $msg = $3;
        $report{callSign} = { s => $1, id => $1 } if defined $1;
        $report{stationId} = { s => $2, id => $2 } if defined $2;
    }
    if ($report{obsStationType}{s} eq 'OOXX') {
        if ($msg !~ m@^($re_D____D) (.*)@) {
            $report{ERROR} = _makeErrorMsg 'callSign', $corr_msg, $msg;
            return %report;
        }
        $msg = $2;
        $report{callSign} = { s => $1, id => $1 };
    }

=begin html

=item

group YYGGi<sub>w</sub> <!--

=end html

=item group YYGGiw

=for html --><dd>

day and hour of observation, indicator for wind speed (unit)

=for html </dd>

=cut

    # group YYGGiw
    if ($msg !~ m@^(($re_day)($re_hour)([0134/])) (.*)@) {
        $report{ERROR} = _makeErrorMsg 'obsTimeWindInd', $corr_msg, $msg;
        return %report;
    }
    $msg = $5;
    $report{obsTime}{s} = "$2$3";
    @{$report{obsTime}}{'day', 'hour', 'minute'} = ($2, $3, '00');
    $report{windIndicator}{s} = $4;
    if ($4 ne '/') {
        $windUnit = $4 < 2 ? 'MPS' : 'KT';
        $windEST = $4 == 0 || $4 == 3;
        $report{windIndicator}{windUnit} = $windUnit;
        $report{windIndicator}{isEstimated} = undef if $windEST;
    } else {
        $report{windIndicator}{notAvailable} = undef;
    }
    if ($report{obsTime}{hour} =~ /00|06|12|18/) {
        $period = '6h';
    } elsif ($report{obsTime}{hour} =~ /03|09|15|21/) {
        $period = '3h';
    } else {
        $period = '1h';
    }

=begin html

=item

group IIiii or 99L<sub>a</sub>L<sub>a</sub>L<sub>a</sub> Q<sub>c</sub>L<sub>o</sub>L<sub>o</sub>L<sub>o</sub>L<sub>o</sub> <!--

=end html

=item group IIiii or 99LaLaLa QcLoLoLoLo

=for html --><dd>

for fixed land stations: station identification

for sea or mobile land stations: position of the station

=for html </dd>

=cut

    if ($report{obsStationType}{s} eq 'AAXX') {
        # group IIiii
        if ($msg !~ m@^($re_IIiii) (.*)@ || !($region = _IIiii2region $1)) {
            $report{ERROR} = _makeErrorMsg 'stationId', $corr_msg, $msg;
            return %report;
        }
        $msg = $2;
        $report{stationId} = { s => $1, id => $1, region => $region };
    } else {
        my ($lat_unit, $lon_unit);

        # group 99LaLaLa QcLoLoLoLo
        if ($msg =~ m@^(99/// /////) (.*)@) {
            $msg = $2;
            $report{stationPosition}{s} = $1;
            $report{stationPosition}{notAvailable} = undef;
            $lat_unit = '';
            $lon_unit = '';
        } elsif ($msg =~ m@^(99(\d(\d)\d) ([1357])(\d\d(\d)\d)) (.*)@) {
            $msg = $7;
            $report{stationPosition}{s} = $1;
            $report{stationPosition}{lat} = $2 / 10;
            $report{stationPosition}{lat} *= -1 if $4 eq '3' || $4 eq '5';
            $report{stationPosition}{lon} = $5 / 10;
            $report{stationPosition}{lon} *= -1 if $4 eq '5' || $4 eq '7';
            $lat_unit = $3;
            $lon_unit = $6;
        } else {
            $report{ERROR} = _makeErrorMsg 'stationPosition', $corr_msg, $msg;
            return %report;
        }

=begin html

=item

group MMMU<sub>La</sub>U<sub>Lo</sub> h<sub>0</sub>h<sub>0</sub>h<sub>0</sub>h<sub>0</sub>i<sub>m</sub> <!--

=end html

=item group MMMULaULo h0h0h0h0im

=for html --><dd>

for mobile land stations: position of the station (Marsden square, height)

=for html </dd>

=back

=cut

        if ($report{obsStationType}{s} eq 'OOXX') {
            # group MMMULaULo h0h0h0h0im
            if (   $msg !~ m@^((?:///|($re_MMM))(?://|($re_ULaULo)) (?:////|(\d{4}))([1-8])) (.*)@
                || (   defined $2
                    && !(($2 >= 1 && $2 <= 623) || ($2 >= 901 && $2 <= 936)))
                || (defined $3 && "$lat_unit$lon_unit" ne $3))
            {
                $report{ERROR} = _makeErrorMsg 'stationPosition',$corr_msg,$msg;
                return %report;
            }
            $msg = $6;
            $report{stationPosition}{s} .= " $1";
            $report{stationPosition}{marsdenSquare} = $2 + 0 if defined $2;
            if (defined $4) {
                $report{stationPosition}{elevation} = {
                    v => $4 + 0,
                    u => ($5 <= 4 ? 'M' : 'FT'),
                    q => 'confidenceIs' .
                         { 1 => 'Excellent', 2 => 'Good',
                           3 => 'Fair',      0 => 'Poor' }->{$5 % 4}
                };
            }
        }
    }

=for comment ###########################################################

=head3 Section 1: data for global exchange common for all code forms

=for comment ###########################################################

    iRixhVV Nddff (00fff) 1snTTT (2snTdTdTd|29UUU) 3P0P0P0P0 (4PPPP|4a3hhh) 5appp 6RRRtR (7wwW1W2|7wawaWa1Wa2) 8NhCLCMCH 9GGgg

=over

=begin html

=item

group i<sub>R</sub>i<sub>x</sub>hVV <!--

=end html

=item group iRixhVV

=for html --><dd>

precipitation indicator, weather indicator, base of lowest cloud

=for html </dd>

=cut

    # group iRixhVV
    # EXTENSION: iR can be '/': notAvailable
    # EXTENSION: ix can be '/': notAvailable
    if ($msg !~ m@^(([0-4/])([1-7/])([\d/])([0-46-9]\d|5[06-9]|//)) (.*)@) {
        $report{ERROR} = _makeErrorMsg 'indicatorCloudVis', $corr_msg, $msg;
        return %report;
    }
    $msg = $6;
    $report{precipInd}{s} = $2;
    if ($2 eq '/') {
        $report{precipInd}{notAvailable} = undef;
    } else {
        # ( '13', '1', '3', '0', 'NA' )[$2];
        $report{precipInd}{precipIndVal} = $2;
    }
    $report{wxInd}{s} = $3;
    if ($3 eq '/') {
        $report{wxInd}{notAvailable} = undef;
    } else {
        $report{wxInd}{wxIndVal} = $3;
    }

    # 1 Manned Included
    # 2 Manned Omitted (no significant phenomenon to report)
    # 3 Manned Omitted (no observation, data not available)
    # 4 Automatic Included using Code tables 4677 and 4561
    # 5 Automatic Omitted (no significant phenomenon to report)
    # 6 Automatic Omitted (no observation, data not available)
    # 7 Automatic Included using Code tables 4680 and 4531
    if (   exists $report{wxInd}{wxIndVal}
        && $report{wxInd}{wxIndVal} >= 4)
    {
        $report{reportModifier}{s} = 'AUTO';
        $report{reportModifier}{modifierType} = $report{reportModifier}{s};
    }

    $report{baseLowestCloud} = _codeTable1600 $4;
    $report{baseLowestCloud}{s} = $4;

    _codeTable4377 \%report, $5;

=begin html

=item

group Nddff (00fff) <!--

=end html

=item group Nddff (00fff)

=for html --><dd>

total cloud cover, wind direction and speed

=cut

    # group Nddff (00fff)
    if ($msg !~ m@^([\d/])($re_wind_dir|//|99)(\d\d|//) (.*)@) {
        $report{ERROR} = _makeErrorMsg 'cloudWind', $corr_msg, $msg;
        return %report;
    }
    $msg = $4;
    %{$report{totalCloudCover}} = _codeTable2700 $1;
    $report{sfcWind}{s} = "$2$3";
    if ("$2$3" eq '////') {
        $report{sfcWind}{wind}{notAvailable} = undef;
    } elsif ($2 eq '00') {
        $report{sfcWind}{wind}{isCalm} = undef;
        $report{sfcWind}{wind}{isEstimated} = undef if $windEST;
    } else {
        if ($2 eq '//') {
            $report{sfcWind}{wind}{dirNotAvailable} = undef;
        } elsif ($2 eq '99') {
            $report{sfcWind}{wind}{dirVarAllUnk} = undef;
        } else {
            $report{sfcWind}{wind}{dir} = $2 * 10;
        }
        if ($3 eq '//') {
            $report{sfcWind}{wind}{speedNotAvailable} = undef;
        } else {
            $report{sfcWind}{wind}{speed} = { v => $3 + 0, u => $windUnit };
        }
        $report{sfcWind}{wind}{isEstimated} = undef if $windEST;
    }

    if (   exists $report{sfcWind}{wind}{speed}
        && $report{sfcWind}{wind}{speed} == 99)
    {
        if ($msg !~ m@^(00([1-9]\d\d)) (.*)@) {
            $report{ERROR} = _makeErrorMsg 'wind', $corr_msg, $msg;
            return %report;
        }
        $msg = $3;
        $report{sfcWind}{s} .= " $1";
        $report{sfcWind}{wind}{speed}{v} = $2 + 0;
    }

=item

=begin html

group 1s<sub>n</sub>TTT <!--

=end html

=item group 1snTTT

=for html --><dd>

temperature

=for html </dd>

=cut

    # group 1snTTT
    if ($msg =~ m@^(1(?:////|([01]\d{3}))) (.*)@) {
        $msg = $3;
        $report{temperature}{s} = $1;
        if (defined $2) {
            $report{temperature}{air}{temp} = _parseTemp $2;
        } else {
            $report{temperature}{air}{notAvailable} = undef;
        }
    }

=begin html

=item

group 2s<sub>n</sub>T<sub>d</sub>T<sub>d</sub>T<sub>d</sub> or 29UUU <!--

=end html

=item group 2snTdTdTd or 29UUU

=for html --><dd>

dewpoint or relative humidity

=for html </dd>

=cut

    # group 2snTdTdTd|29UUU
    if ($msg =~ m@^(2(?:////|([01]\d{3})|9(100|0\d\d))) (.*)@) {
        $msg = $4;
        $report{temperature}{s} .= ' ' if exists $report{temperature};
        $report{temperature}{s} .= $1;
        if (defined $2) {
            $report{temperature}{dewpoint}{temp} = _parseTemp $2;
            if (   exists $report{temperature}{air}
                && exists $report{temperature}{air}{temp})
            {
                @{$report{temperature}}{'relHumid1', 'relHumid2',
                                        'relHumid3', 'relHumid4'}
                    = _getHumidity $report{temperature}{air}{temp}{v},
                                   $report{temperature}{dewpoint}{temp}{v};
            }
        } elsif (defined $3) {
            $report{temperature}{relHumid1} = $3 + 0;
        } else {
            $report{temperature}{dewpoint}{notAvailable} = undef;
        }
    }

=begin html

=item

group 3P<sub>0</sub>P<sub>0</sub>P<sub>0</sub>P<sub>0</sub> <!--

=end html

=item group 3P0P0P0P0

=for html --><dd>

station level pressure

=for html </dd>

=cut

    # group 3P0P0P0P0
    if ($msg =~ m@^(3(?:(\d{4})|////)) (.*)@) {
        $msg = $3;
        $report{stationPressure}{s} = $1;
        if (defined $2) {
            $report{stationPressure}{hPa} = $2 / 10;
            $report{stationPressure}{hPa} += 1000
                if $report{stationPressure}{hPa} < 500;
            # force .0
            $report{stationPressure}{hPa} =
                                  sprintf '%.1f', $report{stationPressure}{hPa};
        } else {
            $report{stationPressure}{notAvailable} = undef;
        }
    }
    if (   !exists $report{stationPressure}
        && $report{obsStationType}{s} eq 'AAXX')
    {
        push @{$report{warning}}, { warningType => 'pressureMissing' };
    }

=begin html

=item

group 4PPPP or 4a<sub>3</sub>hhh <!--

=end html

=item group 4PPPP or 4a3hhh

=for html --><dd>

sea level pressure or geopotential height of an agreed standard isobaric surface

=for html </dd>

=cut

    # group 4PPPP|4a3hhh
    # TODO: how to detect if it is 4a3hhh?
    if ($msg =~ m@^(4(?:(\d{4})|////)) (.*)@) {
        $msg = $3;
        $report{SLPhPa}{s} = $1;
        if (defined $2) {
            $report{SLPhPa}{hPa} = $2 / 10;
            $report{SLPhPa}{hPa} += 1000 if $report{SLPhPa}{hPa} < 500;
            # force .0
            $report{SLPhPa}{hPa} = sprintf '%.1f', $report{SLPhPa}{hPa};
        } else {
            $report{SLPhPa}{notAvailable} = undef;
        }
    }

=begin html

=item

group 5appp <!--

=end html

=item group 5appp

=for html --><dd>

pressure tendency and amount of change

=cut

    # group 5appp
    if ($msg =~ m@^(5([0-8])(\d{3})) (.*)@) {
        $msg = $4;
        $report{pressureChange} = {
            s                 => $1,
            period            => '3h',
            pressureTendency  => $2,
            pressureChangeVal => {
                v => $3 / 10 * ($2 >= 5 ? -1 : 1),
                u => 'hPa'
        }};
    } elsif ($msg =~ m@^(5////) (.*)@) {
        $msg = $2;
        $report{pressureChange} = {
            s            => $1,
            period       => '3h',
            notAvailable => undef
        };
    }

=begin html

=item

group 6RRRt<sub>R</sub> <!--

=end html

=item group 6RRRtR

=for html --><dd>

amount of precipitation for given period

=for html </dd>

=cut

    # group 6RRRtR
    if ($msg =~ m@^(6(\d{3})(\d)) (.*)@) {
        if (   exists $report{precipInd}{precipIndVal}
            && $report{precipInd}{precipIndVal} != 0
            && $report{precipInd}{precipIndVal} != 1)
        {
            push @{$report{warning}}, { warningType => 'precipNotOmitted1' };
        }

        $msg = $4;
        $report{precipHourly}{s} = $1;
        if ($3 > 0) {
            $report{precipHourly}{period} = _codeTable4019 $3;
        } else {
            $report{precipHourly}{periodNotAvailable} = undef;
        }
        if ($2 == 0) {
            $report{precipHourly}{precipAmount} = { v => 0, u => 'MM' };
        } elsif ($2 <= 989) {
            $report{precipHourly}{precipAmount} = { v => $2 + 0, u => 'MM' };
        } elsif ($2 == 990) {
            $report{precipHourly}{precipAmountTraces} = undef;
        } else {
            $report{precipHourly}{precipAmount} =
                                            { v => ($2 - 990) / 10, u => 'MM' };
        }
    } elsif (   exists $report{precipInd}{precipIndVal}
             && (   $report{precipInd}{precipIndVal} == 0
                 || $report{precipInd}{precipIndVal} == 1))
    {
        push @{$report{warning}}, { warningType => 'precipOmitted1' };
    }

=begin html

=item

group 7wwW<sub>1</sub>W<sub>2</sub> or 7w<sub>a</sub>w<sub>a</sub>W<sub>a1</sub>W<sub>a2</sub> <!--

=end html

=item group 7wwW1W2 or 7wawaWa1Wa2

=for html --><dd>

present and past weather

=for html </dd>

=cut

    # group (7wwW1W2|7wawaWa1Wa2)
    if ($msg =~ m@^(7(\d\d|//)([\d/])([\d/])) (.*)@) {
        if (   exists $report{wxInd}{wxIndVal}
            && $report{wxInd}{wxIndVal} != 1
            && $report{wxInd}{wxIndVal} != 4
            && $report{wxInd}{wxIndVal} != 7)
        {
            push @{$report{warning}}, { warningType => 'weatherNotOmitted' };
        }

        $msg = $5;
        $report{weatherSynop}{s} = $1;
        if ($2 eq '//') {
            $report{weatherSynop}{weatherPresentNotAvailable} = undef;
        } else {
            $report{weatherSynop}{weatherPresent} = $2;
        }
        $report{weatherSynop}{period} = $period;
        if ($3 eq '/') {
            $report{weatherSynop}{weatherPast1NotAvailable} = undef;
        } else {
            $report{weatherSynop}{weatherPast1} = $3;
        }
        if ($4 eq '/') {
            $report{weatherSynop}{weatherPast2NotAvailable} = undef;
        } else {
            $report{weatherSynop}{weatherPast2} = $4;
        }
    } elsif (   exists $report{wxInd}{wxIndVal}
             && (   $report{wxInd}{wxIndVal} == 1
                 || $report{wxInd}{wxIndVal} == 4
                 || $report{wxInd}{wxIndVal} == 7))
    {
        push @{$report{warning}}, { warningType => 'weatherOmitted' };
    }

=begin html

=item

group 8N<sub>h</sub>C<sub>L</sub>C<sub>M</sub>C<sub>H</sub> <!--

=end html

=item group 8NhCLCMCH

=for html --><dd>

cloud type for each level and cloud cover of lowest reported cloud type

=for html </dd>

=cut

    # group 8NhCLCMCH
    if ($msg =~ m@^(8////|8([0-8/])([\d/])([\d/])([\d/])) (.*)@) {
        $msg = $6;
        if (defined $2) {
            $report{cloudTypes} = {
                $3 eq '/' ? ('cloudTypeLowNA', undef)   :('cloudTypeLow', $3),
                $4 eq '/' ? ('cloudTypeMiddleNA', undef):('cloudTypeMiddle',$4),
                $5 eq '/' ? ('cloudTypeHighNA', undef)  :('cloudTypeHigh', $5),
            };
            if ($2 eq '/') {
                $report{cloudTypes}{oktasNotAvailable} = undef;
            } elsif ($3 ne '/' and $3 != 0) {
                $report{cloudTypes}{oktasLow} = $2;
            } elsif ($4 ne '/' and $4 != 0) {
                $report{cloudTypes}{oktasMiddle} = $2;
            }
        } else {
            $report{cloudTypes}{notAvailable} = undef;
        }
        $report{cloudTypes}{s} = $1;
    }

=begin html

=item

group 9GGgg <!--

=end html

=item group 9GGgg

=for html --><dd>

exact observation time

=back

=cut

    # group 9GGgg
    if ($msg =~ m@^(9($re_hour)($re_min)) (.*)@) {
        $msg = $4;
        @{$report{exactObsTime}}{'s', 'hour', 'minute'} = ($1, $2, $3);
    }

=for comment ###########################################################

=head3 Section 2: maritime data for global exchange (optional)

=for comment ###########################################################

=cut

    if ($msg =~ m@^(222[\d/]{2}(?: (?:[\d/]{5}|ICE))+) (.*)@) {
        $msg = $2;
        push @{$report{warning}}, { warningType => 'notProcessed', s => $1 };
    }

=for comment ###########################################################

=head3 Section 3: data for regional exchange (optional)

=for comment ###########################################################

The WMO regions are:

=over

=item

I (Africa)

=item

II (Asia)

=item

III (South America)

=item

IV (North and Central Amerika)

=item

V (South-West Pacific)

=item

VI (Europe)

=item

Antarctic

=back

    region I: 0TgTgRcRt 1snTxTxTx 2snTnTnTn 4E'sss 5j1j2j3j4 (j5j6j7j8j9) 6RRRtR 7R24R24R24R24 8NsChshs 9SPSPspsp 80000 0LnLcLdLg (1sLdLDLve)

    region II: 0EsnT'gT'g 1snTxTxTx 2snTnTnTn 3EsnTgTg 4E'sss 5j1j2j3j4 (j5j6j7j8j9) 6RRRtR 7R24R24R24R24 8NsChshs 9SPSPspsp

    region III: 1snTxTxTx 2snTnTnTn 3EsnTgTg 4E'sss 5j1j2j3j4 (j5j6j7j8j9) 6RRRtR 7R24R24R24R24 8NsChshs 9SPSPspsp

    region V: 1snTxTxTx 2snTnTnTn 4E'sss 5j1j2j3j4 (j5j6j7j8j9) 6RRRtR 7R24R24R24R24 8NsChshs 9SPSPspsp

    region VI: 1snTxTxTx 2snTnTnTn 3EsnTgTg 4E'sss 5j1j2j3j4 (j5j6j7j8j9) 6RRRtR 7R24R24R24R24 8NsChshs 9SPSPspsp

=cut

    if (defined $region && $msg =~ m@^333 (.*)@) {
        my (@s3, $msg5, $r);

        $msg = $1;
        @s3  = ();

        if ($region =~ /^(?:I|II|III|V|VI)$/) {

=over

=begin html

=item

group 0T<sub>g</sub>T<sub>g</sub>R<sub>c</sub>R<sub>t</sub> (region I) <!--

=end html

=item group 0TgTgRcRt (region I)

=for html --><dd>

minumum ground temperature last night, character and start/end of precipitation

=for html </dd>

=cut

            # region I: group 0TgTgRcRt
            if ($region eq 'I' && $msg =~ m@^(?:0(//|\d\d)(\d)(\d)) (.*)@) {
                my $begin_end;

                $msg = $4;
                $r->{s} = "0$1";
                if ($1 eq '//') {
                    $r->{notAvailable} = undef;
                } else {
                    $r->{period} = 'n';
                    $r->{temp}{v} = $1 > 50 ? -($1 - 50) : $1 + 0;
                    $r->{temp}{u} = 'C';
                }
                push @s3, { tempMinGround => $r };
                $r = undef;

                # WMO-No. 306 Vol II, code table 167
                $r->{s} = $2;
                if ($2 == 1 || $2 == 5) {
                    $r->{phenomDescr} = 'isLight';
                } elsif ($2 == 2 || $2 == 6) {
                    $r->{phenomDescr} = 'isModerate';
                } elsif ($2 == 3 || $2 == 7) {
                    $r->{phenomDescr} = 'isHeavy';
                } elsif ($2 == 4 || $2 == 8) {
                    $r->{phenomDescr} = 'isVeryHeavy';
                } elsif ($2 == 9) {
                    $r->{phenomDescr} = 'isVariable';
                }
                if ($2 == 0) {
                    $r->{noPrecip} = undef;
                } elsif ($2 < 5) {
                    $r->{phenomDescr2} = 'isIntermittent';
                } elsif ($2 < 9) {
                    $r->{phenomDescr2} = 'isContinuous';
                }
                push @s3, { precipCharacter => $r };
                $r = undef;

                # WMO-No. 306 Vol II, code table 168
                $r->{s} = $3;
                if ($3 == 0) {
                    $r->{noPrecip} = undef;
                } elsif ($3 == 1) {
                    $r->{hoursTill} = { v => 1, q => 'isLess' };
                } elsif ($3 < 7) {
                    $r->{hoursFrom} = $3 - 1;
                    $r->{hoursTill} = $3;
                } elsif ($3 < 9) {
                    $r->{hoursFrom} = ($3 - 7) * 2 + 6;
                    $r->{hoursTill} = ($3 - 7) * 2 + 8;
                } else {
                    $r->{hoursFrom} = { v => 10, q => 'isGreater' };
                }
                # WMO-No. 306 Vol I.1 Part A, code table 4677:
                #   00..49 is not precipitation
                $begin_end =       exists $report{weatherSynop}
                                && exists $report{weatherSynop}{weatherPresent}
                                && $report{weatherSynop}{weatherPresent} >= 50
                             ? 'beginPrecip' : 'endPrecip';
                push @s3, { $begin_end => $r };
                $r = undef;
            }

=begin html

=item

group 0Es<sub>n</sub>T'<sub>g</sub>T'<sub>g</sub> (region II) <!--

=end html

=item group 0EsnT'gT'g (region II)

=for html --><dd>

state of ground, minimum ground temperature last night

=for html </dd>

=cut

            # EXTENSION: allow E for state of the ground
            if ($region eq 'II' && $msg =~ m@^(0([\d/E]))(///|[01]\d\d) (.*)@) {
                $msg = $4;
                $r->{s} = $1;
                if ($2 eq '/' || $2 eq 'E') {
                    $r->{notAvailable} = undef;
                } else {
                    $r->{stateOfGroundVal} = $2;
                }
                push @s3, { stateOfGround => $r };
                $r = undef;

                $r->{s} = $3;
                if ($3 eq '///') {
                    $r->{notAvailable} = undef;
                } else {
                    $r->{period} = 'n';
                    $r->{temp} = _parseTemp "${3}0";
                    $r->{temp}{v} += 0;
                }
                push @s3, { tempMinGround => $r };
            }
            $r = undef;


=begin html

=item

group 1s<sub>n</sub>T<sub>x</sub>T<sub>x</sub>T<sub>x</sub> <!--

=end html

=item group 1snTxTxTx

=for html --><dd>

maximum temperature (regions I, II, VI: day-time in the last 12 hours, region III: day-time, region V: in the last 24 hours)

=for html </dd>

=cut

            # group 1snTxTxTx
            if ($msg =~ m@^(1([01]\d{3})) (.*)@) {
                $msg = $3;

                $r->{s} = $1;
                $r->{temp} = _parseTemp $2;
                $r->{period} = '12h' if $region =~ /^(?:I|II|VI)$/;
                if ($region eq 'V') {
                    $r->{period} = '24h';
                    push @s3, { tempMaxPeriod => $r };
                } else {
                    push @s3, { tempMaxDaytime => $r };
                }
            } elsif ($msg =~ m@^(1////) (.*)@) {
                $msg = $2;

                $r->{s} = $1;
                $r->{notAvailable} = undef;
                $r->{period} = '12h' if $region =~ /^(?:I|II|VI)$/;
                if ($region eq 'V') {
                    $r->{period} = '24h';
                    push @s3, { tempMaxPeriod => $r };
                } else {
                    push @s3, { tempMaxDaytime => $r };
                }
            }
            $r = undef;

=begin html

=item

group 2s<sub>n</sub>T<sub>n</sub>T<sub>n</sub>T<sub>n</sub> <!--

=end html

=item group 2snTnTnTn

=for html --><dd>

minimum temperature (regions I, II, VI: night-time in the last 12 hours, region III: night-time, region V: in the last 24 hours)

=for html </dd>

=cut

            # group 2snTnTnTn
            if ($msg =~ m@^(2([01]\d{3})) (.*)@) {
                $msg = $3;

                $r->{s} = $1;
                $r->{temp} = _parseTemp $2;
                if ($region eq 'V') {
                    $r->{period} = '24h',
                    push @s3, { tempMinPeriod => $r };
                } else {
                    $r->{period} = '12h' if $region =~ /^(?:I|II|VI)$/;
                    push @s3, { tempMinNighttime => $r };
                }
            } elsif ($msg =~ m@^(2////) (.*)@) {
                $msg = $2;

                $r->{s} = $1;
                $r->{notAvailable} = undef;
                if ($region eq 'V') {
                    $r->{period} = '24h',
                    push @s3, { tempMinPeriod => $r };
                } else {
                    $r->{period} = '12h' if $region =~ /^(?:I|II|VI)$/;
                    push @s3, { tempMinNighttime => $r };
                }
            }
            $r = undef;

=begin html

=item

group 3Es<sub>n</sub>T<sub>g</sub>T<sub>g</sub> (region VI) <!--

=end html

=item group 3EsnTgTg (regions II, VI)

=for html --><dd>

state of ground, minimum ground temperature last night

=for html </dd>

=cut

            # region II: group 3EsnTgTg
            # region VI: group 3EsnTgTg
            if (   ($region eq 'II' || $region eq 'VI')
                && $msg =~ m@^(?:3([\d/])(///|[01]\d\d)) (.*)@)
            {
                $msg = $3;
                $r->{s} = "3$1";
                if ($1 eq '/') {
                    $r->{notAvailable} = undef;
                } else {
                    $r->{stateOfGroundVal} = $1;
                }
                push @s3, { stateOfGround => $r };
                $r = undef;

                $r->{s} = $2;
                if ($2 eq '///') {
                    $r->{notAvailable} = undef;
                } else {
                    $r->{period} = 'n';
                    $r->{temp} = _parseTemp "${2}0";
                    $r->{temp}{v} += 0;
                }
                push @s3, { tempMinGround => $r };
                $r = undef;
            }

=begin html

=item

group 4E'sss <!--

=end html

=item group 4E'sss

=for html --><dd>

state of ground if covered with snow or ice, snow depth

=cut

            # group 4E'sss
            if ($msg =~ m@^(?:4([\d/])(///|\d{3})) (.*)@) {
                $msg = $3;

                # WMO-No. 306 Vol I.1 Part A, code table 0975
                $r->{s} = "4$1";
                if ($1 eq '/') {
                    $r->{notAvailable} = undef;
                } else {
                    $r->{stateOfGroundSnowVal} = $1;
                }
                push @s3, { stateOfGroundSnow => $r };
                $r = undef;

                # WMO-No. 306 Vol I.1 Part A, code table 3889
                $r->{s} = $2;
                if ($2 eq '///') {
                    $r->{notAvailable} = undef;
                } elsif ($2 eq '000') {
                    $r->{invalidFormat} = undef;
                } elsif ($2 eq '997') {
                    $r->{snowDepthVal} = { v => 0.5, u => 'CM', q => 'isLess' };
                } elsif ($2 eq '998') {
                    $r->{snowDepthCoverNotCont} = undef;
                } elsif ($2 eq '999') {
                    $r->{snowDepthNoMeasurement} = undef;
                } else {
                    $r->{snowDepthVal} = { v => $2 + 0, u => 'CM' };
                }
                push @s3, { snowDepth => $r };
                $r = undef;
            }

=begin html

=item

group 5j<sub>1</sub>j<sub>2</sub>j<sub>3</sub>j<sub>4</sub> (j<sub>5</sub>j<sub>6</sub>j<sub>7</sub>j<sub>8</sub>j<sub>9</sub>) <!--

=end html

=item group 5j1j2j3j4 (j5j6j7j8j9)

=for html --><dd>

evaporation, temperature change, duration of sunshine, radiation type and amount, direction of cloud drift, direction and elevation of cloud, pressure change

=for html </dd>

=cut

            # group 5j1j2j3j4 (j5j6j7j8j9)

            # determine all 5xxxx groups with 0xxxx .. 6xxxx suppl. groups:
            # 1. use only consecutive groups 0xxxx .. 6xxxx
            $msg5 = ($msg =~ /^(5[\d\/]{4} (?:[0-6][\d\/]{4} )*)/)[0];
            $msg5 = '' unless defined $msg5;
            # 2. remove trailing 6xxxx if precipitation indicator is 0 or 2
            if (   exists $report{precipInd}{precipIndVal}
                && (   $report{precipInd}{precipIndVal} == 0
                    || $report{precipInd}{precipIndVal} == 2))
            {
                $msg5 =~ s/6[\d\/]{4} $//;
            }

            # strip $msg5 from $msg, use $msg5 for group(s) 5j1j2j3j4
            $msg = substr $msg, length $msg5;

            # group 5EEEiE
            if ($msg5 =~ m@^(5([0-3]\d\d)(\d)) (.*)@) {
                $msg5 = $4;
                $r->{s} = $1;
                $r->{evapoAmount} = $2 / 10;
                $r->{evapoIndicator} = $3;
                push @s3, { evapo => $r };
                $r = undef;
            } elsif ($msg5 =~ m@^(5////) (.*)@) {
                $msg5 = $2;
                push @s3, { evapo => {
                    s            => $1,
                    notAvailable => undef
                }};
            }

            # group 54g0sndT
            if ($msg5 =~ m@^(54([0-5])([01])(\d)) (.*)@) {
                $msg5 = $5;
                $r->{s} = $1;
                $r->{hoursFrom} = $2;
                $r->{hoursTill} = $2 + 1;

                # WMO-No. 306 Vol I.1 Part A, code table 0822
                $r->{temp}{v} = $4 < 5 ? $4 + 10 : $4 + 0;
                $r->{temp}{v} *= -1 if $3 == 1;
                $r->{temp}{u} = 'C';
                push @s3, { tempChange => $r };
                $r = undef;
            }

            # group 55SSS (j5j6j7j8j9)*, SSS = 000..240
            if ($msg5 =~ m@(55(?:///|(${re_hour}\d|240))) (.*)@) {
                $msg5 = $3;
                $r->{s} = $1;
                $r->{sunshinePeriod} = '1d';
                if (defined $2) {
                    $r->{sunshine} = { v => $2 / 10, u => 'hour' };
                } else {
                    $r->{sunshineNotAvailable} = undef;
                }
                for ($msg5 =~ m@^(0(?:////|\d{4}) )?(1(?:////|\d{4}) )?(2(?:////|\d{4}) )?(3(?:////|\d{4}) )?(4(?:////|\d{4}) )?(5[0-4]\d{3} )?(6(?:////|\d{4}) )?@)
                {
                    next unless defined $_;

                    $msg5 = substr $msg5, 6;

                    /(.)(.*) /;
                    $r->{s} .= " $1$2";
                    $r->{radiationPeriod} = '24h';
                    if ($2 eq '////') {
                        $r->{_radiationType $1}{notAvailable} = undef;
                    } else {
                        $r->{_radiationType $1}{radiationValue} =
                                                  { v => $2 + 0, u => 'Jcmcm' };
                    }
                }
                push @s3, { radiationSun => $r };
                $r = undef;
            }

            # group 553SS (j5j6j7j8j9)*, SS = 00..10
            if ($msg5 =~ m@(553(?://|(0\d|10))) (.*)@) {
                $msg5 = $3;
                $r->{s} = $1;
                $r->{sunshinePeriod} = '1h';
                if (defined $2) {
                    $r->{sunshine} = { v => $2 / 10, u => 'hour' };
                } else {
                    $r->{sunshineNotAvailable} = undef;
                }
                for ($msg5 =~ m@^(0(?:////|\d{4}) )?(1(?:////|\d{4}) )?(2(?:////|\d{4}) )?(3(?:////|\d{4}) )?(4(?:////|\d{4}) )?(5[0-4]\d{3} )?(6(?:////|\d{4}) )?@)
                {
                    next unless defined $_;

                    $msg5 = substr $msg5, 6;

                    /(.)(.*) /;
                    $r->{s} .= " $1$2";
                    $r->{radiationPeriod} = '1h';
                    if ($2 eq '////') {
                        $r->{_radiationType $1}{notAvailable} = undef;
                    } else {
                        $r->{_radiationType $1}{radiationValue} =
                                                   { v => $2 + 0, u => 'kJmm' };
                    }
                }
                push @s3, { radiationSun => $r };
                $r = undef;
            }

            while ($msg5 =~ m@^(55([45])0([78])) \2(\d{4}|////) (.*)@) {
                $msg5 = $5;
                $r->{s} = "$1 $2$4";
                $r->{radiationPeriod} = $2 eq '4' ? '1h' : '24h';
                if ($4 eq '////') {
                    $r->{notAvailable} = undef;
                } else {
                    $r->{radiationValue} =
                             { v => $4 + 0, u => $2 eq '4' ? 'kJmm' : 'Jcmcm' };
                }
                push @s3,
                        { ($3 eq '7' ? 'radShortWave' : 'radDirectSolar'), $r };
                $r = undef;
            }

            # group 56DLDMDH, direction of cloud drift
            if ($msg5 =~ m@^(56([\d/])([\d/])([\d/])) (.*)@) {
                $msg5 = $5;
                $r->{s} = $1;
                _codeTable0700 $r, 'TypeLow', $2;
                _codeTable0700 $r, 'TypeMiddle', $3;
                _codeTable0700 $r, 'TypeHigh', $4;
                push @s3, { cloudTypesDrift => $r };
                $r = undef;
            }

            # group 57CDaeC, direction and elevation of cloud
            while ($msg5 =~ m@^(57([\d/])(\d)(\d)) (.*)@) {
                $msg5 = $5;
                $r->{s} = $1;
                _codeTable0500 $r, $2;
                _codeTable0700 $r, '', $3;
                _codeTable1004 $r, $4;
                push @s3, { cloudLocation => $r };
                $r = undef;
            }

            # group 58p24p24p24
            if ($msg5 =~ m@^(58(?:(\d{3})|///)) (.*)@) {
                $msg5 = $3;
                push @s3, { pressureChange => {
                    s      => $1,
                    period => '24h',
                    defined $2
                           ? (pressureChangeVal => { v => $2 / 10, u => 'hPa' })
                           : ('notAvailable', undef)
                }};
            # group 59p24p24p24
            } elsif ($msg5 =~ m@^(59(?:(\d{3})|///)) (.*)@) {
                $msg5 = $3;
                push @s3, { pressureChange => {
                    s      => $1,
                    period => '24h',
                    defined $2
                        # avoid "-0"
                      ? (pressureChangeVal => { v => (-$2 / 10)*1, u => 'hPa' })
                      : ('notAvailable', undef)
                }};
            }

            # EXTENSION: allow un-announced 6xxxx group
            if ($msg5 =~ m@^6(?:\d{3}|///)\d $@) {
                $msg = $msg5 . $msg;
                $msg5 = '';
            }
            if ($msg5 ne '') {
                $msg = $msg5 . $msg;
                $report{ERROR} = _makeErrorMsg 'invalid333-5xxxx',
                                               $corr_msg, $msg;
                $report{section3} = \@s3;
                return %report;
            }

=begin html

=item

group 6RRRt<sub>R</sub> <!--

=end html

=item group 6RRRtR

=for html --><dd>

amount of precipitation for given period

=for html </dd>

=cut

            # group 6RRRtR
            if ($msg =~ m@^(6(\d{3}|///)(\d)) (.*)@) {
                if (   exists $report{precipInd}{precipIndVal}
                    && $report{precipInd}{precipIndVal} != 0
                    && $report{precipInd}{precipIndVal} != 2)
                {
                    push @{$report{warning}},
                                         { warningType => 'precipNotOmitted3' };
                }

                $msg = $4;
                $r->{s} = $1;
                if ($3 > 0) {
                    $r->{period} = _codeTable4019 $3;
                } else {
                    $r->{periodNotAvailable} = undef;
                }
                if ($2 eq '///') {
                    $r->{precipAmountNotAvailable} = undef;
                } elsif ($2 == 0) {
                    # EXTENSION: allow '000' (should not be used)
                    $r->{precipAmount} = { v => 0, u => 'MM' };
                } elsif ($2 <= 989) {
                    $r->{precipAmount} = { v => $2 + 0, u => 'MM' };
                } elsif ($2 == 990) {
                    $r->{precipAmountTraces} = undef;
                } else {
                    $r->{precipAmount} = { v => ($2 - 990) / 10, u => 'MM' };
                }
                push @s3, { precipHourly => $r };
                $r = undef;
            } elsif (   exists $report{precipInd}{precipIndVal}
                     && (   $report{precipInd}{precipIndVal} == 0
                         || $report{precipInd}{precipIndVal} == 2))
            {
                push @{$report{warning}}, { warningType => 'precipOmitted3' };
            }

=begin html

=item

group 7R<sub>24</sub>R<sub>24</sub>R<sub>24</sub>R<sub>24</sub> <!--

=end html

=item group 7R24R24R24R24

=for html --><dd>

amount of precipitation in the last 24 hours

=for html </dd>

=cut

            if ($msg =~ m@^(7(\d{4})) (.*)@) {
                $msg = $3;
                $r->{s} = $1;
                $r->{period} = '24h';
                if ($2 == 9999) {
                    $r->{precipAmountTraces} = undef;
                } else {
                    $r->{precipAmount} = { v => $2 / 10, u => 'MM' };
                }
                push @s3, { precipHourly => $r };
                $r = undef;
            }

=begin html

=item

group 8N<sub>s</sub>Ch<sub>s</sub>h<sub>s</sub> <!--

=end html

=item group 8NsChshs

=for html --><dd>

cloud cover and height for cloud layers

=for html </dd>

=cut

            # group 8NsChshs (but not 80000!)
            while (   $msg =~ m@^(8(\d)([\d/])([0-46-9]\d|5[06-9]|//)) (.*)@
                   && $1 ne '80000')
            {
                my ($tag, $dist);

                $msg = $5;
                $r->{s} = $1;
                %{$r->{totalCloudCover}} = _codeTable2700 $2;
                _codeTable0500 $r, $3;
                $tag = $2 eq '9' && $3 eq '/' ? 'visVert': 'cloudBase';
                if ($4 eq '//') {
                    $r->{cloudBaseNotAvailable} = undef;
                } else {
                    $dist = _codeTable1677 $4;
                    if (ref $dist eq 'HASH') {
                        if ($tag eq 'visVert') {
                            $r->{$tag}{distance} = $dist;
                            $r->{$tag}{distance}{u} = 'M';
                            $r->{$tag}{s} = $4;
                        } else {
                            $r->{$tag} = $dist;
                            $r->{$tag}{u} = 'M';
                        }
                    } else {
                        if ($tag eq 'visVert') {
                            $r->{"${tag}From"}{distance}{v} = $dist->[0];
                            $r->{"${tag}From"}{distance}{u} = 'M';
                            $r->{"${tag}To"}{distance}{v} = $dist->[1];
                            $r->{"${tag}To"}{distance}{u} = 'M';
                        } else {
                            $r->{"${tag}From"}{v} = $dist->[0];
                            $r->{"${tag}From"}{u} = 'M';
                            $r->{"${tag}To"}{v} = $dist->[1];
                            $r->{"${tag}To"}{u} = 'M';
                        }
                    }
                }
                push @s3, { cloudInfo => $r };
                $r = undef;
            }

=begin html

=item

group 9S<sub>P</sub>S<sub>P</sub>s<sub>p</sub>s<sub>p</sub> <!--

=end html

=item group 9SPSPspsp

=for html --><dd>

supplementary information

=cut

            # WMO-No. 306 Vol I.1 Part A, code table 3778:
            # group 9SPSPspsp
            while ($msg =~ m@^9\d{4} @) {
                $r = undef;

                # group 90(0(tt|zz)|[15]tt)
                if ($msg =~ m@^(90(?:0\d\d|[15]$re_synop_tt)) (.*)@) {
                    my ($s, $type);
                    $s = $1;
                    $msg = $2;
                    $s =~ /..(.)(..)/;
                    $r = _codeTable4077 $2, 'timeBeforeObs';
                    if (exists $r->{timeBeforeObs}) {
                        $type = { 0 => 'weatherSynopStart',
                                  1 => 'weatherSynopEnd',
                                  5 => 'weatherSynopDuration' }->{$1};
                    } else {
                        $type = 'weatherSynopCharacter';
                    }
                    $r->{s} = $s;
                    push @s3, { $type => $r };
                # group 910ff
                } elsif ($msg =~ m@^(910(\d\d)) (.*)@) {
                    $msg = $3;
                    $r->{s} = $1;
                    $r->{period}  = '10m';
                    $r->{wind}{speed}{u} = $windUnit;
                    $r->{wind}{speed}{v} = $2 + 0;
                    if ($r->{wind}{speed}{v} == 99) {
                        if ($msg !~ m@^(00([1-9]\d\d)) (.*)@) {
                            $report{ERROR} = _makeErrorMsg 'wind',
                                                           $corr_msg, $msg;
                            $report{section3} = \@s3;
                            return %report;
                        }
                        $msg = $3;
                        $r->{s} .= " $1";
                        $r->{wind}{speed}{v} = $2 + 0;
                    }
                    push @s3, { highestGust => $r };
                # group (90[2347]tt) 91[1-4]ff (00fff)
                } elsif ($msg =~ m@^((?:90([2347])($re_synop_period) )?91([1-4])(\d\d)) (.*)@)
                {
                    my $type;
                    $msg = $6;
                    if (defined $2) {
                        $r = _codeTable4077 $3, (
                            { 2 => 'timeBeforeObsStart',
                              3 => 'timeBeforeObsEnd',
                              4 => 'timeBeforeObs',
                              7 => 'timeBeforeObs'}->{$2});
                        $r->{period} = '10m' if $2 eq '4';
                    } else {
                        $r->{period} = $period;
                    }
                    $r->{s} = $1;
                    $type = { 1 => 'highestGust',
                              2 => 'highestMeanSpeed',
                              3 => 'meanSpeed',
                              4 => 'lowestMeanSpeed' }->{$4};
                    $r->{wind}{speed}{u} = $windUnit;
                    $r->{wind}{speed}{v} = $5 + 0;
                    if ($r->{wind}{speed}{v} == 99) {
                        if ($msg !~ m@^(00([1-9]\d\d)) (.*)@) {
                            $report{ERROR} = _makeErrorMsg 'wind',
                                                           $corr_msg, $msg;
                            $report{section3} = \@s3;
                            return %report;
                        }
                        $msg = $3;
                        $r->{s} .= " $1";
                        $r->{wind}{speed}{v} = $2 + 0;
                    }
                    push @s3, { $type => $r };
                # group (907tt) 931ss
                } elsif ($msg =~ m@^((?:907($re_synop_tt) )?931(\d\d)) (.*)@) {
                    $msg = $4;
                    if (defined $2) {
                        $r = _codeTable4077 $2, 'timeBeforeObs';
                    } else {
                        $r->{period} = $period;
                    }
                    $r->{s} = $1;
                    _codeTable3870 \$r, $3;
                    push @s3, { snowDepthNew => $r };
                # group 96[024]ww
                } elsif ($msg =~ m@^(96([024])(\d\d)) (.*)@) {
                    $msg = $4;
                    push @s3, { { 0 => 'weatherSynopAdd',
                                  2 => 'weatherSynopAmplPast',
                                  4 => 'weatherSynopAmpl' }->{$2} => {
                                    s              => $1,
                                    weatherPresent => $3
                    }};
                # group 90[23467]tt 966ww
                } elsif ($msg =~ m@^(90([23467])($re_synop_period) 966(\d\d)) (.*)@)
                {
                    $msg = $5;
                    $r = _codeTable4077 $3, 'timeBeforeObs';
                    $r->{s} = $1;
                    $r->{weatherPresent} = $4;
                    push @s3, { weatherSynopPast => $r };
                # TODO: more 9xxxx stuff
                # } elsif () {
                } else {
                    last;
                }
            }

            # region I: group 80000 0LnLcLdLg (1sLdLDLve)
            if ($region eq 'I' && $msg =~ m@^80000 (.*)@) {
               # TODO
               #while ($msg =~ m@^(?:0(\d{4}))(?: 1(\d{4}))? (.*)@) {
               #}
            }

            $report{section3} = \@s3;
        }
    }

=back

=for comment ###########################################################

=head3 Section 4: data for national use for clouds with base below station level (optional)

=for comment ###########################################################

=cut

=for comment ###########################################################

=head3 Section 5: data for national use (optional)

=for comment ###########################################################

=cut

    push @{$report{warning}}, { warningType => 'notProcessed', s => $msg }
        if $msg ne '';

    return %report;
}

########################################################################
# _parseMetarTaf
########################################################################
sub _parseMetarTaf {
    my ($msg, $msg_hdr, $is_speci, $is_taf) = @_;
    my (%metar, $corr_msg);
    my ($day, $hour, $minute);
    my ($curr_day, $curr_hour, $curr_minute);
    my ($ii, $td, @cy, $delim, $tt);

    $metar{msg} = $msg_hdr . $msg;
    $metar{isSpeci} = undef if $is_speci;
    $metar{isTaf} = undef if $is_taf;

=pod

=head2 Parsing of METAR and TAF messages

The message is checked for typical errors and corrected. Errors for METARs could
be:

=over

=item *

invalid characters (everything except capital letters, digits, dot, slash,
dollar sign, space, plus sign, minus sign)

=item *

QNH with spaces

=item *

temperature or dew point with spaces

=item *

misspelled key words, or with missing or additional spaces

=item *

missing key words

=item *

wrong order of keywords

=item *

removal of slashes before and after some components

=back

If the message is modified there will be a C<warning>.

=cut

    $msg .= ' '; # this makes parsing much easier

    # EXTENSION: preprocessing
    # remove trailing =
    $msg =~ s/= $/ /os;

    # replace invalid characters
    $msg =~ s/[^A-Z\d.,\/\$ +-]/?/gos;

    # QNH with spaces
    $msg =~ s/(?<= A) (?=[23]\d{3} )//o;
    $msg =~ s/(?<= Q) (?=[01]\d{3} )//o;
    $msg =~ s/(?<= (?:A[23]|Q[01])) (?=\d{3} )//o;
    $msg =~ s/(?<= (?:A[23]|Q[01])\d) (?=\d\d )//o;
    $msg =~ s/(?<= (?:A[23]|Q[01])\d\d) (?=\d )//o;

    # misspelled key words, or with missing or additional spaces
    $msg =~ s/ (?:CAMOK|CC?VOK) / CAVOK /o;
    $msg =~ s/ (?:NO(?: S)?IG|NOSI(?: G)?|N[L ]OSIG|NSI?G|(?:MO|N0)SIG|NOS I ?G) / NOSIG /o;
    $msg =~ s/(?<! )(?=(?:TEMPO|BECMG) )/ /o; # BLU+BLU+(TEMPO|BECMG)
    $msg =~ s/ (?:R MK|RMKS|RRMK) / RMK /o;
    $msg =~ s@(?<= RMK)(?=[^ /])@ @o;
    $msg =~ s@ 0VC(?=$re_cloud_base$re_cloud_type?|$re_cloud_type )@ OVC@og;
    $msg =~ s@ BNK(?=$re_cloud_base$re_cloud_type?|$re_cloud_type )@ BKN@og;
    $msg =~ s@( $re_cloud_type| FULYR)[/-](?=$re_compass_dir16 )@$1 @og;
    $msg =~ s@(?<= VR) (?=B$re_wind_speed$re_wind_speed_unit )@@og;
    $msg =~ s@(?<= R) (?=WY)@@og;
    $msg =~ s@(?<= RW) (?=Y)@@og;
    $msg =~ s@(?<= RWY$re_rwy) (?=[LCR] )@@og;
    $msg =~ s@ RNW @ RWY @og;
    $msg =~ s@/(?=AURBO )@ @o;
    $msg =~ s@(?<= A)0(?=[12]A? )@O@o;
    $msg =~ s@(?<= ALQ)(?=S )@D@o;
    $msg =~ s@(?<= O)VH(?= )@HD@o;
    $msg =~ s@(?<= OCNL)Y(?= )@@o;
    $msg =~ s@(?<= PK)(?=WND )@ @o;
    $msg =~ s@( $re_phen_desc L)GT(?=$re_ltg_types+)@$1TG@o;
    $msg =~ s@( $re_phen_desc LTG) (?=$re_ltg_types+)@$1@o;
    $msg =~ s@ LIGHTNING @ LTG @og;
    $msg =~ s@ LTNG @ LTG @og;
    $msg =~ s@(?<= I)N(?=OVC )@@o;
    $msg =~ s@(?<= SP)O(?=TS )@@o;
    $msg =~ s@(?<= W) (?=IND )@@o;
    $msg =~ s@(?<= WIND RWY) (?=$re_rwy_des )@@o;
    $msg =~ s@(?<= (?:RWY|THR))($re_rwy_des ${re_wind_dir}0$re_wind_speed) (?=$re_wind_speed_unit )@$1@og;
    $msg =~ s@(?<= N)ORTH(?= )@@o;
    $msg =~ s@(?<= S)OUTH(?= )@@o;
    $msg =~ s@(?<= E)AST(?= )@@o;
    $msg =~ s@(?<= W)EST(?= )@@o;
    $msg =~ s@(?<= B)A(?=NK )@@o;
    $msg =~ s@(?<= F)(?= BNK )@G@o;
    $msg =~ s@( [+-]?)(RA|SN)SH @$1SH$2 @go;
    $msg =~ s@(?<= ISOL)D(?= )@@o;
    $msg =~ s@(?<= H)A(?=ZY )@@o;
    $msg =~ s@(?<= DS)(?=T )@N@o;
    $msg =~ s@(?<= D)IS(?=T )@SN@o;
    $msg =~ s@(?<= PL)(?=ME )@U@o;
    $msg =~ s@(?<= BL)K @ACK@o;
    $msg =~ s@(?<= INVIS)T(?= )@@o;
    $msg =~ s@(?<= FROIN)/ ?@ @o;
    $msg =~ s@(?<= AS)(?=CTD )@O@o;
    $msg =~ s@(?<=^[A-Z][A-Z\d]{3} $re_day$re_hour$re_min)(?= )@Z@o
        unless $is_taf;
    $msg =~ s@(?<=^[A-Z][A-Z\d]{3} $re_day$re_hour$re_min)KT(?= )@Z@o;
    $msg =~ s@(?<=^[A-Z][A-Z\d]{3} $re_day$re_hour${re_min}Z )(${re_wind_dir}0)/(?=${re_wind_speed}$re_wind_speed_unit )@$1@o;
    $msg =~ s@(?<= [0-2]) (?=\d0${re_wind_speed}$re_wind_speed_unit )@@go;
    $msg =~ s@(?<= 3) (?=[0-6]0${re_wind_speed}$re_wind_speed_unit )@@go;
    $msg =~ s@(?<=M)/(?=S )@P@o;
    $msg =~ s@(?<=[+]) (?=$re_weather_w_i )@@o;
    $msg =~ s@(?<=$re_cloud_base )(M?\d\d/)/(?=M?\d\d )@$1@o;
    $msg =~ s@(?<= BECMG| TEMPO)(?=$re_hour(?:$re_hour|24) )@ @go;
    $msg =~ s@ TMPO @ TEMPO @go;
    $msg =~ s@ PRECIP @ PCPN @o;
    $msg =~ s@( T[XN]M?\d\d/(?:$re_hour|24)Z)(?=T[XN]M?\d\d/(?:$re_hour|24)Z )@$1 @o;
    $msg =~ s@/($re_snw_cvr_title)@ $1@;
    $msg =~ s@(?<= RWY0) (?=[1-9] )@@og;
    $msg =~ s@(?<= RWY[12]) (?=\d )@@og;
    $msg =~ s@(?<= RWY3) (?=[0-6] )@@og;
    $msg =~ s@ DRFTG SNW @ DRSN @og;
    $msg =~ s@(?<= T)O(?=\d{3}[01]\d{3} )@0@og;

    # missing key words
    $msg =~ s@ (?:TEMPO|INTER) ($re_hour$re_min)/($re_hour$re_min)(?= )@ TEMPO FM$1 TL$2@og;
    $msg =~ s@(?<! (?:BECMG|TEMPO) )(?=FM$re_hour$re_min )@BECMG @og
        unless $is_taf;

    # wrong order of keywords
    $msg =~ s@( DSNT| DISTANT)( (?:$re_weather|SMOKE|HAZE|TSTMS|$re_cloud_type|LTG)(?: TO)?)(?=$re_phen_locdirs )@$2$1@og;
    $msg =~ s@( $re_weather| SMOKE| HAZE| TSTMS| $re_cloud_type| LTG)((?: TO)?$re_phen_locdirs|$re_num_quadrant) (DSNT|DISTANT) @$1 $3$2 @og;

    # komma removed
    $msg =~ s@^((?:U|ZM).* RMK.*QFE\d{3}) (?=\d )@$1,@;
    $msg =~ s@^(FQ.* RMK.*TX/\d\d) (?=\d )@$1,@;

    $corr_msg = $msg;
    if ($metar{msg} . ' ' ne $msg_hdr . $msg) {
        # we appended a blank, don't show it in warning
        my $msg_mod;
        $msg_mod = "$msg_hdr$msg";
        $msg_mod =~ s/ $//;
        push @{$metar{warning}}, { warningType => 'msgModified', s => $msg_mod};
    }

=pod

Then the components of the message are checked starting with the observation
station and the observation time (or issue time for TAF).

=cut

    if ($is_taf && $msg =~ /^(COR|AMD) (.*)/o) {
        $msg = $2;
        $metar{reportModifier}{s} = $1;
        $metar{reportModifier}{modifierType} = $1;
    }

    if ($msg !~ /^([A-Z][A-Z\d]{3}) (.*)/o) {
        $metar{ERROR} = _makeErrorMsg 'obsStation', $corr_msg, $msg;
        return %metar;
    }
    $metar{obsStation}{id} = $1;
    $metar{obsStation}{s} = $1;
    $msg = $2;

    @cy = $metar{obsStation}{id} =~ /(((.).)..)/;

    $tt = $is_taf ? 'issueTime' : 'obsTime';
    if ($msg =~ /^($re_day)($re_hour)($re_min)Z (.*)/o) {
        $msg = $4;
        $metar{$tt}{s} = "$1$2$3Z";
        @{$metar{$tt}}{'day', 'hour', 'minute'} = ($1, $2, $3);
    } elsif ($msg =~ /^($re_day$re_hour${re_min}KT) (.*)/o) {
        $msg = $2;
        $metar{$tt}{s} = $1;
        $metar{$tt}{invalidFormat} = $1;
    } elsif ($is_taf && $msg =~ /^$re_day$re_hour$re_min /o) {
        $metar{issueTime}{s} = '';
        $metar{issueTime}{invalidFormat} = '';
    } elsif ($msg =~ /^(NIL) $/o) {
        # EXTENSION: NIL instead of issueTime
        $metar{reportModifier}{modifierType} = $metar{reportModifier}{s} = $1;
        return %metar;
    } else {
        $metar{ERROR} = _makeErrorMsg 'obsTime', $corr_msg, $msg;
        return %metar;
    }

=pod

For METAR messages the next component may optionally be the report modifier to
indicate a missing message (C<NIL>), a message created by an automated station
(C<AUTO>), or an corrected (C<COR>) or retarded (C<RTD>) message. For Canada,
the component BBB may appear to indicate that the report has been retarded
(C<RR?>), corrected (C<CC?>), amended (C<AA?>), or segmented (C<P??>).

=cut

    # EXTENSION: BBB (for Canada)
    if (!$is_taf && $msg =~ /^(NIL|AUTO|COR|RTD|$re_bbb) (.*)/o) {
        $msg = $2;
        $metar{reportModifier}{s} = $1;
        if ($metar{reportModifier}{s} =~ 'NIL|AUTO|COR|RTD') {
            $metar{reportModifier}{modifierType} = $metar{reportModifier}{s};
        } else {
            $metar{reportModifier}{s} =~ '(.)(.)(.)';
            if ($1 eq 'P') {
                $metar{reportModifier}{modifierType} = $1;
                $metar{reportModifier}{segment} = "$2$3";
                $metar{reportModifier}{isLastSegment} = undef if $2 eq 'Z';
            } else {
                $metar{reportModifier}{modifierType} = "$1$2";
                if ($3 eq 'Z') {
                    $metar{reportModifier}{over24hLate}  = undef;
                } elsif ($3 eq 'Y') {
                    $metar{reportModifier}{sequenceLost} = undef;
                } else {
                    $metar{reportModifier}{bulletinSeq} = $3;
                }
            }
        }
        return %metar if $metar{reportModifier}{modifierType} eq 'NIL';
    }

=pod

For TAF messages there may follow the modifier C<NIL> (which terminates the
message), C<COR> (corrected), or C<AMD> (amended). Then follows the forecast
period and optionally C<CNL> to indicate a cancelled forecast.

=cut

    if ($is_taf) {
        if ($msg =~ /^(NIL|COR|AMD) (.*)/o) {
            $msg = $2;
            $metar{reportModifier}{s} = $1;
            $metar{reportModifier}{modifierType} = $1;
            return %metar if $1 eq 'NIL';
        }

        if ($msg =~ m/^($re_day)($re_hour)($re_hour|24) (.*)/o) {
            $msg = $4;
            $metar{fcstPeriod}{s} = "$1$2$3";
            @{$metar{fcstPeriod}}{'day', 'hourFrom', 'hourTill'} = ($1, $2, $3);
        } else {
            $metar{ERROR} = _makeErrorMsg 'fcstPeriod', $corr_msg, $msg;
            return %metar;
        }

        if ($msg =~ m/^(CNL) (.*)/o) {
            $msg = $2;
            $metar{fcstCancelled}{s} = "$1";
            return %metar;
        }

        # EXTENSION: NIL after fcstPeriod
        if ($msg =~ m/^(NIL) (.*)/o) {
            $msg = $2;
            $metar{reportModifier}{s} = $1;
            $metar{reportModifier}{modifierType} = $1;
            return %metar;
        }

        # EXTENSION: NOT AVBL after fcstPeriod. Remarks may follow
        if ($msg =~ m/^(FCST NOT AVBL DUE NO OBS) (.*)/o) {
            $msg = $2;
            $metar{fcstNotAvbl}{s} = $1;
            $metar{fcstNotAvbl}{fcstNotAvblReason} = 'NOOBS';
        }
        if ($msg =~ m/^(FCST NOT AVBL DUE INSUFFICIENT OBS) (.*)/o) {
            $msg = $2;
            $metar{fcstNotAvbl}{s} = $1;
            $metar{fcstNotAvbl}{fcstNotAvblReason} = 'INSUFFICIENTOBS';
        }
    }

=pod

The first weather group is the surface wind. It may include a gust speed; the
wind speed and/or the wind direction may also be missing. There may be a second
group to report a varying wind direction.

=cut

    # EXTENSION: allow ///// (wind not available)
    # EXTENSION: allow /// for wind direction (not available)
    # EXTENSION: allow // for wind speed (not available)
    if ($msg =~ m@^(/////|(?:$re_wind)) (.*)@o) {
        $msg = $6;
        $metar{sfcWind}{s} = $1;
        $metar{sfcWind}{wind} = _parseWind $1;
    # EXTENSION: allow invalid formats
    } elsif ($msg =~ m@^(${re_wind_dir}0$re_wind_speed|\d{6}$re_wind_speed_unit) (.*)@o)
    {
        $metar{sfcWind}{s} = $1;
        $metar{sfcWind}{wind}{invalidFormat} = $1;
        $msg = $2;
    # EXTENSION: allow missing wind
    } else {
        push @{$metar{warning}}, { warningType => 'windMissing' };
    }

    # EXTENSION:
    # Annex 3 Appendix 3 4.1.4.2.b.1 requires wind speed >=3 kt and
    #   variation 60-180 for this group: not checked
    if ($msg =~ /^($re_wind_dir\d)V($re_wind_dir\d) (.*)/o) {
        if (exists $metar{sfcWind}) {
            $metar{sfcWind}{s} .= ' ';
        } else {
            # TODO?: actually not "not available" but missing
            $metar{sfcWind}{wind}{dirNotAvailable} = undef;
            $metar{sfcWind}{wind}{speedNotAvailable} = undef;
            $metar{sfcWind}{s} = '';
        }
        $metar{sfcWind}{wind}{windVarLeft} = $1 + 0;
        $metar{sfcWind}{wind}{windVarRight} = $2 + 0;
        $metar{sfcWind}{s} .= "$1V$2";
        $msg = $3;
    }

=pod

If the weather conditions allow (visbility more than 10 km, no relevant weather,
no clouds below 5000 ft) this may be indicated by the keyword C<CAVOK>. The next
component after that should be the temperature (see below).

=cut

    # Annex 3 Appendix 3 2.2:
    #   vis >10km, no relevant weather, no cloud below 5000 ft
    if ($msg =~ '^CAVOK (.*)') {
        $metar{CAVOK} = undef; # cancels visibility, weather and cloud
        $msg = $1;
    }

    if (!exists $metar{CAVOK}) {

=pod

Otherwise, the prevailing visibilty follows. It can have a compass direction
attached or C<NDV> if no directional variations can be given. There may be an
additional group for the minimum visibilty.

=cut

        # EXTENSION: allow 16 compass directions
        # EXTENSION: allow 'M' after re_vis_m
        if ($msg =~ /^((?:(P)?($re_vis_m)M?(NDV)?|($re_vis_km))($re_compass_dir16)?)(?: ($re_vis_m)($re_compass_dir))? (.*)/o)
        {
            $msg = $9;
            $metar{visPrev}{s} = $1;
            $metar{visPrev}{compassDir} = $6 if defined $6;
            if ($7) {
                $metar{visMin}{s}          = $7 . $8;
                $metar{visMin}{distance}   = { v => $7 + 0, u => 'M' };
                $metar{visMin}{compassDir} = $8;
            }
            if (defined $3) {
                $metar{visPrev}{distance}    = { v => $3 + 0, u => 'M' };
                $metar{visPrev}{distance}{q} = 'isEqualGreater' if defined $2;
                $metar{visPrev}{NDV}         = undef if defined $4;
            } else {
                $metar{visPrev}{distance} = { v => $5, u => 'KM' };
                $metar{visPrev}{distance}{v} =~ s/KM//;
            }
        } elsif ($msg =~ m@^(${re_vis_sm})SM (.*)@o) {
            $msg = $2;
            $metar{visPrev}{s} = "$1SM";
            _parseFraction $metar{visPrev}, $1, 'SM';
        } elsif ($is_taf && $msg =~ m@^P([1-9]\d*)SM (.*)@o) {
            $msg = $2;
            $metar{visPrev}{s} = "P$1SM";
            _parseFraction $metar{visPrev}, $1, 'SM';
            $metar{visPrev}{distance}{q} = 'isEqualGreater';
        # EXTENSION: allow //// and misformatted entry
        } elsif ($msg =~ m@^//// (.*)@o) {
            $metar{visPrev}{notAvailable} = undef;
            $metar{visPrev}{s} = '////';
            $msg = $1;
        } elsif ($msg =~ m@^(\d{3})(?: ($re_vis_m)($re_compass_dir))? (.*)@o) {
            $metar{visPrev}{invalidFormat} = $1;
            $metar{visPrev}{s} = $1;
            if ($2) {
                $metar{visMin}{s}        = $2 . $3;
                $metar{visMin}{distance} = { v => $2 + 0, u => 'M' };
                $metar{visMin}{compassDir} = $3;
            }
            $msg = $4;
        # EXTENSION: allow invalid formats
        } elsif ($msg =~ m@^(${re_vis_m}U) (.*)@o) {
            $metar{visPrev}{invalidFormat} = $1;
            $metar{visPrev}{s} = $1;
            $msg = $2;
        } else {
            # EXTENSION: allow missing visibility even without CAVOK
            push @{$metar{warning}}, { warningType => 'visibilityMissing' };
        }

=pod

After that may follow one or more runway visibilty ranges or C<RVRNO> if they
are not available.

=cut

        while (_parseRwyVis \%metar, \$msg) {};

        # EXTENSION: allow RVRNO (not available) (KADW, RKSG, PAED)
        while ($msg =~ m@^RVRNO (.*)@o) {
            $metar{RVRNO} = undef;
            $msg = $1;
        }

=pod

Then there can be up to 3 groups to describe the present weather.

=cut

        # EXTENSION: allow // (not available)
        while ($msg =~ m@^($re_weather|//|([+-]?(?:RATS|BRHZ|RESH|RETS|VC|VCRA|RABR))) (.*)@o) {
            if (defined $2) {
                push @{$metar{weather}}, {
                    s             => $2,
                    invalidFormat => $2
                };
            } else {
                push @{$metar{weather}}, _parseWeather $1;
            }
            $msg = $3;
        }

=pod

Then follow up to 3 groups for the sky condition (cloud cover and base or
vertical visibilty) optionally with cloud type. The keywords C<CLR>, C<SKC>, or
C<NSC> may indicate different sky conditions if no cloud cover is given.

=cut

        # EXTENSION: allow ////// (not available) without CB, TCU
        # WMO-No. 306 Vol II
        #   CLR: no clouds below 10000 (FMH-1: 12000) ft detected by autom. st.
        # Annex 3 Appendix 3 4.5.4.1.e, f
        #   SKC: no clouds + VV not restricted but not CAVOK
        #   NSC: no significant clouds + no CB + VV not restr. but not CAVOK,SKC
        # AMOSSG/5-SoD Appendix C 4.9.1.4.b
        #   NCD: no clouds + CB, TCU detected by automatic observation system
        while ($msg =~ m@^((SKC|NSC|CLR|NCD)|VV($re_cloud_base)|(/{6}|$re_cloud_cov(?:$re_cloud_base(?: ?$re_cloud_type)?|$re_cloud_type))|($re_cloud_cov\d{1,2}|$re_cloud_cov(?: \d{1,3})?)) (.*)@o)
        {
            $msg = $6;
            if (defined $5) {
                push @{$metar{cloud}}, {
                    s             => $1,
                    invalidFormat => $1
                };
            } elsif (defined $2) {
                push @{$metar{cloud}}, {
                    s        => $1,
                    noClouds => $2
                };
            } elsif (defined $3) {
                $metar{visVert}{s} = $1;
                if ($3 eq '///') {
                    $metar{visVert}{notAvailable} = undef;
                } else {
                    $metar{visVert}{distance} = { v => $3 * 100, u => 'FT' };
                }
            } else {
                push @{$metar{cloud}}, _parseCloud $4;
            }
        }
        _determineCeiling $metar{cloud} if exists $metar{cloud};
    }

=pod

The following group contains the current air temperature and optionally the dew
point. If both are given the humidity can be determined.

=cut

    # EXTENSION: FMH-1: dew point is optional
    # EXTENSION: allow // for temperature and dew point
    # EXTENSION: allow XX for temperature and dew point
    if (!$is_taf && $msg =~ '^((?:(M)?(\d ?\d)|(?://|XX))/((M)? ?(\d ?\d)|(?://|XX))?) (.*)')
    {
        $msg = $7;
        $metar{temperature}{s} = $1;
        if (defined $3) {
            $metar{temperature}{air}{temp} = { v => $3, u => 'C' };
            $metar{temperature}{air}{temp}{v} =~ tr / //d;
            $metar{temperature}{air}{temp}{v} += 0;
            $metar{temperature}{air}{temp}{v} *= -1 if defined $2;
        } else {
            $metar{temperature}{air}{notAvailable} = undef;
        }
        if (defined $4) {
            if (defined $6) {
                $metar{temperature}{dewpoint}{temp} = { v => $6, u => 'C' };
                $metar{temperature}{dewpoint}{temp}{v} =~ tr / //d;
                $metar{temperature}{dewpoint}{temp}{v} += 0;
                $metar{temperature}{dewpoint}{temp}{v} *= -1 if defined $5;
            } else {
                $metar{temperature}{dewpoint}{notAvailable} = undef;
            }
        }
        if (   exists $metar{temperature}{air}
            && exists $metar{temperature}{air}{temp}
            && exists $metar{temperature}{dewpoint}
            && exists $metar{temperature}{dewpoint}{temp})
        {
            @{$metar{temperature}}{'relHumid1', 'relHumid2',
                                   'relHumid3', 'relHumid4'}
                = _getHumidity $metar{temperature}{air}{temp}{v},
                               $metar{temperature}{dewpoint}{temp}{v};
        }
    } else {
        push @{$metar{warning}}, { warningType => 'tempMissing'} unless $is_taf;
    }

=pod

The last regular item is the QNH given in hectopascal or in. Hg.

=cut

    if ($msg =~ m@^($re_qnh) (.*)@o) {
        $msg = $2;
        $metar{QNH} = _parseQNH $1;
    }
    # EXTENSION: allow Altimeter after QNH
    if (_isValidCountry('QNH_Alt', \@cy) && $msg =~ m@^($re_qnh) (.*)@o) {
        my $qnh;

        $msg = $2;
        $qnh = _parseQNH $1;
        for (keys %$qnh) {
            if ($_ eq 's') {
                $metar{QNH}{s} .= ' '. $qnh->{$_};
            } else {
                $metar{QNH}{$_} .= $qnh->{$_};
            }
        }
    }
    push @{$metar{warning}}, { warningType => 'QNHMissing' }
        unless $is_taf || exists $metar{QNH};

=pod

After that may follow country or station specific information: pressure, worst
cloud cover. There may also be groups indicating recent weather, wind shear,
runway conditions and runway winds.

=cut

    # EXTENSION: allow some (sea level?) pressure (OIxx, OPxx)
    if ($cy[1] =~ /O[IP]/ && $msg =~ m@^([789]\d\d[./]\d) (.*)@o) {
        $msg = $2;
        $metar{somePressure}{s} = $1;
        $metar{somePressure}{hPa} = $1;
        $metar{somePressure}{hPa} =~ s@/@.@;
    }

    # EXTENSION: allow worst cloud cover
    if (   _isValidCountry('cloudMaxCover', \@cy)
        && $msg =~ m@^($re_cloud_cov|SKC) (.*)@o)
    {
        $msg = $2;
        $metar{cloudMaxCover}{s} = $1;
        if ($1 eq 'SKC') {
            $metar{cloudMaxCover}{SKC} = undef;
        } else {
            $metar{cloudMaxCover}{cloudCover} = $1;
        }
    }

    while ($msg =~ m@^RE($re_weather_re) (.*)@o) {
        $msg = $2;
        push @{$metar{recentWeather}}, _parseWeather($1, 1);
    }

    if ($msg =~ '^(WS ALL RWY) (.*)') {
        $msg = $2;
        push @{$metar{windShear}}, {
            s           => $1,
            rwyDesigAll => undef
        };
    }
    while ($msg =~ m@^(WS RWY($re_rwy_des)) (.*)@o) {
        $msg = $3;
        push @{$metar{windShear}}, {
            s        => $1,
            rwyDesig => $2
        };
    }

    while (_parseRwyState \%metar, \$msg) {};

    # EXTENSION: allow colour code
    if ($msg =~ m@^($re_colour) (.*)@o) {
        $msg = $2;
        $metar{colourCode} = _parseColourCode $1;
    }

    # EXTENSION: country specific things
    if (   _isValidCountry('NEFO_PLAYA', \@cy)
        && $msg =~ m@^(NEFO PLAYA (?:([1-9]\d+0)FT|(SKC))) (.*)@o)
    {
        $msg = $4;
        $metar{NEFO_PLAYA}{s} = $1;
        if (defined $2) {
            $metar{NEFO_PLAYA}{cloudBase} = { v => $2 + 0, u => 'FT'};
        } else {
            $metar{NEFO_PLAYA}{SKC} = undef;
        }
    }

    # EXTENSION:
    while (   _isValidCountry('rwyWind', \@cy)
           && $msg =~ m@^($re_rwy_wind) (.*)@o)
    {
        my $r;
        $msg = $7;
        $r->{s} = $1;
        $r->{wind} = _parseWind $4;
        $r->{wind}{windVarLeft} = $5 + 0 if defined $5;
        $r->{wind}{windVarRight} = $6 + 0 if defined $6;
        $r->{rwyDesig} = defined $2 ? $2 : $3;
        push @{$metar{rwyWind}}, $r;
    }

    # EXTENSION:
    while (   _isValidCountry('rsRwyWind', \@cy)
           && $msg =~ m@^($re_rs_rwy_wind) (.*)@o)
    {
        my $r;
        $msg = $5;
        $r->{s} = $1;
        $r->{wind} = _parseWind $3 . '0' . $4;
        if ($2 eq 'RS') {
            $r->{windLocation} = $2;
            $metar{windAtLoc} = $r;
        } else {
            $r->{rwyDesig} = $2;
            push @{$metar{rwyWind}}, $r;
        }
    }

=pod

Finally, there may be trends and remarks.

=cut

    if ($msg =~ '^(NOSIG) (.*)') {
        my $td;

        $td->{s} = $1;
        $td->{trendType} = $1;
        push @{$metar{trend}}, $td;
        $msg = $2;

        # EXTENSION: allow rwyState after NOSIG. Not considered trend data!
        while (_parseRwyState \%metar, \$msg) {};
    }

    # EXTENSION: allow RH after NOSIG. Not considered trend data!
    if (_isValidCountry('relHumid', \@cy) && $msg =~ m@^(RH(\d\d)) (.*)@o) {
        $msg = $3;
        $metar{RH}{s} = $1;
        $metar{RH}{relHumid} = $2 + 0;
    }

    # "supplementary" section of TAFs
    while ($is_taf && _parseTAFsuppl \%metar, \$msg, \%metar) {};

    # additional info in TAFs
    while ($is_taf && _parseTAFinfo \%metar, \$msg) {};

    # changes at midnight: 0000 with FM/AT, 2400 with TL
    # EXTENSION: allow missing period
    # EXTENSION: FM: allow time with Z
    # EXTENSION: FM: allow time without minutes
    # EXTENSION: FM: allow 24Z? or 2400Z?
    while (   (!$is_taf && $msg =~ '^(BECMG|TEMPO) (.*)')
           || ($is_taf && $msg =~ m@^(FM(?:($re_hour)($re_min)?|(24)(?:00)?)Z?|(?:(BECMG|TEMPO|INTER)|(?:PROB([34]0)(?: (TEMPO|INTER))?))(?: ($re_hour|24)($re_hour|24))?) (.*)@o))
    {
        my $td;

        if ($is_taf) {
            $msg = $10;
            $td->{s} = $1;
            if (defined $2) {
                $td->{trendType} = 'BECMG';
                @{$td->{trendTime1}}{'timeSpec', 'hour'} = ('FM', $2);
                $td->{trendTime1}{minute} = defined $3 ? $3 : '00';
            } elsif (defined $4) {
                $td->{trendType} = 'BECMG';
                @{$td->{trendTime1}}{'timeSpec', 'hour'} = ('FM', $4);
                $td->{trendTime1}{minute} = '00';
            } else {
                if (defined $5) {
                    $td->{trendType} = $5;
                } else {
                    $td->{probability} = $6;
                    $td->{trendType} = defined $7 ? $7 : 'PROB';
                }
                if (defined $8) {
                    @{$td->{trendTime1}}{'timeSpec', 'hour', 'minute'}
                                                             = ('FM', $8, '00');
                    @{$td->{trendTime2}}{'timeSpec', 'hour', 'minute'}
                                                             = ('TL', $9, '00');
                } else {
                    push @{$metar{warning}},
                                 { warningType => 'periodMissing', s => $1 };
                }
            }
        } else {
            $msg = $2;
            $td->{s} = $1;
            $td->{trendType} = $1;

            if ($msg =~ m@^((?:TL2400|(?:FM|TL|AT)$re_hour$re_min)Z?) (.*)@o) {
                $msg = $2;
                $td->{s} .= " $1";
                @{$td->{trendTime1}}{'timeSpec', 'hour', 'minute'} =
                                                           $1 =~ /(..)(..)(..)/;

                if ($msg =~ m@^TL(2400Z?|$re_hour${re_min}Z?) (.*)@o) {
                    $td->{s} .= " TL$1";
                    $msg = $2;
                    $td->{trendTime2}{timeSpec} = 'TL';
                    @{$td->{trendTime2}}{'hour', 'minute'} = $1 =~ /(..)(..)/;
                }
            }
        }

        if ($msg =~ m@^($re_wind) (.*)@o) {
            $msg = $6;
            $td->{sfcWind}{s} = $1;
            $td->{sfcWind}{wind} = _parseWind $1;
        # EXTENSION: allow invalid formats
        } elsif ($msg =~ m@^(${re_wind_dir}0$re_wind_speed|\d{6}$re_wind_speed_unit) (.*)@o)
        {
            $td->{sfcWind}{s} = $1;
            $td->{sfcWind}{wind}{invalidFormat} = $1;
            $msg = $2;
        }

        # EXTENSION: allow variation
        if ($msg =~ /^($re_wind_dir\d)V($re_wind_dir\d) (.*)/o) {
            if (exists $td->{sfcWind}) {
                $td->{sfcWind}{s} .= ' ';
            } else {
                # TODO?: actually not "not available" but missing
                $td->{sfcWind}{wind}{dirNotAvailable} = undef;
                $td->{sfcWind}{wind}{speedNotAvailable} = undef;
                $td->{sfcWind}{s} = '';
            }
            $td->{sfcWind}{wind}{windVarLeft} = $1 + 0;
            $td->{sfcWind}{wind}{windVarRight} = $2 + 0;
            $td->{sfcWind}{s} .= "$1V$2";
            $msg = $3;
        }

        if ($msg =~ '^CAVOK (.*)') {
            $td->{CAVOK} = undef;
            $msg = $1;
        }

        if (!exists $td->{CAVOK}) {
            # EXTENSION: allow 'M' after re_vis_m
            if ($msg =~ m@^(($re_vis_m)M?|($re_vis_km)|(${re_vis_sm})SM) (.*)@o)
            {
                $msg = $5;
                $td->{visPrev}{s} = $1;
                if (defined $2) {
                    $td->{visPrev}{distance} = { v => $2 + 0, u => 'M' };
                } elsif (defined $3) {
                    $td->{visPrev}{distance} = { v => $3, u => 'KM' };
                    $td->{visPrev}{distance}{v} =~ s/KM//;
                } else {
                    _parseFraction $td->{visPrev}, $4, 'SM';
                }
            } elsif ($is_taf && $msg =~ m@^P([1-9]\d*)SM (.*)@o) {
                $msg = $2;
                $td->{visPrev}{s} = "P$1SM";
                _parseFraction $td->{visPrev}, $1, 'SM';
                $td->{visPrev}{distance}{q} = 'isEqualGreater';
            } elsif ($msg =~ m@^(\d{3}) (.*)@o) {
                $td->{visPrev}{s} = $1;
                $td->{visPrev}{invalidFormat} = $1;
                $msg = $2;
            }

            # EXTENSION: allow VC..
            while ($msg =~ m@^($re_weather|NSW) (.*)@o) {
                push @{$td->{weather}}, _parseWeather $1;
                $msg = $2;
            }

            $ii = 0;
            while ($msg =~ m@^((SKC|NSC)|VV($re_cloud_base)|($re_cloud_cov$re_cloud_base$re_cloud_type?|$re_cloud_cov$re_cloud_type)) (.*)@o)
            {
                $msg = $5;
                if (defined $2) {
                    push @{$td->{cloud}}, {
                        s        => $1,
                        noClouds => $2
                    };
                } elsif (defined $3) {
                    $td->{visVert}{s} = $1;
                    if ($3 eq '///') {
                        $td->{visVert}{notAvailable} = undef;
                    } else {
                        $td->{visVert}{distance} = { v => $3 * 100, u => 'FT' };
                    }
                } else {
                    push @{$td->{cloud}}, _parseCloud $4;
                }
            }
            _determineCeiling $td->{cloud} if exists $td->{cloud};

            # EXTENSION: allow colour code
            if ($msg =~ m@^($re_colour) (.*)@o) {
                $msg = $2;
                $td->{colourCode} = _parseColourCode $1;
            }
        }

        # EXTENSION: allow rwyState after trend BECMG / TEMPO
        while (_parseRwyState $td, \$msg) {};

        # "supplementary" section of TAFs
        while ($is_taf && _parseTAFsuppl $td, \$msg, \%metar) {};

        # additional info in TAFs
        while ($is_taf && _parseTAFinfo \%metar, \$msg) {};

        push @{$metar{trend}}, $td;
    }

    if ($msg =~ '^RMK[ /] ?(.*)') {
        my ($notRecognised, $parsed);

        @{$metar{remark}} = ();
        $msg = $1;

        $notRecognised = '';
        while ($msg ne '') {
            my $r;

            $parsed = 1;
            $r = {};
            if (   $notRecognised =~ /DUE(?: TO)?(?: $re_phen_desc| GROUND)?$/
                || $notRecognised =~ /BY(?: $re_phen_desc| GROUND)?$/)
            {
                $parsed = 0;
            } elsif (_parseRwyState $r, \$msg) {
                push @{$metar{remark}}, { rwyState => $r->{rwyState}[0] };
            } elsif (_parseRwyVis $r, \$msg) {
                push @{$metar{remark}}, { visRwy => $r->{visRwy}[0] };
            } elsif (   _isValidCountry('remarkCloudMaxCov', \@cy)
                && $msg =~ m@^($re_cloud_cov) (.*)@o)
            {
                $msg = $2;
                push @{$metar{remark}}, { cloudMaxCover => {
                    s          => $1,
                    cloudCover => $1
                }};
            } elsif (   _isValidCountry('remarkColour', \@cy)
                && $msg =~ m@^($re_colour) (.*)@o)
            {
                $msg = $2;
                push @{$metar{remark}}, { colourCode => _parseColourCode $1 };
            } elsif (   _isValidCountry('remarkCloudTypes', \@cy)
                     && $msg =~ m@^(8/([\d/])([\d/])([\d/])) (.*)@o)
            {
                $msg = $5;
                push @{$metar{remark}}, { cloudTypes => {
                    s => $1,
                    $2 eq '/' ? ('cloudTypeLowAboveOvercast', undef)
                              : ('cloudTypeLow', $2),
                    $3 eq '/' ? ('cloudTypeMiddleAboveOvercast', undef)
                              : ('cloudTypeMiddle', $3),
                    $4 eq '/' ? ('cloudTypeHighAboveOvercast', undef)
                              : ('cloudTypeHigh', $4),
                }};
            } elsif ($msg =~ m@^(PWINO|FZRANO|TSNO|PNO|RVRNO|NO ?SPECI|VIA PHONE|RCRNR|HALO(?: VISBL)?|(?:LGT |HVY )?FROIN|SD[FGP]/HD[FGP]|VRBL CONDS|ACFT MSHP)[/ ](.*)@o)
            {
                $msg = $2;
                $r->{s} = $1;
                ($r->{v} = $1) =~ tr/ /_/;
                $r->{v} =~ s/NO_?SPECI/NOSPECI/;
                $r->{v} =~ s/_VISBL$//;
                push @{$metar{remark}}, { keyword => $r };
            } elsif ($msg =~ m@^(?:/) (.*)@o){
                $msg = $1;
                push @{$metar{remark}}, { keyword => {
                    s => '/',
                    v => 'slash'
                }};
            } elsif ($msg =~ m@^(?:\$) (.*)@o){
                $msg = $1;
                push @{$metar{remark}}, { needMaint => { s => '$' }};
            } elsif (   _isValidCountry('remarkCIGRAG', \@cy)
                     && $msg =~ m@^(CIG (?:RAG|RGD)) (.*)@o)
            {
                $msg = $2;
                $r->{s} = $1;
                ($r->{v} = $1) =~ tr/ /_/;
                $r->{v} =~ s/RGD/RAG/;
                push @{$metar{remark}}, { keyword => $r };
            } elsif (   _isValidCountry('remarkCIGVRBL', \@cy)
                     && $msg =~ m@^(CIG VRBL? ($re_vis_sm)-($re_vis_sm)) (.*)@o)
            {
                $msg = $4;
                $r->{s} = $1;
                $r->{visibilityFrom} = {};
                $r->{visibilityTo} = {};
                _parseFraction $r->{visibilityFrom}, $2, 'SM';
                _parseFraction $r->{visibilityTo}, $3, 'SM';
                push @{$metar{remark}}, { ceilVisVariable => $r };
            } elsif (   _isValidCountry('remarkSLPhPa', \@cy)
                     && $msg =~ m@^(SLP ?(?:(\d\d)(\d)|NO|///)) (.*)@o)
            {
                $msg = $4;

                if (!defined $2) {
                    push @{$metar{remark}}, { SLPhPa => {
                        s            => $1,
                        notAvailable => undef
                    }};
                # TODO: really need QNH? but values are REALLY close:
                # KLWM 161454Z AUTO 31012G21KT 10SM CLR M10/M18 A2808 RMK SLP511
                #   -> 951.1 hPa
                # KQHT 021355Z 26003KT 8000 IC M15/M16 A3070 RMK SLP484
                #   -> 1048.4 hPa
                # higher QNHs possible:
                # ETNS 161520Z 11006KT 7000 FEW020 M02/M05 Q1078 WHT WHT
                #   -> SLP78?
                # UNWW 021400Z 00000MPS CAVOK M25/M29 Q1056 NOSIG RMK QFE759
                #   -> SLP56?
                } elsif (!exists $metar{QNH} || !exists $metar{QNH}{inHg}) {
                    if ($2 > 80 || $2 < 20) { # only if within sensible range
                        my $slp;
                        $slp = "$2.$3";
                        # threshold 55 taken from mdsplib
                        $slp += $slp < 55 ? 1000 : 900;
                        push @{$metar{remark}}, { SLPhPa => {
                            s => $1,
                            hPa => sprintf '%.1f', $slp
                        }};
                    } else {
                        push @{$metar{remark}}, { SLPhPa => {
                            s => $1,
                            invalidFormat => "no QNH, x$2.$3 hPa"
                        }};
                    }
                } else {
                    my ($slp, @slp, @qnh, $qnhHPa);
                    my $INHG2HPA = 33.86388640341;

                    @slp = ($1, $2, $3);
                    $qnhHPa = _rnd($metar{QNH}{inHg} * $INHG2HPA, 1);

                    $slp = $qnhHPa;                       # start with given QNH
                    $slp =~ s/..$//;                  # remove 2 trailing digits
                    $slp .= "$slp[1].$slp[2]";              # append given value
                    $slp += 100 if $slp + 50 < $qnhHPa;     # ensure SLP > QNH
                    push @{$metar{remark}}, { SLPhPa => {
                        s => $slp[0],
                        hPa => sprintf '%.1f', $slp
                    }};
                }
            } elsif (   _isValidCountry('remarkSPinHg', \@cy)
                     && $msg =~ m@^(SP([23]\d\.\d{3})) (.*)@o)
            {
                $msg = $3;
                push @{$metar{remark}}, { SLPinHg => {
                    s    => $1,
                    inHg => $2
                }};
            } elsif (   _isValidCountry('remarkSLPinHg', \@cy)
                     && $msg =~ m@^(SLP(?:(\d{3})|NO)) (.*)@o)
            {
                $msg = $3;

                if (!defined $2) {
                    push @{$metar{remark}}, { SLPinHg => {
                        s => $1,
                        notAvailable => undef
                    }};
                } elsif (!exists $metar{QNH} || !exists $metar{QNH}{inHg}) {
                    push @{$metar{remark}}, { SLPinHg => {
                        s => $1,
                        invalidFormat => "no QNH, x$2 in. Hg"
                    }};
                } else {
                    my ($qnhInHg, @slp);

                    @slp = ($1, $2);
                    $qnhInHg = $metar{QNH}{inHg} * 100;
                    $qnhInHg =~ s/...$//;
                    $qnhInHg .= "$slp[1]";
                    $qnhInHg /= 100;
                    $qnhInHg += 10 if $qnhInHg + 5 < $metar{QNH}{inHg};
                    push @{$metar{remark}}, { SLPinHg => {
                        s => $slp[0],
                        inHg => $qnhInHg
                    }};
                }
            } elsif (   _isValidCountry('remarkGRIDWind', \@cy)
                     && $msg =~ m@^GRID(${re_wind_dir}0${re_wind_speed}KT) (.*)@o)
            {
                $msg = $2;
                push @{$metar{remark}}, { gridWind => {
                    s    => "GRID$1",
                    wind => _parseWind $1
                }};
            } elsif (   _isValidCountry('remarkRwyWind', \@cy)
                     && $msg =~ m@^($re_rwy_wind) (.*)@o)
            {
                $msg = $7;
                $r->{rwyWind}{s} = $1;
                $r->{rwyWind}{rwyDesig} = defined $2 ? $2 : $3;
                $r->{rwyWind}{wind} = _parseWind $4;
                $r->{rwyWind}{wind}{windVarLeft} = $5 + 0 if defined $5;
                $r->{rwyWind}{wind}{windVarRight} = $6 + 0 if defined $6;
                push @{$metar{remark}}, $r;
            } elsif (   _isValidCountry('remarkRwyWind2', \@cy)
                     && $msg =~ m@^($re_rwy_wind2) (.*)@o)
            {
                $msg = $4;
                $r->{rwyWind}{s} = $1;
                $r->{rwyWind}{wind} = _parseWind $2;
                $r->{rwyWind}{rwyDesig} = $3;
                push @{$metar{remark}}, $r;
            } elsif (   _isValidCountry('remarkRsc', \@cy)
                     && $msg =~ m@^($re_rsc) (.*)@o)
            {
                $msg = $2;
                $r->{s} = $1;
                for ($1 =~ /(SLR|LSR|PSR|P|SANDED|WET|DRY|IR|WR|\/\/|\d\d)+?/g){
                    if ($_ eq '//') {
                        push @{$r->{rwySfcCondArr}}, { notAvailable => undef };
                    } elsif ($_ =~ /\d/) {
                        push @{$r->{rwySfcCondArr}}, { decelerometer => $_ };
                    } else {
                        push @{$r->{rwySfcCondArr}}, { rwySfc => $_ };
                    }
                }
                push @{$metar{remark}}, { rwySfcCondition => $r };
            } elsif (   _isValidCountry('remarkOpacityPhenom', \@cy)
                     && $msg =~ m@^(((?:$re_opacity_phenom[0-8])+)(?: ($re_cloud_type) (?:(ASOCTD?)|(EMBD)))?) (.*)@o)
            {
                $msg = $6;
                $r->{s} = $1;
                $r->{opacityPhenomArr} = ();
                _parseOpacityPhenom $r, $2;
                $r->{cloudTypeAsoctd} = $3 if defined $4;
                $r->{cloudTypeEmbd}   = $3 if defined $5;
                push @{$metar{remark}}, { opacityPhenom => $r };
            } elsif (   _isValidCountry('remarkOpacityPhenomRev', \@cy)
                     && $msg =~ m@^([0-8])($re_opacity_phenom) (.*)@o)
            {
                $msg = $3;
                $r->{s} = "$1$2";
                $r->{opacityPhenomArr} = ();
                _parseOpacityPhenom $r, "$2$1";
                push @{$metar{remark}}, { opacityPhenom => $r };
            } elsif (   _isValidCountry('remarkCloudOpacityLvl', \@cy)
                     && $msg =~ m@^(([0-8])(?:((?:BL)?SN|FG)|($re_cloud_type))(\d{3})(?: TO)?($re_phen_locdirs)?) (.*)@o)
            {
                $msg = $7;
                $r->{s} = $1;
                $r->{oktas} = $2;
                $r->{weather} = _parseWeather $3 if defined $3;
                $r->{cloudType} = $4 if defined $4;
                $r->{cloudBase} = { v => $5 * 100, u => 'FT' };
                $r->{locationAnd} = _parseLocations $6 if defined $6;
                push @{$metar{remark}}, { cloudOpacityLvl => $r };
            } elsif (   _isValidCountry('remarkCloudCoverVar', \@cy)
                     && $msg =~ m@^(($re_cloud_cov$re_cloud_base?) V ($re_cloud_cov)) (.*)@o)
            {
                $msg = $4;
                $r->{cloudCoverVar} = _parseCloud $2;
                $r->{cloudCoverVar}{s} = $1;
                $r->{cloudCoverVar}{cloudCover2} = $3;
                push @{$metar{remark}}, $r;
            } elsif (   _isValidCountry('remarkCloudCoverLvl', \@cy)
                     && $msg =~ m@^($re_cloud_cov$re_cloud_base) (.*)@o)
            {
                $msg = $2;
                push @{$metar{remark}}, { cloud => _parseCloud $1 };
            } elsif (   _isValidCountry('remarkCloudTypeLvl', \@cy)
                     && $msg =~ m@^($re_cloud_type)($re_cloud_base) (.*)@o)
            {
                $msg = $3;
                push @{$metar{remark}}, { cloudTypeLvl => {
                    s         => "$1$2",
                    cloudType => $1,
                    cloudBase => { v => $2 * 100, u => 'FT' },
                }};
            } elsif (   _isValidCountry('remarkCloudTraces', \@cy)
                     && $msg =~ m@^(TRS?( LWR)?(?: CLD|((?:[/ ]$re_trace_cloud)+))($re_phen_locdirs)?) (.*)@o)
            {
                $msg = $5;
                $r->{s} = $1;
                $r->{isLower} = undef if defined $2;
                if (defined $3) {
                    for ($3 =~ m@[/ ]($re_trace_cloud)@g) {
                        push @{$r->{cloudType}}, $_;
                    }
                } else {
                    $r->{cloudTypeNotAvailable} = undef;
                }
                $r->{locationAnd} = _parseLocations $4 if defined $4;
                push @{$metar{remark}}, { cloudTrace => $r };
            } elsif (   _isValidCountry('remarkCloudTraces', \@cy)
                     && $msg =~ m@^(((?:$re_trace_cloud[/ ])+)TR) (.*)@o)
            {
                $msg = $3;
                $r->{s} = $1;
                for ($2 =~ m@($re_trace_cloud)[/ ]@g) {
                    push @{$r->{cloudType}}, $_;
                }
                push @{$metar{remark}}, { cloudTrace => $r };
            } elsif (   _isValidCountry('remarkReWeather', \@cy)
                     && $msg =~ m@^RE($re_weather_re) (.*)@o)
            {
                $msg = $2;
                push @{$metar{remark}},
                                  { recentWeather => [ _parseWeather($1, 1) ] };
            } elsif (   _isValidCountry('remarkSfcVisM', \@cy)
                     && $msg =~ m@^SFC VIS (($re_vis_m)|($re_vis_km)) (.*)@o)
            {
                $msg = $4;
                $r->{s} = "SFC VIS $1";
                $r->{locationAt} = 'SFC';
                if (defined $2) {
                    $r->{visibility}{distance} = { v => $2 + 0, u => 'M' };
                } else {
                    $r->{visibility}{distance} = { v => $3, u => 'KM' };
                    $r->{visibility}{distance}{v} =~ s/KM//;
                }
                push @{$metar{remark}}, { visibilityAtLoc => $r };
            } elsif (   _isValidCountry('remarkVisSfcTwr', \@cy)
                     && $msg =~ m@^((SFC|TWR) VIS ($re_vis_sm)(?:SM)?) (.*)@o)
            {
                $msg = $4;
                $r->{s} = $1;
                $r->{locationAt} = $2;
                $r->{visibility} = {};
                _parseFraction $r->{visibility}, $3, 'SM';
                push @{$metar{remark}}, { visibilityAtLoc => $r };
            } elsif (   _isValidCountry('remarkVisVarK', \@cy)
                     && $msg =~ m@^(VIS ($re_vis_sm)V($re_vis_sm)) (.*)@o)
            {
                $msg = $4;
                $r->{visVar1}{s} = $1;
                $r->{visVar2} = {};
                _parseFraction $r->{visVar1}, $2, 'SM';
                _parseFraction $r->{visVar2}, $3, 'SM';
                push @{$metar{remark}}, $r;
            } elsif (   _isValidCountry('remarkVisVarCY', \@cy)
                     && $msg =~ m@^(VIS VRB ($re_vis_sm)-($re_vis_sm)) (.*)@o)
            {
                $msg = $4;
                $r->{visVar1}{s} = $1;
                $r->{visVar2} = {};
                _parseFraction $r->{visVar1}, $2, 'SM';
                _parseFraction $r->{visVar2}, $3, 'SM';
                push @{$metar{remark}}, $r;
            } elsif (   _isValidCountry('remarkVisVarCYCZ', \@cy)
                     && $msg =~ m@^(VSBY VRBL (\d\.\d)V(\d\.\d)) (.*)@o)
            {
                $msg = $4;
                $r->{visVar1}{s} = $1;
                $r->{visVar2} = {};
                _parseFraction $r->{visVar1}, $2, 'SM';
                _parseFraction $r->{visVar2}, $3, 'SM';
                push @{$metar{remark}}, $r;
            } elsif (   _isValidCountry('remarkPCPNPastHour', \@cy)
                     && $msg =~ m@^(PCPN (\d+\.\d)MM PAST HR) (.*)@o)
            {
                $msg = $3;
                push @{$metar{remark}}, { precipHourly => {
                    s            => $1,
                    period       => '1h',
                    precipAmount => { v => $2 + 0, u => 'MM' }
                }};
            } elsif (   _isValidCountry('remarkVisLocM', \@cy)
                     && $msg =~ m@^VIS((?:$re_phen_locdirs $re_vis_m_km_remark)(?:(?: AND)?$re_phen_locdirs $re_vis_m_km_remark)*) (.*)@o)
            {
                my @visLocData;

                $msg = $2;
                $r->{visListAtLoc}{s} = "VIS$1";
                $r->{visListAtLoc}{visLocData} = \@visLocData;
                for ($1 =~ m@(?:AND)?($re_phen_locdirs $re_vis_m_km_remark)@og) {
                    my $v;

                    m@($re_phen_locdirs) (?:($re_vis_m)|([1-9]\d{2,3})M|($re_vis_km))@o;
                    $v->{locationAnd} = _parseLocations $1;
                    if (defined $2) {
                        $v->{visibility}{distance} = { v => $2 + 0, u => 'M' };
                    } elsif (defined $3) {
                        $v->{visibility}{distance} = { v => $3 + 0, u => 'M' };
                    } else {
                        $v->{visibility}{distance} = { v => $4, u => 'KM' };
                        $v->{visibility}{distance}{v} =~ s/KM//;
                    }
                    push @visLocData, $v;
                }
                push @{$metar{remark}}, $r;
            } elsif ($msg =~ m@^(VIS($re_phen_locdirs) LWR) (.*)@o) {
                $msg = $3;
                push @{$metar{remark}}, { phenomenonAtLoc => {
                    s           => $1,
                    otherPhenom => 'VIS_LWR',
                    locationAnd => _parseLocations $2
                }};
            } elsif (   _isValidCountry('remarkQNH', \@cy)
                     && $msg =~ m@^(A)([23]\d)(\d\d) (.*)@o)
            {
                $msg = $4;
                push @{$metar{remark}}, { QNH => {
                    s    => "$1$2$3",
                    inHg => "$2.$3"
                }};
            } elsif (   _isValidCountry('remarkQNHMB', \@cy)
                     && $msg =~ m@^(Q([01]\d{3})MB) (.*)@o)
            {
                $msg = $3;
                push @{$metar{remark}}, { QNH => {
                    s => $1,
                    hPa => $2
                }};
            } elsif (   _isValidCountry('remarkQFF', \@cy)
                     && $msg =~ m@^([789]\d\d) (.*)@o)
            {
                $msg = $2;
                push @{$metar{remark}}, { QFF => {
                    s => $1,
                    hPa => $1
                }};
            } elsif ($msg =~ m@^(QNH([01]\d{3}\.\d)) (.*)@o) {
                $msg = $3;
                push @{$metar{remark}}, { QNH => {
                    s => $1,
                    hPa => $2
                }};
            } elsif (   _isValidCountry('remarkCorrected', \@cy)
                     && $msg =~ m@^(COR ($re_hour)($re_min)) (.*)@o)
            {
                $msg = $4;
                push @{$metar{remark}}, { correctedAt => {
                    s      => $1,
                    hour   => $2,
                    minute => $3
                }};
            } elsif (   _isValidCountry('remarkSnowIncr', \@cy)
                     && $msg =~ m@^(SNINCR (\d+)/(\d+)) (.*)@o)
            {
                $msg = $4;
                push @{$metar{remark}}, { snowIncr => {
                    s        => $1,
                    pastHour => $2,
                    onGround => $3
                }};
            } elsif (   _isValidCountry('remarkHourly', \@cy)
                     && $msg =~ m@^1([01]\d{3}) (.*)@o)
            {
                $msg = $2;
                push @{$metar{remark}}, { tempMaxPeriod => {
                    s      => "1$1",
                    period => '6h',
                    temp   => _parseTemp($1)
                }};
            } elsif (   _isValidCountry('remarkHourly', \@cy)
                     && $msg =~ m@^2([01]\d{3}) (.*)@o)
            {
                $msg = $2;
                push @{$metar{remark}}, { tempMinPeriod => {
                    s     => "2$1",
                    period => '6h',
                    temp  => _parseTemp($1)
                }};
            } elsif (   _isValidCountry('remarkHourly', \@cy)
                     && $msg =~ m@^(5(?:([0-8])(\d{3})|////)) (.*)@o)
            {
                $msg = $4;
                push @{$metar{remark}}, { pressureChange => {
                    s      => $1,
                    period => '3h',
                    defined $2 ? ('pressureTendency', $2,
                                  'pressureChangeVal', {
                                              v => $3 / 10 * ($2 >= 5 ? -1 : 1),
                                              u => 'hPa'
                                 })
                               : ('notAvailable', undef)
                }};
            } elsif (   _isValidCountry('remarkPeakWind', \@cy)
                     && $msg =~ m@^(PK WND ($re_wind_dir\d$re_wind_speed)/(?:($re_hour)?($re_min))) (.*)@o)
            {
                $msg = $5;
                push @{$metar{remark}}, { peakWind => {
                    s      => $1,
                    wind   => _parseWind("$2KT"),
                    defined $3 ? ('hour', $3) : (),
                    minute => $4
                }};
            } elsif (   $cy[0] eq 'KEHA'
                     && $msg =~ m@^(PK WND ($re_wind_speed) 000) (.*)@o)
            {
                $msg = $3;
                push @{$metar{remark}}, { peakWind => {
                    s    => $1,
                    wind => _parseWind("///$2KT")
                }};
            } elsif (   $cy[0] eq 'EDLP'
                     && $msg =~ m@^RWY($re_rwy_des) (TRL ?(\d\d)) ([A-Z]{2}) ATIS ([A-Z]) (.*)@o)
            {
                $msg = $6;
                push @{$metar{remark}}, { activeRwy => {
                    s        => "RWY$1",
                    rwyDesig => $1
                }};
                push @{$metar{remark}}, { transitionLvl => {
                    s     => $2,
                    level => $3 + 0
                }};
                push @{$metar{remark}}, { obsInitials => { s => $4 }};
                push @{$metar{remark}}, { currentATIS => {
                    s    => "ATIS $5",
                    ATIS => $5
                }};
            } elsif (   _isValidCountry('remarkHourlyTemp', \@cy)
                     && $msg =~ m@^(T([01]\d{3})([01]\d{3})?) (.*)@o)
            {
                $msg = $4;
                $r->{s} = $1;
                $r->{air}{temp} = _parseTemp $2;
                $r->{dewpoint}{temp} = _parseTemp $3
                    if defined $3;
                push @{$metar{remark}}, { tempHourly => $r };
            } elsif (   _isValidCountry('remarkSunshine', \@cy)
                     && $msg =~ m@^(98(?:(\d{3})|///)) (.*)@o)
            {
                $msg = $3;
                $r->{s} = $1;
                $r->{sunshinePeriod} = '1d';
                if (defined $2) {
                    $r->{sunshine} = { v => $2 + 0, u => 'min' };
                } else {
                    $r->{sunshineNotAvailable} = undef;
                }
                push @{$metar{remark}}, { radiationSun => $r };
            } elsif (   _isValidCountry('remarkQFEhPa', \@cy)
                     && $msg =~ m@^(QFE([01]?\d{3})) (.*)@o)
            {
                $msg = $3;
                push @{$metar{remark}}, { QFE => {
                    s => $1,
                    hPa => $2 + 0
                }};
            } elsif (   _isValidCountry('remarkDA_PA', \@cy)
                     && $msg =~ m@^(/?(DA|PA)[ /]?([+-]?\d+)(?:FT)?/?) (.*)@o)
            {
                $msg = $4;
                push @{$metar{remark}}, { ({ DA => 'densityAlt',
                                             PA => 'pressureAlt'
                                           }->{$2})
                                          => {
                    s        => $1,
                    altitude => { v => $3 + 0, u => 'FT' }
                }};
            } elsif (   _isValidCountry('remarkKNIP', \@cy)
                     && $msg =~ m@^((RH|SST|AI|OAT)/(\d\d)F?) (.*)@o)
            {
                $msg = $4;
                $r->{s} = $1;
                if ($2 eq 'SST' || $2 eq 'OAT') {
                    $r->{temp} = { v => $3 + 0, u => 'F' };
                } elsif ($2 eq 'RH') {
                    $r->{relHumid} = $3 + 0;
                } else {
                    $r->{"$2Val"} = $3 + 0;
                }
                push @{$metar{remark}}, { $2 => $r };
            } elsif (   _isValidCountry('remarkKN', \@cy)
                     && $msg =~ m@^((RH|SST)/(\d+)) (.*)@o)
            {
                $msg = $4;
                $r->{s} = $1;
                if ($2 eq 'SST') {
                    $r->{temp} = { v => $3 + 0, u => 'C' };
                } else {
                    $r->{relHumid} = $3 + 0;
                }
                push @{$metar{remark}}, { $2 => $r };
            } elsif (   _isValidCountry('remarkQFEmmHg', \@cy)
                     && $msg =~ m@^(QFE ?(\d{3})(?:,(\d))?(?:/(\d{3,4}))?) (.*)@o)
            {
                my $qfe;
                $msg = $5;
                $qfe->{s} = $1;
                $qfe->{mmHg} = $2 + 0;
                $qfe->{mmHg} += $3 / 10 if defined $3;
                $qfe->{hPa} = $4 + 0 if defined $4;
                push @{$metar{remark}}, { QFE => $qfe };
            } elsif ($msg =~ m@^(((?:VIS|CHI)NO) RWY ?($re_rwy_des)) (.*)@o) {
                $msg = $4;
                push @{$metar{remark}}, { $2 => {
                    s => $1,
                    rwyDesig => $3
                }};
            } elsif (   _isValidCountry('remarkSNWCVR', \@cy)
                     && $msg =~ m@^($re_snw_cvr) (.*)@o)
            {
                $msg = $3;
                $r->{s} = $1;
                $r->{snowCoverType} = defined $2 ? $2 : 'NIL';
                $r->{snowCoverType} =~ s/ONE /ONE_/;
                $r->{snowCoverType} =~ s/MU?CH /MUCH_/;
                $r->{snowCoverType} =~ s/TR(?:ACE ?)? ?/TRACE_/;
                $r->{snowCoverType} =~ s/MED(?:IUM)?/MEDIUM/;
                $r->{snowCoverType} =~ s/ PACK(?:ED)?/_PACKED/;
                push @{$metar{remark}}, { snowCover => $r };
            } elsif (   _isValidCountry('remarkObsTimeOffset', \@cy)
                     && $msg =~ m@^(OBS TAKEN [+](\d+)) (.*)@o)
            {
                $msg = $3;
                push @{$metar{remark}}, { obsTimeOffset => {
                    s       => $1,
                    minutes => $2
                }};
            } elsif (   _isValidCountry('remarkBalloon', \@cy)
                     && $msg =~ m@^((?:BLN (?:DSA?PRD (\d+) ?FT[.]?|VI?SBL(?: TO)? (\d+) ?FT))) (.*)@o)
            {
                $msg = $4;
                $r->{s} = $1;
                if (defined $2) {
                    $r->{disappearedAt} = { distance => { v => $2, u => 'FT' }};
                } else {
                    $r->{visibleTo} = { distance => { v => $3, u => 'FT' }};
                }
                push @{$metar{remark}}, { balloon => $r };
            } elsif (   _isValidCountry('remarkRVR', \@cy)
                     && $msg =~ m@^RVR((?: RWY ?$re_rwy_des $re_rwy_vis ?FT[.]?)+) (.*)@o)
            {
                my $hdr = 'RVR ';
                $msg = $2;
                for ($1 =~ / (RWY ?$re_rwy_des $re_rwy_vis ?FT[.]?)/g) {
                    my $v;

                    $v->{s} = "$hdr$_";
                    $hdr    = '';

                    /RWY ?($re_rwy_des) ($re_rwy_vis) ?(FT)/;

                    $v->{rwyDesig}      = $1;
                    $v->{RVR}{distance} = { v => $2, u => $3 };
                    $v->{RVR}{distance}{q} = 'isLess'
                        if $v->{RVR}{distance}{v} =~ s/^M//;
                    $v->{RVR}{distance}{q} = 'isEqualGreater'
                        if $v->{RVR}{distance}{v} =~ s/^P//;
                    $v->{RVR}{distance}{v} += 0;
                    push @{$metar{remark}}, { visRwy => $v };
                }
            } elsif (   _isValidCountry('remarkTX', \@cy)
                     && $msg =~ m@^(TX/(\d\d[,.]\d)) (.*)@o)
            {
                $msg = $3;
                $r->{tempMaxFQ}{s} = $1;
                $r->{tempMaxFQ}{temp} = { v => $2, u => 'C' };
                $r->{tempMaxFQ}{temp}{v} =~ s/,/./;
                push @{$metar{remark}}, $r;

            } elsif (   _isValidCountry('remarkTAFNxtFcstBy', \@cy)
                     && $msg =~ m@^(NXT FCST BY ($re_hour)Z) (.*)@o)
            {
                $msg = $3;
                push @{$metar{remark}}, { nextFcstBy => {
                    s      => $1,
                    hour   => $2,
                    minute => '00'
                }};
            } elsif (   _isValidCountry('remarkTAFNxtFcstAt', \@cy)
                     && $msg =~ m@^(NXT FCST WILL BE ISSUED AT ($re_day)($re_hour)($re_min)Z) (.*)@o)
            {
                $msg = $5;
                push @{$metar{remark}}, { nextFcstAt => {
                    s      => $1,
                    day    => $2,
                    hour   => $3,
                    minute => $4
                }};
            } elsif (   _isValidCountry('remarkTAFAmdAt', \@cy)
                     && $msg =~ m@^(AMD AT ($re_day)($re_hour)($re_min)Z) (.*)@o)
            {
                $msg = $5;
                push @{$metar{remark}}, { amdAt => {
                    s      => $1,
                    day    => $2,
                    hour   => $3,
                    minute => $4
                }};
            } elsif (   _isValidCountry('remarkTAFFcstAutoObs', \@cy)
                     && $msg =~ m@^(FCST BASED ON AUTO OBS(?: ($re_hour)-($re_hour)Z)?[.]?) (.*)@o)
            {
                $msg = $4;
                push @{$metar{remark}}, { fcstAutoObs => {
                    s => $1,
                    defined $2 ? ('hourFrom', $2,
                                  'hourTill', $3)
                               : ()
                }};
            } elsif (   _isValidCountry('remarkTAFFcstAutometar', \@cy)
                     && $msg =~ m@^(BASED ON AUTOMETAR) (.*)@o)
            {
                $msg = $2;
                push @{$metar{remark}}, { fcstAutoMETAR => { s => $1 }};
            } elsif (   _isValidCountry('remarkATIS', \@cy)
                     && $msg =~ m@^(ATIS ([A-Z])) (.*)@o)
            {
                $msg = $3;
                push @{$metar{remark}}, { currentATIS => {
                    s    => $1,
                    ATIS => $2
                }};
            } elsif ($msg =~ m@^(((?:$re_data_estmd[ /])+)(?:VISUALLY )?$re_estmd) (.*)@o)
            {
                $msg = $3;
                $r->{s} = $1;
                for ($2 =~ /$re_data_estmd/og) {
                    s/WI?NDS?(?: DATA)?/WND/;
                    s/CEILING/CIG/;
                    s/(?:CIG )?BLN/CIG BLN/;
                    s/ /_/g;
                    push @{$r->{estimatedItem}}, $_;
                }
                push @{$metar{remark}}, { estimated => $r };
            } elsif ($msg =~ m@^($re_estmd((?:[ /]$re_data_estmd)+)) (.*)@o) {
                $msg = $3;
                $r->{s} = $1;
                for ($2 =~ /$re_data_estmd/og) {
                    s/WI?NDS?(?: DATA)?/WND/;
                    s/CEILING/CIG/;
                    s/(?:CIG )?BLN/CIG BLN/;
                    s/ /_/g;
                    push @{$r->{estimatedItem}}, $_;
                }
                push @{$metar{remark}}, { estimated => $r };
            } elsif ($msg =~ m@^((?:(THN) )?($re_cloud_cov) ABV ($re_cloud_base)) (.*)@o)
            {
                $msg = $5;
                $r->{cloudAbove}{s} = $1;
                $r->{cloudAbove}{isThin} = undef if defined $2;
                $r->{cloudAbove}{cloudCover} = $3;
                $r->{cloudAbove}{cloudBase}  = { v => $4 * 100, u => 'FT' };
                push @{$metar{remark}}, $r;
            } elsif (   _isValidCountry('remarkPhenCloud', \@cy)
                     && $msg =~ m@^((FU|(?:FZ)?FG|BR|PWR PLNT(?: PLUME)?) ($re_cloud_cov$re_cloud_base)) (.*)@o)
            {
                my ($cl, $phen);

                $msg = $4;
                $r->{s} = $1;
                $r->{cloud} = _parseCloud $3;
                $phen = $2;
                if ($phen =~ m@^$re_weather$@) {
                    $r->{weather} = _parseWeather $phen;
                } else {
                    ($r->{cloudPhenom} = $phen) =~ s/ /_/g;
                }
                push @{$metar{remark}}, { obscuration => $r };
            } elsif (   _isValidCountry('remarkRainfall', \@cy)
                     && $msg =~ m@^(RF(\d\d\.\d)/(\d{3}\.\d)) (.*)@o)
            {
                $msg = $4;
                push @{$metar{remark}}, { rainfall => {
                    s             => $1,
                    rainfall10min => $2 + 0,
                    rainfall0900  => $3 + 0
                }};
            } elsif (   _isValidCountry('remarkRSNK', \@cy)
                     && $msg =~ m@^(RSNK (M)?(\d\d(?:\.\d)?)/(${re_wind_dir}0$re_wind_speed(?:G$re_wind_speed)?)($re_wind_speed_unit)?) (.*)@o)
            {
                my ($temp, $unit);

                $msg = $6;
                $temp = $3 + 0;
                $temp *= -1 if defined $2;
                $unit = defined $5 ? $5 : 'KT';
                push @{$metar{remark}}, { RSNK => {
                    s    => $1,
                    air  => { temp => { v => $temp, u => 'C' } },
                    wind => _parseWind "$4$unit"
                }};
            } elsif (   _isValidCountry('remarkLAGPK', \@cy)
                     && $msg =~ m@^(LAG ?PK (M)?(\d\d)/(M)?(\d\d)/(${re_wind_dir}0$re_wind_speed(?:G$re_wind_speed)?)) (.*)@o)
            {
                my ($temp, $dewpoint);

                $msg = $7;
                $temp = $3 + 0;
                $temp *= -1 if defined $2;
                $dewpoint = $5 + 0;
                $dewpoint *= -1 if defined $4;
                push @{$metar{remark}}, { LAG_PK => {
                    s        => $1,
                    air      => { temp => { v => $temp,     u => 'C' } },
                    dewpoint => { temp => { v => $dewpoint, u => 'C' } },
                    wind     => _parseWind "$6KT"
                }};
            } elsif (   _isValidCountry('remarkPrecipHourly3', \@cy)
                     && $msg =~ m@^(P(\d{3})) (.*)@o)
            {
                $msg = $3;
                $r->{s} = $1;
                $r->{precipAmount} = { v => $2 / 100, u => 'IN' };
                $r->{period}       = '1h';
                push @{$metar{remark}}, { precipHourly => $r };

# nearly the same pattern twice except for the position of $re_phen_desc
# (do not match too much but don't miss anything, either)

            } elsif ($msg =~ m@^(((?:$re_phen_desc[/ ])+)$re_phenomenon4( BBLO)?(?:$re_loc_quadr3$re_wx_mov_d2?|$re_loc_quadr3?$re_wx_mov_d2)(?: ($re_cloud_type) (?:(ASOCTD?)|(EMBD)))?) (.*)@o)
            {
                $msg = $21;
                $r->{s} = $1;
                _parsePhenomDescr $r, 'phenomDescrPre', $2;
                if (defined $3) {
                    ($r->{otherPhenom} = $3) =~ tr/ /_/;
                    $r->{otherPhenom} =~ s/REDUCED/RDCD/;
                    $r->{otherPhenom} =~ s/AURORA/AURBO/;
                    $r->{otherPhenom} =~ s/SHWR/SH/;
                } elsif (defined $4) {
                    @{$r->{cloudType}} = split m@[/-]@, $4;
                } elsif (defined $5) {
                    $r->{weather} = _parseWeather $5;
                } else {
                    $r->{cloudCover} = $6;
                }
                _parsePhenomDescr $r, 'phenomDescrPost', $7 if defined $7;
                $r->{locationAnd} = _parseLocations $8 if defined $8;
                $r->{locationAnd} = _parseQuadrants $10, $9 if defined $10;
                $r->{$11}{locationAnd} = _parseLocations $12 if defined $12;
                $r->{locationAnd} = _parseLocations $13 if defined $13;
                $r->{locationAnd} = _parseQuadrants $15, $14 if defined $15;
                $r->{$16}{locationAnd} = _parseLocations $17 if defined $17;
                $r->{cloudTypeAsoctd} = $18 if defined $19;
                $r->{cloudTypeEmbd}   = $18 if defined $20;
                push @{$metar{remark}}, { phenomenonAtLoc => $r };
            } elsif ($msg =~ m@^($re_phenomenon4(?: IS)?((?:[/ ](?:$re_phen_desc|BBLO))*)(?:$re_loc_quadr3$re_wx_mov_d2?|$re_loc_quadr3?$re_wx_mov_d2)(?: ($re_cloud_type) (?:(ASOCTD?)|(EMBD)))?) (.*)@o)
            {
                $msg = $20;
                $r->{s} = $1;
                if (defined $2) {
                    if ($2 eq 'VIS HIER') {
                        $r->{otherPhenom} = 'VIS_HYR';
                    } else {
                        ($r->{otherPhenom} = $2) =~ tr/ /_/;
                        $r->{otherPhenom} =~ s/REDUCED/RDCD/;
                        $r->{otherPhenom} =~ s/AURORA/AURBO/;
                        $r->{otherPhenom} =~ s/SHWR/SH/;
                    }
                } elsif (defined $3) {
                    @{$r->{cloudType}} = split m@[/-]@, $3;
                } elsif (defined $4) {
                    $r->{weather} = _parseWeather $4;
                } else {
                    $r->{cloudCover} = $5;
                }
                _parsePhenomDescr $r, 'phenomDescrPost', $6 if defined $6;
                $r->{locationAnd} = _parseLocations $7 if defined $7;
                $r->{locationAnd} = _parseQuadrants $9, $8 if defined $9;
                $r->{$10}{locationAnd} = _parseLocations $11 if defined $11;
                $r->{locationAnd} = _parseLocations $12 if defined $12;
                $r->{locationAnd} = _parseQuadrants $14, $13 if defined $14;
                $r->{$15}{locationAnd} = _parseLocations $16 if defined $16;
                $r->{cloudTypeAsoctd} = $17 if defined $18;
                $r->{cloudTypeEmbd}   = $17 if defined $19;
                push @{$metar{remark}}, { phenomenonAtLoc => $r };
            } elsif ($msg =~ m@^((?:PR(?:ESENT )?WX )?((?:$re_phen_desc[/ ])+)$re_phenomenon4( BBLO)?(?: ($re_cloud_type) (?:(ASOCTD?)|(EMBD)))?) (.*)@o)
            {
                $msg = $11;
                $r->{s} = $1;
                _parsePhenomDescr $r, 'phenomDescrPre', $2 if defined $2;
                if (defined $3) {
                    ($r->{otherPhenom} = $3) =~ tr/ /_/;
                    $r->{otherPhenom} =~ s/REDUCED/RDCD/;
                    $r->{otherPhenom} =~ s/AURORA/AURBO/;
                    $r->{otherPhenom} =~ s/SHWR/SH/;
                } elsif (defined $4) {
                    @{$r->{cloudType}} = split m@[/-]@, $4;
                } elsif (defined $5) {
                    $r->{weather} = _parseWeather $5;
                } else {
                    $r->{cloudCover} = $6;
                }
                _parsePhenomDescr $r, 'phenomDescrPost', $7 if defined $7;
                $r->{cloudTypeAsoctd} = $8 if defined $9;
                $r->{cloudTypeEmbd}   = $8 if defined $10;
                push @{$metar{remark}}, { phenomenonOnly => $r };

            # first check a restricted set of phenomenon and descriptions!

            } elsif ($msg =~ m@^(GR ($re_gs_size)) (.*)@o) {
                $msg = $3;
                _parseFraction $r, $2, '';
                push @{$metar{remark}}, { hailStones => {
                    s => $1,
                    hailStoneSize => {
                        v => $r->{distance}{v},
                        u => 'IN',
                        exists $r->{distance}{q} ? ('q', $r->{distance}{q}) : ()
                }}};
            } elsif ($msg =~ m@^((PRES[FR]R)( PAST HR)?) (.*)@o) {
                $msg = $4;
                $r->{s} = $1;
                $r->{otherPhenom} = $2;
                _parsePhenomDescr $r, 'phenomDescrPost', $3 if defined $3;
                push @{$metar{remark}}, { phenomenonOnly => $r };
            } elsif ($msg =~ m@^(($re_phenomenon_other)(?: IS)?((?:[/ ](?:$re_phen_desc_other|BBLO))*)(?: ($re_cloud_type) (?:(ASOCTD?)|(EMBD)))?) (.*)@o)
            {
                $msg = $7;
                $r->{s} = $1;
                ($r->{otherPhenom} = $2) =~ tr/ /_/;
                $r->{otherPhenom} =~ s/REDUCED/RDCD/;
                $r->{otherPhenom} =~ s/AURORA/AURBO/;
                $r->{otherPhenom} =~ s/SHWR/SH/;
                _parsePhenomDescr $r, 'phenomDescrPost', $3 if defined $3;
                $r->{cloudTypeAsoctd} = $4 if defined $5;
                $r->{cloudTypeEmbd}   = $4 if defined $6;
                push @{$metar{remark}}, { phenomenonOnly => $r };
            } elsif ($msg =~ m@^((?:PR(?:ESENT )?WX )?$re_phenomenon4((?: IS)?(?:[/ ](?:$re_phen_desc|BBLO))*)(?: ($re_cloud_type) (?:(ASOCTD?)|(EMBD)))?) (.*)@o)
            {
                $msg = $10;
                $r->{s} = $1;
                if (defined $2) {
                    ($r->{otherPhenom} = $2) =~ tr/ /_/;
                    $r->{otherPhenom} =~ s/REDUCED/RDCD/;
                    $r->{otherPhenom} =~ s/AURORA/AURBO/;
                    $r->{otherPhenom} =~ s/SHWR/SH/;
                } elsif (defined $3) {
                    @{$r->{cloudType}} = split m@[/-]@, $3;
                } elsif (defined $4) {
                    $r->{weather} = _parseWeather $4;
                } else {
                    $r->{cloudCover} = $5;
                }
                _parsePhenomDescr $r, 'phenomDescrPost', $6 if defined $6;
                $r->{cloudTypeAsoctd} = $7 if defined $8;
                $r->{cloudTypeEmbd}   = $7 if defined $9;
                push @{$metar{remark}}, { phenomenonOnly => $r };
            } elsif ($msg =~ m@^(((?:$re_phen_desc[/ ])*)LTG($re_ltg_types+)(?: TO)?($re_phen_locdirs)?$re_wx_mov_d2?(?: ($re_cloud_type) (?:(ASOCTD?)|(EMBD)))?) (.*)@o)
            {
                $msg = $10;
                $r->{s} = $1;
                _parsePhenomDescr $r, 'phenomDescrPre', $2 if defined $2;
                $r->{locationAnd} = _parseLocations $4 if defined $4;
                $r->{$5}{locationAnd} = _parseLocations $6 if defined $6;
                $r->{cloudTypeAsoctd} = $7 if defined $8;
                $r->{cloudTypeEmbd}   = $7 if defined $9;
                $r->{lightningType} = ();
                for ($3 =~ /$re_ltg_types/og) {
                    push @{$r->{lightningType}}, $_;
                }
                push @{$metar{remark}}, { phenomenonAtLoc => $r };
            } elsif ($msg =~ m@^(AO(?:1|2A?)) (.*)@o) {
                $msg = $2;
                push @{$metar{remark}}, { obsStationType => {
                    s           => $1,
                    stationType => $1
                }};
            } elsif ($msg =~ m@^(CIG ($re_cloud_base)(?:(?:(?: (APCH))?(?: RWY ?| R?)?($re_rwy_des))(?: TO)?($re_phen_locdirs)?|(?: TO)?($re_phen_locdirs))) (.*)@o)
            {
                $msg = $7;
                $r->{s} = $1;
                $r->{cloudBase}  = { v => $2 * 100, u => 'FT' };
                $r->{isApproach} = undef if defined $3;
                $r->{rwyDesig} = $4 if defined $4;
                $r->{locationAnd} =_parseLocations $5 if defined $5;
                $r->{locationAnd} =_parseLocations $6 if defined $6;
                push @{$metar{remark}}, { ceilingAtLoc => $r };
            } elsif ($msg =~ m@^(VIS ($re_vis_sm)(?: (APCH))?(?: RWY ?| R?)?($re_rwy_des)) (.*)@o)
            {
                $msg = $5;
                $r->{s} = $1;
                $r->{visibility} = {};
                _parseFraction $r->{visibility}, $2, 'SM';
                $r->{isApproach} = undef if defined $3;
                $r->{rwyDesig} = $4;
                push @{$metar{remark}}, { visibilityAtLoc => $r };
            } elsif ($msg =~ m@^(NOSIG|CAVU|$re_estmd PASS (?:OPEN|CLOSED|CLSD|MARGINAL|MRGL)|PASS $re_estmd CLSD|EPC|EPO|EPM|RTS) (.*)@o) {
                $msg = $2;
                $r->{s} = $1;
                $r->{v} = $1;
                if ($r->{v} =~ m@PASS OP@o) {
                    $r->{v} = 'EPO';
                } elsif ($r->{v} =~ m@PASS .*CL@o) {
                    $r->{v} = 'EPC';
                } elsif ($r->{v} =~ m@PASS M@o) {
                    $r->{v} = 'EPM';
                }
                push @{$metar{remark}}, { keyword => $r };
            } elsif ($msg =~ m@^((?:CLIMAT ?)?($re_temp)/($re_temp)(?:/(TR|$re_precip)(?:/(NIL|$re_precip))?)?) (.*)@o)
            {
                my ($precip1, $precip2);

                $msg = $6;
                $r->{s} = $1;
                $r->{temp1}{temp} = { v => $2 + 0, u => 'C' };
                $r->{temp2}{temp} = { v => $3 + 0, u => 'C' };
                if (defined $4) {
                    if ($4 eq 'TR') {
                        $r->{precipAmountTraces} = undef;
                    } else {
                        $precip1 = $4;
                    }
                    if (defined $5) {
                        if ($5 eq 'NIL') {
                            $r->{precipAmount2MM} = 0;
                        } else {
                            $precip2 = $5;
                            if ($precip2 =~ s/ ?([MC]M)//) {
                                $r->{precipAmount2MM} = $precip2 + 0;
                                $r->{precipAmount2MM} *= 10 if $1 eq 'CM';
                            } else {
                                $r->{precipAmount2Inch} = $precip2 + 0;
                            }
                        }
                    }
                    if (defined $precip1) {
                        if ($precip1 =~ s/ ?([MC]M)//) {
                            $r->{precipAmount1MM} = $precip1 + 0;
                            $r->{precipAmount1MM} *= 10 if $1 eq 'CM';
                        } else {
                            $r->{precipAmount1Inch} = $precip1 + 0;
                        }
                    }
                }
                push @{$metar{remark}}, { climate => $r };
            } elsif ($msg =~ m@^((TORNADO|FUNNEL CLOUDS?|WATERSPOUT) ($re_be_prec_be+)(?: TO)?($re_phen_locdirs)$re_wx_mov_d2?) (.*)@o) {
                $msg = $7;
                $r->{s} = $1;
                $r->{tornadicActivityType} = lc $2;
                $r->{locationAnd} = _parseLocations $4;
                $r->{$5}{locationAnd} = _parseLocations $6 if defined $6;
                $r->{startEndArr} = ();
                for ($3 =~ /$re_be_prec_be+?/og) {
                    my %s_e;
                    /(.)(..)?(..)/;
                    $s_e{hour} = $2 if defined $2;
                    $s_e{minute} = $3;
                    if ($1 eq 'B') {
                        push @{$r->{startEndArr}}, { startTime => \%s_e };
                    } else {
                        push @{$r->{startEndArr}}, { endTime => \%s_e };
                    }
                }
                $r->{tornadicActivityType} =~ s/(CLOUD)S?/$1/;
                $r->{tornadicActivityType} =~ s/ /_/;
                push @{$metar{remark}}, { tornadicActivity => $r };
            } elsif ($msg =~ m@^((?:FIRST|FST)(?:( STAFFED| STFD)|( MANNED))?(?: OBS?)?)[ /](.*)@o)
            {
                $msg = $4;
                $r->{s} = $1;
                $r->{isStaffed} = undef if defined $2;
                $r->{isManned} = undef if defined $3;
                push @{$metar{remark}}, { firstObs => $r };
            } elsif ($msg =~ m@^(NEXT ($re_day)($re_hour)($re_min)(?: ?UTC| ?Z)?) (.*)@o)
            {
                $msg = $5;
                $r->{s} = $1;
                @{$r->{obsAt}}{'day', 'hour', 'minute'} = ($2, $3, $4)
                    if defined $2;
                push @{$metar{remark}}, { nextObs => $r };
            } elsif ($msg =~ m@^(LAST(?:( STAFFED)|( MANNED))?(?: OBS?)?(?: ($re_day)($re_hour)($re_min)(?: ?UTC| ?Z)?)?)[ /](.*)@o)
            {
                $msg = $7;
                $r->{s} = $1;
                $r->{isStaffed} = undef if defined $2;
                $r->{isManned} = undef if defined $3;
                @{$r->{obsAt}}{'day', 'hour', 'minute'} = ($4, $5, $6)
                    if defined $4;
                push @{$metar{remark}}, { lastObs => $r };
            } elsif ($msg =~ m@^((QBB|QBJ)(\d\d0)) (.*)@o) {
                $msg = $4;
                push @{$metar{remark}}, { $2 => {
                    s => $1,
                    altitude => { v => $3 + 0, u => 'M' }
                }};
            # from mdsplib: RADAT: elevation information
            } elsif ($msg =~ m@^(RADAT (?:(\d\d)(\d{3})|MISG)) (.*)@o) {
                $msg = $4;
                $r->{s} = $1;
                if (defined $2) {
                    $r->{relHumid} = $2 + 0;
                    $r->{distance} = { v => $3 * 100, u => 'FT' };
                } else {
                    $r->{isMissing} = undef;
                }
                push @{$metar{remark}}, { RADAT => $r };
            } elsif ($cy[1] eq 'LF' && $msg =~ m@^([MB])(\d) (.*)@o) {
                $msg = $3;
                push @{$metar{remark}}, { reportConcerns => {
                    s       => "$1$2",
                    change  => $1,
                    subject => $2
                }};
            } elsif ($msg =~ m@^(WI?ND (850 ?FT) ($re_wind)(?: ($re_wind_dir\d)V($re_wind_dir\d))?) (.*)@o)
            {
                $msg = $10;
                push @{$metar{remark}}, _parseWindAtLoc($1, $2, $3, $8, $9);
            } elsif ($msg =~ m@^((2000FT|CNTR RWY|HARBOR|ROOF|BAY TWR) WI?ND ($re_wind)(?: ($re_wind_dir\d)V($re_wind_dir\d))?) (.*)@o)
            {
                $msg = $10;
                push @{$metar{remark}}, _parseWindAtLoc($1, $2, $3, $8, $9);
            } elsif ($msg =~ m@^((BAY TWR|WNDY HILL|KAUKAU) ($re_wind)(?: ($re_wind_dir\d)V($re_wind_dir\d))?) (.*)@o)
            {
                $msg = $10;
                push @{$metar{remark}}, _parseWindAtLoc($1, $2, $3, $8, $9);
            } elsif ($cy[1] eq 'LI') {
                my $re_cond_moun =
                    '(?:LIB|CLD SCT|VERS INC|CNS POST|CLD CIME'
                    . '|CIME INC|GEN INC|INC|INVIS)';
                my $re_chg_moun =
                    '(?:NC|CUF'
                     . '|ELEV (?:SLW|RAPID|STF)'
                     . '|ABB (?:SLW|RAPID)'
                     . '|STF(?: ABB)?'
                     . '|VAR RAPID)';
                my $re_cond_vall =
                    '(?:NIL|FOSCHIA(?: SKC SUP)?|NEBBIA(?: SCT)?'
                    . '|CLD SCT(?: NEBBIA INF)?|MAR CLD|INVIS)';
                my $re_chg_vall =
                    '(?:NC|DIM(?: ELEV| ABB)?|AUM(?: ELEV| ABB)?|ABB'
                    . '|NEBBIA INTER)';
                my $re_moun =
                            "$re_phen_locdirs? $re_cond_moun(?: $re_chg_moun)?";
                my $re_vall =
                            "$re_phen_locdirs? $re_cond_vall(?: $re_chg_vall)?";
                if ($msg =~ m@^(MON((?:$re_moun)+)) (.*)@o) {
                    $msg = $3;
                    $r->{s} = $1;
                    $r->{condMountain} = ();
                    for ($2 =~ m@($re_moun)@og) {
                        my $m;
                        m@($re_phen_locdirs)? ($re_cond_moun)(?: ($re_chg_moun))?@o;

                        $m->{locationAnd}     =_parseLocations $1 if defined $1;
                        ($m->{condMounType}   = $2) =~ tr/ /_/;
                        ($m->{condMounChange} = $3) =~ tr/ /_/ if defined $3;
                        push @{$r->{condMountain}}, $m;
                    }
                    push @{$metar{remark}}, { conditionMountain => $r };
                } elsif ($msg =~ m@^(VAL((?:$re_vall)+)) (.*)@o) {
                    $msg = $3;
                    $r->{s} = $1;
                    $r->{condValley} = ();
                    for ($2 =~ m@($re_vall)@og) {
                        my $m;
                        m@($re_phen_locdirs)? ($re_cond_vall)(?: ($re_chg_vall))?@o;

                        $m->{locationAnd}     =_parseLocations $1 if defined $1;
                        ($m->{condVallType}   = $2) =~ tr/ /_/;
                        ($m->{condVallChange} = $3) =~ tr/ /_/ if defined $3;
                        push @{$r->{condValley}}, $m;
                    }
                    push @{$metar{remark}}, { conditionValley => $r };
                } elsif ($msg =~ m@^(QU(L|K) ?([\d/])(?: ?($re_compass_dir16))?) (.*)@o)
                {
                    my $key = $2 eq 'K' ? 'sea' : 'swell';

                    $msg = $5;
                    $r->{s} = $1;
                    if ($3 eq '/') {
                        $r->{notAvailable} = undef;
                    } else {
                        $r->{"${key}CondVal"} = $3;
                    }
                    $r->{locationAnd} = _parseLocations $4 if defined $4;
                    push @{$metar{remark}}, { "${key}Condition" => $r };
                } elsif ($msg =~ m@^(VIS (MAR) ([1-9]\d*) KM) (.*)@o) {
                    $msg = $4;
                    $r->{s} = $1;
                    $r->{locationAt} = $2;
                    $r->{visibility}{distance} = { v => $3, u => 'KM' };
                    push @{$metar{remark}}, { visibilityAtLoc => $r };
                } elsif ($msg =~ m@^((?:WIND THR ?|WT)($re_rwy_des) ($re_wind)) (.*)@o)
                {
                    $msg = $8;
                    push @{$metar{remark}}, { thrWind => {
                        s        => $1,
                        rwyDesig => $2,
                        wind     => _parseWind $3
                    }};
                } elsif ($msg =~ m@^(VIS MIN ($re_vis_m)($re_compass_dir)?) (.*)@o) {
                    $msg = $4;
                    $r->{s}          = $1;
                    $r->{distance}   = { v => $2 + 0, u => 'M' };
                    $r->{compassDir} = $3 if defined $3;
                    push @{$metar{remark}}, { visMin => $r };
                } else {
                    $parsed = 0;
                }
            } elsif($cy[1] eq 'LK' && $msg =~ m@^(REG QNH ([01]\d{3})) (.*)@)
            {
                $msg = $3;
                push @{$metar{remark}}, { regQNH => {
                    s   => $1,
                    hPa => $2
                }};
            } elsif ($cy[0] eq 'ZMUB' && $msg =~ m@^(\d\d) (.*)@) {
                $msg = $2;
                push @{$metar{remark}}, { RH => {
                    s        => $1,
                    relHumid => $1
                }};
            } else {
                $parsed = 0;
            }
            if (!$parsed && _isValidCountry('remarkC_K_P_NS_RJ_RO_TI_TJ_ET', \@cy))
            {
                $parsed = 1;
                if ($msg =~ m@^(4([01]\d{3})([01]\d{3})?) (.*)@o) {
                    $msg = $4;
                    $r->{s} = $1;
                    $r->{tempMaxPeriod}{s}      = $2;
                    $r->{tempMaxPeriod}{period} = '24h';
                    $r->{tempMaxPeriod}{temp}   = _parseTemp $2;
                    if (defined $3) {
                        $r->{tempMinPeriod}{s}      = $3;
                        $r->{tempMinPeriod}{period} = '24h';
                        $r->{tempMinPeriod}{temp}   = _parseTemp $3;
                    }
                    push @{$metar{remark}}, { temp24h => $r };
                } elsif ($msg =~ m@^(4/(\d{3})) (.*)@o) {
                    $msg = $3;
                    push @{$metar{remark}}, { snowOnGround => {
                        s            => $1,
                        precipAmount => { v => $2 + 0, u => 'IN' }
                    }};
                } elsif (   !_isValidCountry('remarkPrecipHourly3', \@cy)
                         && $msg =~ m@^((P|6|7)(?:(\d{4})|////)) (.*)@o)
                {
                    # Air Force Manual 15-111, Dec. '03, 2.8.1
                    $msg = $4;
                    $r->{s} = $1;
                    if (defined $3) {
                        $r->{precipAmount} = { v => $3 / 100, u => 'IN' };
                    } else {
                        $r->{precipAmountNotAvailable} = undef;
                    }
                    if ($2 eq 'P') {
                        $r->{period} = '1h';
                    } elsif ($2 eq '7') {
                        $r->{period} = '24h';
                    } else {
                        # EXTENSION: allow 20 minutes
                        if (exists $metar{obsTime}{hour}) {
                            if ((  $metar{obsTime}{hour}
                                 . $metar{obsTime}{minute}) =~
                                '(?:(?:23|05|11|17)[45]|(?:00|06|12|18)[01])\d')
                            {
                                $r->{period} = '6h';
                            } elsif (  ($metar{obsTime}{hour}
                                     . $metar{obsTime}{minute}) =~
                                '(?:(?:02|08|14|20)[45]|(?:03|09|15|21)[01])\d')
                            {
                                $r->{period} = '3h';
                            } else {
                                $r->{period3o6h} = undef;
                            }
                        } else {
                            $r->{period3o6h} = undef;
                        }
                    }
                    push @{$metar{remark}}, { precipHourly => $r };
                # EXTENSION: allow slash after 933 (PATA)
                } elsif ($msg =~ m@^(933/?(\d{3})) (.*)@o) {
                    $msg = $3;
                    push @{$metar{remark}}, { waterEquivOfSnow => {
                        s            => $1,
                        precipAmount => { v => $2 / 10, u => 'IN' }
                    }};
                } elsif ($msg =~ m@^(CIG (\d{3})V(\d{3})) (.*)@o) {
                    $msg = $4;
                    push @{$metar{remark}}, { variableCeiling => {
                        s => $1,
                        cloudBaseFrom => { v => $2 * 100, u => 'FT' },
                        cloudBaseTo   => { v => $3 * 100, u => 'FT' }
                    }};
                } elsif ($msg =~ m@^(CIG (\d{4})V(\d{4})) (.*)@o) {
                    $msg = $4;
                    push @{$metar{remark}}, { variableCeiling => {
                        s => $1,
                        cloudBaseFrom => { v => $2 + 0, u => 'FT' },
                        cloudBaseTo   => { v => $3 + 0, u => 'FT' }
                    }};
                } elsif ($msg =~ m@^(($re_be_prec+)($re_phen_locdirs)?$re_wx_mov_d2?) (.*)@o)
                {
                    $msg = $6;

                    $r->{beginEndPrecip}{s} = $1;
                    $r->{beginEndPrecip}{locationAnd} = _parseLocations $3
                        if defined $3;
                    $r->{beginEndPrecip}{$4}{locationAnd} = _parseLocations $5
                        if defined $5;
                    $r->{beginEndPrecip}{precip} = ();
                    for ($2 =~ /($re_be_prec)+?/g) {
                        my (@startEndArr, $precip);
                        my ($weather, $times) = /(.*?)($re_be_prec_be+)/g;
                        for ($times =~ /$re_be_prec_be+?/g) {
                            my %s_e;
                            /(.)(..)?(..)/;
                            $s_e{hour} = $2 if defined $2;
                            $s_e{minute} = $3;
                            if ($1 eq 'B') {
                                push @startEndArr, { startTime => \%s_e };
                            } else {
                                push @startEndArr, { endTime => \%s_e };
                            }
                        }
                        $precip = _parseWeather $weather;
                        delete $precip->{s};
                        $precip->{startEndArr} = \@startEndArr;
                        push @{$r->{beginEndPrecip}{precip}}, $precip;
                    }
                    push @{$metar{remark}}, $r;
                } elsif ($msg =~ m@^WSHFT (($re_hour)?($re_min)( FROPA)?) (.*)@o)
                {
                    $msg = $5;
                    $r->{windShift}{s} = "WSHFT $1";
                    $r->{windShift}{hour} = $2 if defined $2;
                    $r->{windShift}{minute} = $3;
                    $r->{windShift}{FROPA} = undef if defined $4;
                    push @{$metar{remark}}, $r;
                } elsif ($msg =~ m@^(VIS ($re_vis_sm) ($re_compass_dir16)) (.*)@o)
                {
                    my (@visLocData, $v);

                    $msg = $4;
                    $r->{visListAtLoc}{s} = $1;
                    $v->{visibility} = {};
                    _parseFraction $v->{visibility}, $2, 'SM';
                    $r->{visListAtLoc}{visLocData} = \@visLocData;
                    $v->{locationAnd} = _parseLocations $3;
                    push @visLocData, $v;
                    push @{$metar{remark}}, $r;
                } elsif ($msg =~ m@^VIS($re_phen_locdirs $re_vis_sm(?: SM)?(?:(?: AND)?$re_phen_locdirs $re_vis_sm(?: SM)?)*) (.*)@o)
                {
                    my @visLocData;

                    $msg = $2;
                    $r->{visListAtLoc}{s} = "VIS$1";
                    $r->{visListAtLoc}{visLocData} = \@visLocData;
                    for ($1 =~ m@(?:AND)?($re_phen_locdirs $re_vis_sm)@og) {
                        my ($v, $is_less, $vis);

                        m@($re_phen_locdirs) ($re_vis_sm)@o;
                        $v->{locationAnd} = _parseLocations $1;
                        $v->{visibility} = {};
                        _parseFraction $v->{visibility}, $2, 'SM';
                        push @visLocData, $v;
                    }
                    push @{$metar{remark}}, $r;
                } else {
                    $parsed = 0;
                }
            }
            if (!$parsed && $msg =~ m@^(BA (?:GOOD|POOR)|THN SPTS IOVC|FUOCTY|ALTM MISG|ALL WNDS GRID|CONTRAILS?|(?:SKY|MT(?:N?S)?|RDGS) OBSC(?:URED|D)?|FOGGY|TWLGT|WEA) (.*)@o)
            {
                $parsed = 1;
                $msg = $2;
                $r->{s} = $1;
                ($r->{v} = $1) =~ tr/\/ /_/;
                $r->{v} =~ s/CONTRAILS?/CONTRAILS/;
                $r->{v} =~ s/OBSC(?:URE)?D/OBSC/;
                $r->{v} =~ s/MT(?:N?S)?/MTNS/;
                push @{$metar{remark}}, { keyword => $r };
            }
            if (!$parsed) {
                $msg =~ '(\S+) (.*)';
                $notRecognised .= ' ' unless $notRecognised eq '';
                $notRecognised .= $1;
                $msg = $2;
            }
            if ($parsed && $notRecognised ne '') {
                my $top = pop @{$metar{remark}};
                push @{$metar{remark}},
                                    { notRecognised => { s => $notRecognised }};
                push @{$metar{remark}}, $top;
                $notRecognised = '';
            }
        }
        push @{$metar{remark}}, { notRecognised => { s => $notRecognised }}
            if $notRecognised ne '';
    }

    $metar{ERROR} = _makeErrorMsg 'other', $corr_msg, $msg if $msg ne '';
    return %metar;
}

=head1 DEPENDENCIES

The function C<POSIX::floor> from the module C<POSIX> is required.

=head1 FUNCTIONS

=cut

########################################################################
# parseReport
########################################################################
sub parseReport {
    my ($msg, $is_taf, $is_synop) = @_;
    my ($is_speci, $msg_hdr);

=head2 parseReport($msg,$is_taf,$is_synop)

This is the main function of the module.

The following arguments are expected:

=over

=item C<msg>

string that contains the message

=item C<is_taf>

boolean value, indicating whether it is a TAF message or not

=item C<is_synop>

boolean value, indicating whether it is a SYNOP message or not

=back

Leading and trailing spaces are removed from C<msg>, multiple spaces are
replaced by a single one. Characters that are invalid in HTML or XML are also
removed.

=cut

    $msg =~ s/ +/ /g;
    $msg =~ s/ $//;
    $msg =~ s/^ //;
    $msg =~ s/[^ -~]/?/gos; # avoid invalid XML and HTML

=pod

If the message starts with C<METAR>, C<SPECI>, C<TAF>, or C<SYNOP>, the
arguments C<is_taf> and C<is_synop> are ignored.

=cut

    $msg_hdr = '';
    if ($msg =~ s/^(METAR )//) {
        $is_taf = 0;
        $is_synop = 0;
        $msg_hdr = $1;
    } elsif ($msg =~ s/^(SPECI )//) {
        $is_taf = 0;
        $is_synop = 0;
        $msg_hdr = $1;
        $is_speci = 1;
    } elsif ($msg =~ s/^(TAF )//) {
        $is_taf = 1;
        $is_synop = 0;
        $msg_hdr = $1;
    } elsif ($msg =~ s/^(SYNOP )//) {
        $is_synop = 1;
        $msg_hdr = $1;
    }

=pod

Then the correct function to parse the message is called.

=cut

    return _parseSynop $msg, $msg_hdr if $is_synop;
    return _parseMetarTaf $msg, $msg_hdr, $is_speci, $is_taf;
}

=pod

The return value is a hash with all the parsed components of the message. It
may also have values for the following keys:

=over

=item C<ERROR>

If a message could not be parsed C<ERROR> is a hash with the the keys
C<errorType> (the type of error) and C<s> - the original message with the
position of the error marked as C<< <@> >>.

=item C<warning>

An array of hashes describing problems which where encountered during
parsing but which didn't prevent complete parsing.

=back

=head1 SEE ALSO

=begin html

<a href=XML.pm.html>metaf2xml::XML(3pm)</a>,
<a href=metaf2xml.pl.html>metaf2xml(1)</a>,
<a href=metaf.pl.html>metaf(1),</a> <!--

=end html

B<metaf2xml::XML>(3pm),
B<metaf2xml>(1),
B<metaf>(1),

=for html -->

L<http://metaf2xml.sourceforge.net/>

=head1 COPYRIGHT, LICENSE, DISCLAIMER

copyright (c) 2006-2008 metaf2xml @ L<http://metaf2xml.sourceforge.net/>

This program is free software; you can redistribute it and/or
modify it under the terms of the GNU General Public License as
published by the Free Software Foundation; either version 2 of the
License, or (at your option) any later version.

This program is distributed in the hope that it will be useful, but
WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
General Public License for more details.

You should have received a copy of the GNU General Public License
along with this program; if not, write to the Free Software
Foundation, 59 Temple Place, Suite 330, Boston, MA 02111-1307, USA

=cut

1;
